/**
 * 
 */
package com.github.ansell.oas.objects.impl;

import java.util.Collection;

import org.openrdf.model.Literal;
import org.openrdf.model.Model;
import org.openrdf.model.Resource;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.impl.LinkedHashModel;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.RDFS;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.utils.OasConstants;

/**
 * An implementation of the Annotation object API.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class AnnotationImpl implements Annotation
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private URI annotatedObjectType;
    private URI annotatedObjectUri;
    private URI annotationUniqueId;
    private URI annotator;
    private String annotatorLabel;
    
    private URI ontologyTermUri;
    private Literal tag;
    private URI ontologyUri;
    
    /**
     * Default constructor, values can be set using setProperty methods.
     */
    public AnnotationImpl()
    {
        // TODO Auto-generated constructor stub
    }
    
    /**
     * Assumes that the annotation will be the only annotation in the collection and that it will
     * have a defined URI as the subject of the relevant RDF statements, not a blank node.
     * 
     * @param annotationStatements
     *            A Collection of OpenRDF Sesame RDF statements that define a single annotation.
     */
    public AnnotationImpl(final Collection<Statement> annotationStatements)
    {
        boolean isValid = false;
        // TODO: Convert this algorithm to use Model.filter
        for(final Statement nextAnnotationStatement : annotationStatements)
        {
            if(!(nextAnnotationStatement.getSubject() instanceof URI))
            {
                throw new IllegalArgumentException(
                        "Cannot use this constructor if the relevant triples do not have URIs as subjects");
            }
            
            if(nextAnnotationStatement.getPredicate().equals(RDF.TYPE)
                    && nextAnnotationStatement.getObject().equals(OasConstants.TAGGING_TAGGING))
            {
                isValid = true;
                
                this.setAnnotationUniqueId((URI)nextAnnotationStatement.getSubject());
            }
            else if(nextAnnotationStatement.getPredicate().equals(OasConstants.FOAF_MAKER))
            {
                this.setAnnotator((URI)nextAnnotationStatement.getObject());
            }
            else if(nextAnnotationStatement.getPredicate().equals(OasConstants.TAGGING_NAME))
            {
                this.setTag((Literal)nextAnnotationStatement.getObject());
            }
            else if(nextAnnotationStatement.getPredicate().equals(OasConstants.MOAT_TAG_MEANING))
            {
                this.setOntologyTermUri((URI)nextAnnotationStatement.getObject());
            }
            else if(nextAnnotationStatement.getPredicate().equals(OasConstants.TAGGING_TAGGED_RESOURCE))
            {
                this.setAnnotatedObjectUri((URI)nextAnnotationStatement.getObject());
            }
            // this is for RDF type declarations that did not occur on the tag subject URI
            // This is after the tagging:Tagging check above so that it will only match other items
            else if(nextAnnotationStatement.getPredicate().equals(RDF.TYPE))
            {
                this.setAnnotatedObjectType((URI)nextAnnotationStatement.getObject());
            }
        }
        
        if(!isValid)
        {
            throw new IllegalArgumentException("The list of statements did not contain an instance of an annotation");
        }
    }
    
    /**
     * Constructor to use when creating annotations where the subject of the triples needs to be
     * replaced.
     * 
     * Warning: The list of annotation statements can only contain triples about a single
     * annotation!
     * 
     * @param annotationStatements
     *            The collection of statements representing the annotation to create.
     * @param nextSubject
     *            The expected subject URI to find in the statements.
     * @param replacementSubjectUri
     *            The replacement URI. If the URI is null, then nextSubject will be used as the
     *            annotations unique ID.
     */
    public AnnotationImpl(final Model annotationStatements, final Resource nextSubject, final URI replacementSubjectUri)
    {
        Model statements;
        
        if(annotationStatements instanceof Model)
        {
            statements = annotationStatements;
        }
        else
        {
            statements = new LinkedHashModel(annotationStatements);
        }
        
        // DebugUtils.printStatements(statements);
        
        if(!statements.contains(nextSubject, RDF.TYPE, OasConstants.TAGGING_TAGGING))
        {
            throw new IllegalArgumentException("Could not find a unique annotation target in the given statements.");
        }
        
        // if they specified a replacement URI, use it here
        if(replacementSubjectUri != null)
        {
            this.setAnnotationUniqueId(replacementSubjectUri);
        }
        else
        {
            this.setAnnotationUniqueId((URI)nextSubject);
        }
        
        this.setAnnotator(statements.filter(nextSubject, OasConstants.FOAF_MAKER, null).objectURI());
        this.setTag(statements.filter(nextSubject, OasConstants.TAGGING_NAME, null).objectLiteral());
        this.setOntologyTermUri(statements.filter(nextSubject, OasConstants.MOAT_TAG_MEANING, null).objectURI());
        this.setAnnotatedObjectUri(statements.filter(nextSubject, OasConstants.TAGGING_TAGGED_RESOURCE, null)
                .objectURI());
        
        // Optional
        if(statements.contains(this.getOntologyTermUri(), RDFS.ISDEFINEDBY, null))
        {
            this.setOntologyUri(statements.filter(this.getOntologyTermUri(), RDFS.ISDEFINEDBY, null).objectURI());
            
            this.log.debug("Found RDFS.ISDEFINEDBY relation for term: {}, {}", this.getOntologyTermUri(),
                    this.getOntologyUri());
        }
        else
        {
            this.log.debug("Did not find RDFS.ISDEFINEDBY relation for term: {}", this.getOntologyTermUri());
        }
        
        // Optional
        if(statements.contains(this.getAnnotatedObjectUri(), RDF.TYPE, null))
        {
            this.setAnnotatedObjectType(statements.filter(this.getAnnotatedObjectUri(), RDF.TYPE, null).objectURI());
        }
    }
    
    public AnnotationImpl(final Model annotationStatements, final URI nextSubject)
    {
        this(annotationStatements, nextSubject, null);
    }
    
    @Override
    public URI getAnnotatedObjectType()
    {
        return this.annotatedObjectType;
    }
    
    @Override
    public URI getAnnotatedObjectUri()
    {
        return this.annotatedObjectUri;
    }
    
    @Override
    public URI getAnnotationUniqueId()
    {
        return this.annotationUniqueId;
    }
    
    @Override
    public URI getAnnotator()
    {
        return this.annotator;
    }
    
    @Override
    public String getAnnotatorLabel()
    {
        return this.annotatorLabel;
    }
    
    @Override
    public URI getOntologyTermUri()
    {
        return this.ontologyTermUri;
    }
    
    @Override
    public URI getOntologyUri()
    {
        return this.ontologyUri;
    }
    
    @Override
    public Literal getTag()
    {
        return this.tag;
    }
    
    @Override
    public void setAnnotatedObjectType(final URI nextAnnotatedObjectType)
    {
        this.annotatedObjectType = nextAnnotatedObjectType;
    }
    
    @Override
    public void setAnnotatedObjectUri(final URI nextAnnotatedObjectUri)
    {
        this.annotatedObjectUri = nextAnnotatedObjectUri;
    }
    
    @Override
    public void setAnnotationUniqueId(final URI nextAnnotationUniqueId)
    {
        this.annotationUniqueId = nextAnnotationUniqueId;
    }
    
    @Override
    public void setAnnotator(final URI nextAnnotator)
    {
        this.annotator = nextAnnotator;
    }
    
    @Override
    public void setAnnotatorLabel(final String annotatorLabel)
    {
        this.annotatorLabel = annotatorLabel;
    }
    
    @Override
    public void setOntologyTermUri(final URI nextOntologyUri)
    {
        this.ontologyTermUri = nextOntologyUri;
    }
    
    @Override
    public void setOntologyUri(final URI ontologyUri)
    {
        this.ontologyUri = ontologyUri;
    }
    
    @Override
    public void setTag(final Literal nextTag)
    {
        this.tag = nextTag;
    }
    
    @Override
    public Model toRdf()
    {
        final Model result = new LinkedHashModel();
        
        result.add(this.getAnnotationUniqueId(), RDF.TYPE, OasConstants.TAGGING_TAGGING, this.getAnnotationUniqueId());
        
        if(this.getTag() != null)
        {
            result.add(this.getAnnotationUniqueId(), OasConstants.TAGGING_NAME, this.getTag(),
                    this.getAnnotationUniqueId());
        }
        
        if(this.getOntologyTermUri() != null)
        {
            result.add(this.getAnnotationUniqueId(), OasConstants.MOAT_TAG_MEANING, this.getOntologyTermUri(),
                    this.getAnnotationUniqueId());
            
            if(this.getOntologyUri() != null)
            {
                result.add(this.getOntologyTermUri(), RDFS.ISDEFINEDBY, this.getOntologyUri(),
                        this.getAnnotationUniqueId());
            }
        }
        
        if(this.getAnnotatedObjectUri() != null)
        {
            result.add(this.getAnnotationUniqueId(), OasConstants.TAGGING_TAGGED_RESOURCE,
                    this.getAnnotatedObjectUri(), this.getAnnotationUniqueId());
        }
        
        if(this.getAnnotator() != null)
        {
            result.add(this.getAnnotationUniqueId(), OasConstants.FOAF_MAKER, this.getAnnotator(),
                    this.getAnnotationUniqueId());
        }
        
        if(this.getAnnotatedObjectUri() != null && this.getAnnotatedObjectType() != null)
        {
            result.add(this.getAnnotatedObjectUri(), RDF.TYPE, this.getAnnotatedObjectType(),
                    this.getAnnotationUniqueId());
        }
        
        return result;
    }
}
