/**
 * 
 */
package com.github.ansell.oas.objects.impl;

import java.util.ArrayList;
import java.util.List;

import org.openrdf.model.Model;
import org.openrdf.model.Resource;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.impl.ValueFactoryImpl;
import org.openrdf.model.vocabulary.OWL;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryException;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyID;
import org.semanticweb.owlapi.profiles.OWLProfile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.utils.OasConstants;
import com.github.ansell.oas.utils.SesameUtils;

/**
 * An implementation of the Ontology interface.
 * 
 * Note that the compareTo, equals, and hashCode methods all only use ontologyUri and
 * ontologyVersionUri to determine equality.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class OntologyImpl implements Ontology
{
    private static final Logger LOG = LoggerFactory.getLogger(OntologyImpl.class);
    
    private volatile URI ontologyUri;
    
    private volatile URI ontologyVersionUri;
    
    private URI inferredOntologyUri;
    
    private List<OWLProfile> profiles = new ArrayList<OWLProfile>(8);
    
    /**
     * Defaults to an OWL:Class variable property path query.
     */
    private String recursiveChildQuery =
            " ?child a <http://www.w3.org/2002/07/owl#Class> . ?child <http://www.w3.org/2000/01/rdf-schema#subClassOf>+ ?parent . FILTER(isIRI(?child) && isIRI(?parent)) . ";
    // "?child a <http://www.w3.org/2002/07/owl#Class> . ?child <http://www.w3.org/2000/01/rdf-schema#subClassOf>+ ?parent . FILTER(isIRI(?child) && isIRI(?parent)) . OPTIONAL{ ?child <http://www.w3.org/2000/01/rdf-schema#label> ?label . }";
    
    /**
     * Defaults to an OWL:Class single result query.
     */
    private String directChildQuery =
            " ?child a <http://www.w3.org/2002/07/owl#Class> . ?child <http://www.w3.org/2000/01/rdf-schema#subClassOf> ?parent . FILTER(isIRI(?child) && isIRI(?parent)) . ";
    // "?child a <http://www.w3.org/2002/07/owl#Class> . ?child <http://www.w3.org/2000/01/rdf-schema#subClassOf> ?parent . FILTER(isIRI(?child) && isIRI(?parent)) . OPTIONAL{ ?child <http://www.w3.org/2000/01/rdf-schema#label> ?label . }";
    
    private int concreteAxiomCount = 0;
    
    private int inferredAxiomCount = 0;
    
    private int concreteLogicalAxiomCount = 0;
    
    private int individualCount = 0;
    
    public OntologyImpl()
    {
    }
    
    public OntologyImpl(final OWLOntology ontology)
    {
        this(ontology.getOntologyID());
    }
    
    public OntologyImpl(final OWLOntologyID ontologyID)
    {
        if(ontologyID == null)
        {
            throw new IllegalArgumentException("Given ontology did not contain an OWLOntologyID");
        }
        
        if(ontologyID.isAnonymous())
        {
            throw new IllegalArgumentException("Cannot create an Ontology object based on an Anonymous OWLOntologyID");
        }
        
        if(ontologyID.getOntologyIRI() == null)
        {
            throw new IllegalStateException(
                    "Given OWLOntologyID was not anonymous and did not contain an OntologyIRI, this should not occur");
        }
        
        this.setOntologyUri(ValueFactoryImpl.getInstance().createURI(ontologyID.getOntologyIRI().toString()));
        
        if(ontologyID.getVersionIRI() != null)
        {
            this.setOntologyVersionUri(ValueFactoryImpl.getInstance().createURI(ontologyID.getVersionIRI().toString()));
        }
    }
    
    public OntologyImpl(final URI nextOntologyUri)
    {
        this.setOntologyUri(nextOntologyUri);
    }
    
    public OntologyImpl(final URI nextOntologyUri, final URI nextOntologyVersionUri)
    {
        this(nextOntologyUri);
        this.setOntologyVersionUri(nextOntologyVersionUri);
    }
    
    @Override
    public int compareTo(final Ontology o)
    {
        // sort null ontologies first
        if(o == null)
        {
            return 1;
        }
        
        if(o == this || o.equals(this))
        {
            return 0;
        }
        
        if(this.getOntologyUri().equals(o.getOntologyUri()))
        {
            if(this.getOntologyVersionUri().equals(o.getOntologyVersionUri()))
            {
                return 0;
            }
            else
            {
                return this.getOntologyVersionUri().stringValue().compareTo(o.getOntologyVersionUri().stringValue());
            }
        }
        else
        {
            return this.getOntologyUri().stringValue().compareTo(o.getOntologyUri().stringValue());
        }
    }
    
    @Override
    public boolean ensureVersionUriUniqueness(final Repository ontologyManagerRepository,
            final boolean forceValidationOfLocalVersionAgainstRepository, final boolean useLocalVersionIfSet,
            final URI... contexts)
    {
        if(this.ontologyUri == null && this.ontologyVersionUri == null)
        {
            // make sure that if both ontology URI and ontology version URI are null that the
            // managed ontology version URI is null as well
            // this.managedOntologyVersionUri = null;
            return false;
        }
        else if(useLocalVersionIfSet && this.ontologyVersionUri != null)
        {
            // this.managedOntologyVersionUri = this.ontologyVersionUri;
            return false;
        }
        else if(forceValidationOfLocalVersionAgainstRepository || !useLocalVersionIfSet
                || this.ontologyVersionUri == null)
        {
            try
            {
                final URI updatedUri =
                        SesameUtils.generateNewOntologyVersionUri(this.ontologyUri, this.ontologyVersionUri,
                                ontologyManagerRepository, contexts);
                if(updatedUri.equals(this.ontologyVersionUri))
                {
                    return false;
                }
                else
                {
                    this.ontologyVersionUri = updatedUri;
                    return true;
                }
            }
            catch(final RepositoryException rex)
            {
                OntologyImpl.LOG.error(
                        "Found repository exception when trying to validate managed ontology version URI", rex);
                // nullify the managed ontology version URI so users do not think that we have
                // authorised it when the process failed
                // this.managedOntologyVersionUri = null;
                return false;
            }
        }
        
        return false;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(final Object obj)
    {
        if(this == obj)
        {
            return true;
        }
        if(obj == null)
        {
            return false;
        }
        if(!(obj instanceof Ontology))
        {
            return false;
        }
        final Ontology other = (Ontology)obj;
        if(this.ontologyUri == null)
        {
            if(other.getOntologyUri() != null)
            {
                return false;
            }
        }
        else if(!this.ontologyUri.equals(other.getOntologyUri()))
        {
            return false;
        }
        if(this.ontologyVersionUri == null)
        {
            if(other.getOntologyVersionUri() != null)
            {
                return false;
            }
        }
        else if(!this.ontologyVersionUri.equals(other.getOntologyVersionUri()))
        {
            return false;
        }
        return true;
    }
    
    @Override
    public int getConcreteAxiomCount()
    {
        return this.concreteAxiomCount;
    }
    
    @Override
    public int getConcreteLogicalAxiomCount()
    {
        return this.concreteLogicalAxiomCount;
    }
    
    @Override
    public URI[] getContextUris()
    {
        if(this.ontologyVersionUri == null)
        {
            throw new IllegalStateException("Cannot call this method until an ontology version URI has been setup.");
        }
        
        if(this.inferredOntologyUri != null)
        {
            return new URI[] { this.ontologyVersionUri, this.inferredOntologyUri };
        }
        else
        {
            return new URI[] { this.ontologyVersionUri };
        }
    }
    
    @Override
    public String getDirectChildQuery()
    {
        return this.directChildQuery;
    }
    
    @Override
    public int getIndividualCount()
    {
        return this.individualCount;
    }
    
    @Override
    public int getInferredAxiomCount()
    {
        return this.inferredAxiomCount;
    }
    
    @Override
    public URI getInferredOntologyUri()
    {
        return this.inferredOntologyUri;
    }
    
    @Override
    public URI getOntologyUri()
    {
        return this.ontologyUri;
    }
    
    @Override
    public URI getOntologyVersionUri()
    {
        return this.ontologyVersionUri;
    }
    
    @Override
    public List<OWLProfile> getProfiles()
    {
        return this.profiles;
    }
    
    @Override
    public String getRecursiveChildQuery()
    {
        return this.recursiveChildQuery;
    }
    
    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((this.ontologyUri == null) ? 0 : this.ontologyUri.hashCode());
        result = prime * result + ((this.ontologyVersionUri == null) ? 0 : this.ontologyVersionUri.hashCode());
        return result;
    }
    
    @Override
    public void setConcreteAxiomCount(final int axiomCount)
    {
        this.concreteAxiomCount = axiomCount;
    }
    
    @Override
    public void setConcreteLogicalAxiomCount(final int logicalAxiomCount)
    {
        this.concreteLogicalAxiomCount = logicalAxiomCount;
    }
    
    @Override
    public void setDirectChildQuery(final String nextChildSearchQueryPattern)
    {
        this.directChildQuery = nextChildSearchQueryPattern;
    }
    
    @Override
    public void setIndividualCount(final int individualCount)
    {
        this.individualCount = individualCount;
    }
    
    @Override
    public void setInferredAxiomCount(final int axiomCount)
    {
        this.inferredAxiomCount = axiomCount;
    }
    
    @Override
    public void setInferredOntologyUri(final URI inferredOntologyUri)
    {
        this.inferredOntologyUri = inferredOntologyUri;
    }
    
    @Override
    public void setOntologyDetails(final Model ontologyGraph)
    {
        if(this.ontologyUri == null)
        {
            final Model filtered = ontologyGraph.filter(null, RDF.TYPE, OWL.ONTOLOGY);
            
            // Keep iterating through until we get to the end of the list of matches or we find a
            // URI in the subject position
            for(final Resource nextSubject : filtered.subjects())
            {
                if(nextSubject instanceof URI)
                {
                    this.setOntologyUri((URI)nextSubject);
                    // break out to avoid setting the ontology URI again
                    break;
                }
            }
        }
        
        // if the ontology URI has been set or discovered at this point, then we can use it to
        // autodiscover the versionIRI
        if(this.ontologyVersionUri == null && this.ontologyUri != null)
        {
            final Model filtered = ontologyGraph.filter(this.ontologyUri, OasConstants.OWL_VERSIONIRI, null);
            
            // FIXME: Need better strategy for finding the version given a list of statements
            // Keep iterating through until we get to the end of the list of matches or we find a
            // URI in the object position
            for(final Value nextObject : filtered.objects())
            {
                if(nextObject instanceof URI)
                {
                    this.setOntologyVersionUri((URI)nextObject);
                    // break out to avoid setting the ontology version URI again
                    break;
                }
            }
        }
        
        // if the ontology URI was set, but the ontology version URI was not, then assign the
        // version a new URI to it at this point
        if(this.ontologyUri != null && this.ontologyVersionUri == null)
        {
            this.ontologyVersionUri =
                    ValueFactoryImpl.getInstance().createURI(OasConstants.AUTOGENERATED_NS,
                            this.ontologyUri.stringValue());
        }
        
        if(ontologyGraph.contains(this.ontologyVersionUri, OasConstants.OAS_CONCRETE_AXIOM_COUNT, null))
        {
            this.setConcreteAxiomCount(ontologyGraph
                    .filter(this.ontologyVersionUri, OasConstants.OAS_CONCRETE_AXIOM_COUNT, null).objectLiteral()
                    .intValue());
        }
        
        if(ontologyGraph.contains(this.ontologyVersionUri, OasConstants.OAS_CONCRETE_LOGICAL_AXIOM_COUNT, null))
        {
            this.setConcreteLogicalAxiomCount(ontologyGraph
                    .filter(this.ontologyVersionUri, OasConstants.OAS_CONCRETE_LOGICAL_AXIOM_COUNT, null)
                    .objectLiteral().intValue());
        }
        
        if(ontologyGraph.contains(this.ontologyVersionUri, OasConstants.OAS_INDIVIDUAL_COUNT, null))
        {
            this.setIndividualCount(ontologyGraph
                    .filter(this.ontologyVersionUri, OasConstants.OAS_INDIVIDUAL_COUNT, null).objectLiteral()
                    .intValue());
        }
        
        if(ontologyGraph.contains(this.ontologyVersionUri, OasConstants.OAS_INFERRED_AXIOM_COUNT, null))
        {
            this.setInferredAxiomCount(ontologyGraph
                    .filter(this.ontologyVersionUri, OasConstants.OAS_INFERRED_AXIOM_COUNT, null).objectLiteral()
                    .intValue());
        }
        
        // null this out so that it will be revalidated the next time someone tries to access it
        // this.managedOntologyVersionUri = null;
    }
    
    @Override
    public void setOntologyUri(final URI nextOntologyUri)
    {
        if(nextOntologyUri == null)
        {
            throw new IllegalArgumentException("Ontology URI cannot be null.");
        }
        
        if(this.ontologyVersionUri != null && nextOntologyUri.equals(this.ontologyVersionUri))
        {
            throw new IllegalStateException("Cannot use same URI for both ontology and version.");
        }
        
        this.ontologyUri = nextOntologyUri;
    }
    
    @Override
    public void setOntologyVersionUri(final URI nextOntologyVersionUri)
    {
        if(nextOntologyVersionUri == null)
        {
            throw new IllegalArgumentException("Ontology version URI cannot be null.");
        }
        
        if(this.ontologyUri != null && nextOntologyVersionUri.equals(this.ontologyUri))
        {
            throw new IllegalStateException("Cannot use same URI for both ontology and version.");
        }
        
        this.ontologyVersionUri = nextOntologyVersionUri;
    }
    
    @Override
    public void setProfiles(final List<OWLProfile> matchingProfiles)
    {
        this.profiles = matchingProfiles;
    }
    
    @Override
    public void setRecursiveChildQuery(final String nextChildSearchQueryPattern)
    {
        this.recursiveChildQuery = nextChildSearchQueryPattern;
    }
    
    @Override
    public OWLOntologyID toOWLOntologyID()
    {
        if(this.ontologyUri == null && this.ontologyVersionUri == null)
        {
            return new OWLOntologyID();
        }
        else if(this.ontologyVersionUri == null)
        {
            return new OWLOntologyID(this.ontologyUri);
        }
        else if(this.ontologyUri == null)
        {
            return new OWLOntologyID(this.ontologyVersionUri, this.ontologyVersionUri);
        }
        else
        {
            return new OWLOntologyID(this.ontologyUri, this.ontologyVersionUri);
        }
    }
    
    @Override
    public String toString()
    {
        if(this.ontologyUri == null && this.ontologyVersionUri == null)
        {
            return super.toString();
        }
        else if(this.ontologyVersionUri != null)
        {
            return this.ontologyUri.stringValue() + " [" + this.ontologyVersionUri.stringValue() + "]";
        }
        else
        {
            return this.ontologyUri.stringValue() + "[no version]";
        }
    }
}
