/**
 * 
 */
package com.github.ansell.oas.tsv;

import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.List;

import org.openrdf.model.BNode;
import org.openrdf.model.Model;
import org.openrdf.model.ValueFactory;
import org.openrdf.model.impl.LinkedHashModel;
import org.openrdf.model.impl.ValueFactoryImpl;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.RDFS;
import org.openrdf.model.vocabulary.XMLSchema;

import au.com.bytecode.opencsv.CSVReader;

import com.github.ansell.oas.utils.OasConstants;

/**
 * Contains Tab Seperated Values utilities to convert TSV files representing Annotations to RDF
 * files which can be used to communicate with an OAS server using the BulkUpdateAnnotationResource.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public final class OasTsvUtility
{
    /**
     * The list of expected fields in the TSV file:
     * 
     * <ol>
     * <li>URI for annotated object : Required</li>
     * <li>URI for ontology term : Required</li>
     * <li>String containing content for anontation : Required</li>
     * <li>URI for annotator : Optional, will default to the user who is logged in when the
     * annotation is submitted if it is not given</li>
     * <li>URI for ontology : Optional, can specify a particular version of an ontology if necessary
     * </li>
     * <li>URI for the type of the annotated object : Optional, can be used to filter annotations
     * based on the type of object they are annotating.</li>
     * </ol>
     */
    public static final List<String> EXPECTED_FIELDS = Arrays.asList("annotated_object", "ontology_term",
            "annotation_content", "annotator", "ontology_uri", "annotated_object_type");
    
    /**
     * Reads annotations from a TSV file, skipping the first line if necessary, and returns the
     * Annotations as RDF statements in a {@link Model}.
     * 
     * @param inputStream
     *            The TSV file to read annotations from, as an {@link InputStream}.
     * @param skipFirstLine
     *            If true the first line from the file will be skipped, otherwise it will be
     *            interpreted as an Annotation.
     * @return A {@link Model} containing the RDF statements for annotations represented in the TSV
     *         file.
     * @throws IOException
     *             If the file was not found or was not about to be read from.
     */
    public static Model parseTsv(final InputStream inputStream, final boolean skipFirstLine) throws IOException
    {
        return OasTsvUtility.parseTsv(new InputStreamReader(inputStream, StandardCharsets.UTF_8), skipFirstLine);
    }
    
    /**
     * Reads annotations from a TSV file, skipping the first line if necessary, and returns the
     * Annotations as RDF statements in a {@link Model}.
     * 
     * @param inputReader
     *            The TSV file to read annotations from, as an {@link Reader}.
     * @param skipFirstLine
     *            If true the first line from the file will be skipped, otherwise it will be
     *            interpreted as an Annotation.
     * @return A {@link Model} containing the RDF statements for annotations represented in the TSV
     *         file.
     * @throws IOException
     *             If the file was not found or was not about to be read from.
     */
    public static Model parseTsv(final Reader inputReader, final boolean skipFirstLine) throws IOException
    {
        final ValueFactory vf = ValueFactoryImpl.getInstance();
        final Model result = new LinkedHashModel();
        
        final int skipLines = skipFirstLine ? 1 : 0;
        
        int currentLine = skipLines;
        
        // Skips first line which contains headers, and autocloses the reader when finished or an
        // exception occurs
        try (final CSVReader reader = new CSVReader(inputReader, '\t', '"', skipLines))
        {
            String[] nextLine = reader.readNext();
            while(nextLine != null)
            {
                currentLine++;
                if(nextLine.length != OasTsvUtility.EXPECTED_FIELDS.size())
                {
                    throw new IOException("Did not find the expected number of fields on line " + currentLine);
                }
                
                final BNode nextLineIdentifier = vf.createBNode("a" + currentLine);
                
                result.add(nextLineIdentifier, RDF.TYPE, OasConstants.TAGGING_TAGGING);
                
                // Annotated object URI
                if(nextLine[0] == null)
                {
                    throw new IOException("Annotated object was not found on line " + currentLine);
                }
                else
                {
                    result.add(nextLineIdentifier, OasConstants.TAGGING_TAGGED_RESOURCE, vf.createURI(nextLine[0]));
                }
                
                // Ontology Term URI
                if(nextLine[1] == null)
                {
                    throw new IOException("Ontology term was not found on line " + currentLine);
                }
                else
                {
                    result.add(nextLineIdentifier, OasConstants.MOAT_TAG_MEANING, vf.createURI(nextLine[1]));
                }
                
                // Annotation content
                if(nextLine[2] == null)
                {
                    throw new IOException("Annotation content was not found on line " + currentLine);
                }
                else
                {
                    result.add(nextLineIdentifier, OasConstants.TAGGING_NAME,
                            vf.createLiteral(nextLine[2], XMLSchema.STRING));
                }
                
                // Annotator URI
                // Optional field
                if(nextLine[3] != null && !nextLine[3].trim().isEmpty())
                {
                    result.add(nextLineIdentifier, OasConstants.FOAF_MAKER, vf.createURI(nextLine[3]));
                }
                
                // Ontology URI
                // Optional field
                if(nextLine[4] != null && !nextLine[4].trim().isEmpty())
                {
                    result.add(vf.createURI(nextLine[1]), RDFS.ISDEFINEDBY, vf.createURI(nextLine[4]));
                }
                
                // Annotated Object Type URI
                // Optional field
                if(nextLine[5] != null && !nextLine[5].trim().isEmpty())
                {
                    result.add(vf.createURI(nextLine[0]), RDF.TYPE, vf.createURI(nextLine[5]));
                }
                
                nextLine = reader.readNext();
            }
        }
        catch(final IllegalArgumentException iae)
        {
            throw new IOException("Unable to create URI for a field found on line " + currentLine, iae);
        }
        
        return result;
    }
    
    /**
     * Reads annotations from a TSV file, skipping the first line, and returns the Annotations as
     * RDF statements in a {@link Model}.
     * 
     * @param filename
     *            The TSV file to read annotations from.
     * @return A {@link Model} containing the RDF statements for annotations represented in the TSV
     *         file.
     * @throws IOException
     *             If the file was not found or was not about to be read from.
     */
    public static Model parseTsv(final String filename) throws IOException
    {
        return OasTsvUtility.parseTsv(filename, true);
    }
    
    /**
     * Reads annotations from a TSV file, skipping the first line if necessary, and returns the
     * Annotations as RDF statements in a {@link Model}.
     * 
     * @param filename
     *            The TSV file to read annotations from.
     * @param skipFirstLine
     *            If true the first line from the file will be skipped, otherwise it will be
     *            interpreted as an Annotation.
     * @return A {@link Model} containing the RDF statements for annotations represented in the TSV
     *         file.
     * @throws IOException
     *             If the file was not found or was not about to be read from.
     */
    public static Model parseTsv(final String filename, final boolean skipFirstLine) throws IOException
    {
        return OasTsvUtility.parseTsv(new FileReader(filename), skipFirstLine);
    }
    
    /**
     * Private default constructor as this is a static class.
     */
    private OasTsvUtility()
    {
    }
    
}
