/**
 * 
 */
package com.github.ansell.oas.utils;

import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Set;

import org.openrdf.model.URI;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.http.HTTPRepository;
import org.openrdf.repository.sail.SailRepository;
import org.openrdf.sail.memory.MemoryStore;
import org.restlet.Context;
import org.restlet.ext.crypto.DigestAuthenticator;
import org.restlet.ext.crypto.DigestVerifier;
import org.restlet.ext.freemarker.ContextTemplateLoader;
import org.restlet.security.ChallengeAuthenticator;
import org.restlet.security.LocalVerifier;
import org.restlet.security.Realm;
import org.restlet.security.Role;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.data.impl.SesameDataHandlerImpl;
import com.github.ansell.oas.webservice.OasWebServiceApplication;
import com.github.ansell.propertyutil.PropertyUtil;
import com.github.ansell.restletutils.FixedRedirectCookieAuthenticator;
import com.github.ansell.restletutils.RestletUtilRoles;
import com.github.ansell.restletutils.RestletUtilSesameRealm;
import com.github.ansell.restletutils.RestletUtilUser;

import freemarker.ext.beans.BeansWrapper;
import freemarker.template.Configuration;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class ApplicationUtils
{
    private static final Logger log = LoggerFactory.getLogger(ApplicationUtils.class);
    
    public static ChallengeAuthenticator getNewAuthenticator(final Realm nextRealm, final Context context,
            final PropertyUtil propertyUtil)
    {
        ChallengeAuthenticator result = null;
        
        final String authMethod =
                propertyUtil.get(OasProps.PROP_CHALLENGE_AUTH_METHOD, OasProps.DEF_CHALLENGE_AUTH_METHOD);
        
        if(authMethod.equalsIgnoreCase("digest"))
        {
            ApplicationUtils.log.info("Using digest authenticator");
            result =
                    new DigestAuthenticator(context, nextRealm.getName(), propertyUtil.get(
                            OasProps.PROP_CHALLENGE_AUTH_SECRET, OasProps.DEF_CHALLENGE_AUTH_SECRET));
            
            if(nextRealm.getVerifier() instanceof DigestVerifier)
            {
                // NOTE: The verifier in this case must support digest verification by being an
                // instance of DigestVerifier
                result.setVerifier(nextRealm.getVerifier());
            }
            else if(nextRealm.getVerifier() instanceof LocalVerifier)
            {
                // else we need to map the verifier in
                ((DigestAuthenticator)result).setWrappedVerifier((LocalVerifier)nextRealm.getVerifier());
            }
            else
            {
                throw new RuntimeException("Verifier was not valid for use with DigestAuthenticator verifier="
                        + nextRealm.getVerifier().toString());
            }
            
            result.setEnroler(nextRealm.getEnroler());
            
            result.setOptional(Boolean.valueOf(propertyUtil.get(OasProps.PROP_CHALLENGE_AUTH_OPTIONAL,
                    OasProps.DEF_CHALLENGE_AUTH_OPTIONAL)));
        }
        else if(authMethod.equalsIgnoreCase("cookie"))
        {
            ApplicationUtils.log.info("Using cookie authenticator");
            
            final byte[] secretKey =
                    propertyUtil.get(OasProps.PROP_CHALLENGE_AUTH_SECRET, OasProps.DEF_CHALLENGE_AUTH_SECRET).getBytes(
                            StandardCharsets.UTF_8);
            
            result = new FixedRedirectCookieAuthenticator(context, nextRealm.getName(), secretKey);
            
            ((FixedRedirectCookieAuthenticator)result).setLoginFormPath(propertyUtil.get(OasProps.PROP_LOGIN_FORM_PATH,
                    OasProps.DEF_LOGIN_FORM_PATH));
            ((FixedRedirectCookieAuthenticator)result).setLoginPath(propertyUtil.get(OasProps.PROP_LOGIN_PATH,
                    OasProps.DEF_LOGIN_PATH));
            ((FixedRedirectCookieAuthenticator)result).setLogoutPath(propertyUtil.get(OasProps.PROP_LOGOUT_PATH,
                    OasProps.DEF_LOGOUT_PATH));
            ((FixedRedirectCookieAuthenticator)result).setRedirectQueryName(propertyUtil.get(
                    OasProps.PROP_LOGIN_REDIRECT_FIELD, OasProps.DEF_LOGIN_REDIRECT_FIELD));
            // FIXME: Make this configurable
            ((FixedRedirectCookieAuthenticator)result).setCookieName("OASAuthentication");
            // FIXME: Make this configurable
            ((FixedRedirectCookieAuthenticator)result).setIdentifierFormName("login");
            // FIXME: Make this configurable
            ((FixedRedirectCookieAuthenticator)result).setSecretFormName("password");
            ((FixedRedirectCookieAuthenticator)result).setInterceptingLogin(true);
            ((FixedRedirectCookieAuthenticator)result).setInterceptingLogout(true);
            ((FixedRedirectCookieAuthenticator)result).setFixedRedirectUri(propertyUtil.get(
                    OasProps.PROP_ONTOLOGY_MANAGER_PATH, OasProps.DEF_ONTOLOGY_MANAGER_PATH));
            
            result.setMultiAuthenticating(false);
            
            result.setVerifier(nextRealm.getVerifier());
            result.setEnroler(nextRealm.getEnroler());
            
            result.setOptional(Boolean.valueOf(propertyUtil.get(OasProps.PROP_CHALLENGE_AUTH_OPTIONAL,
                    OasProps.DEF_CHALLENGE_AUTH_OPTIONAL)));
            
        }
        else if(authMethod.equalsIgnoreCase("http"))
        {
            // FIXME: Implement a stub here
            ApplicationUtils.log.error("FIXME: Implement HTTP ChallengeAuthenticator authMethod={}", authMethod);
            throw new RuntimeException("FIXME: Implement HTTP ChallengeAuthenticator");
        }
        else
        {
            ApplicationUtils.log.error("Did not recognise ChallengeAuthenticator method authMethod={}", authMethod);
            throw new RuntimeException("Did not recognise ChallengeAuthenticator method");
        }
        
        return result;
    }
    
    public static Repository getNewRepository(final PropertyUtil propertyUtil)
    {
        final String repositoryUrl = propertyUtil.get(OasProps.PROP_SESAME_URL, "");
        
        ApplicationUtils.log.debug("Configured using Sesame repository URL: {}", repositoryUrl);
        
        // if we weren't able to find a repository URL in the configuration, we setup an
        // in memory store
        if(repositoryUrl == null || repositoryUrl.trim().isEmpty())
        {
            final Repository repository = new SailRepository(new MemoryStore());
            
            try
            {
                repository.initialize();
                
                ApplicationUtils.log.info("Created an in memory store as repository for OAS data");
                
                return repository;
            }
            catch(final RepositoryException ex)
            {
                throw new RuntimeException("Could not initialise Sesame In Memory repository");
            }
        }
        else
        {
            final Repository repository = new HTTPRepository(repositoryUrl);
            
            try
            {
                repository.initialize();
                
                ApplicationUtils.log.info("Using sesame http repository as repository for OAS data URL={}",
                        repositoryUrl);
                
                return repository;
            }
            catch(final RepositoryException ex)
            {
                throw new RuntimeException("Could not initialise Sesame HTTP repository with URL=" + repositoryUrl);
            }
        }
    }
    
    public static Configuration getNewTemplateConfiguration(final Context newChildContext)
    {
        final Configuration result = new Configuration();
        result.setDefaultEncoding("UTF-8");
        result.setURLEscapingCharset("UTF-8");
        
        // FIXME: Make the template path configurable
        result.setTemplateLoader(new ContextTemplateLoader(newChildContext, "clap://class/templates"));
        
        final BeansWrapper myWrapper = new BeansWrapper();
        myWrapper.setSimpleMapWrapper(true);
        result.setObjectWrapper(myWrapper);
        
        return result;
    }
    
    public static void setupApplication(final OasWebServiceApplication application, final Context applicationContext)
    {
        ApplicationUtils.log.debug("application {}", application);
        ApplicationUtils.log.debug("applicationContext {}", applicationContext);
        
        final PropertyUtil propertyUtil = application.getPropertyUtil();
        
        final List<Role> roles = ((org.restlet.Application)application).getRoles();
        roles.clear();
        roles.addAll(RestletUtilRoles.getRoles());
        
        final Repository nextRepository = ApplicationUtils.getNewRepository(application.getPropertyUtil());
        
        final RestletUtilSesameRealm nextRealm =
                new RestletUtilSesameRealm(nextRepository, SesameUtils.USER_MANAGER_GRAPH);
        
        // FIXME: Make this configurable
        nextRealm.setName("OASRealm");
        
        final List<RestletUtilUser> currentUsers = nextRealm.getUsers();
        
        // If the realm does not have any users then we insert some defaults
        if(currentUsers.isEmpty())
        {
            ApplicationUtils.log.warn("Automatically adding default users as there were no users found in the system");
            
            final RestletUtilUser defaultUser =
                    new RestletUtilUser("anotherUser", "anotherPassword", "Default Non Admin", "User",
                            "default.user@example.com");
            final URI defaultUserUri = nextRealm.addUser(defaultUser);
            
            final RestletUtilUser defaultAdminUser =
                    new RestletUtilUser(propertyUtil.get(OasProps.PROP_ADMIN_USER_IDENTIFIER,
                            OasProps.DEF_ADMIN_USER_IDENTIFIER), propertyUtil.get(OasProps.PROP_ADMIN_USER_SECRET,
                            OasProps.DEF_ADMIN_USER_SECRET), "Default Admin", "User", "default.admin.user@example.com");
            final URI defaultAdminUserUri = nextRealm.addUser(defaultAdminUser);
            nextRealm.map(defaultAdminUser, RestletUtilRoles.ADMIN.getRole());
        }
        else
        {
            boolean adminFound = false;
            for(final RestletUtilUser nextCurrentUser : currentUsers)
            {
                final Set<Role> testAdminUserRoles = nextRealm.findRoles(nextCurrentUser);
                
                if(testAdminUserRoles.contains(RestletUtilRoles.ADMIN.getRole()))
                {
                    adminFound = true;
                    break;
                }
            }
            
            if(!adminFound)
            {
                final RestletUtilUser defaultAdminUser =
                        new RestletUtilUser(propertyUtil.get(OasProps.PROP_ADMIN_USER_IDENTIFIER,
                                OasProps.DEF_ADMIN_USER_IDENTIFIER), propertyUtil.get(OasProps.PROP_ADMIN_USER_SECRET,
                                OasProps.DEF_ADMIN_USER_SECRET), "Default Admin", "User",
                                "default.admin.user@example.com");
                final URI defaultAdminUserUri = nextRealm.addUser(defaultAdminUser);
                nextRealm.map(defaultAdminUser, RestletUtilRoles.ADMIN.getRole());
                
                ApplicationUtils.log
                        .warn("Automatically added an administrator user as there were no admin users found in the system");
            }
        }
        
        final ChallengeAuthenticator newAuthenticator =
                ApplicationUtils.getNewAuthenticator(nextRealm, applicationContext, application.getPropertyUtil());
        application.setAuthenticator(newAuthenticator);
        
        application.setRealm(nextRealm);
        
        application.setDataHandler(new SesameDataHandlerImpl(nextRepository));
        
        final Configuration newTemplateConfiguration = ApplicationUtils.getNewTemplateConfiguration(applicationContext);
        application.setTemplateConfiguration(newTemplateConfiguration);
        
        // create a custom error handler using our overridden OasStatusService together with the
        // freemarker configuration so that we can display error messages nicely inside of the
        // normal interface
        final OasStatusService statusService =
                new OasStatusService(newTemplateConfiguration, application.getPropertyUtil());
        application.setStatusService(statusService);
    }
    
    private ApplicationUtils()
    {
    }
    
}
