/**
 * Utility package for OAS
 */
package com.github.ansell.oas.utils;

import java.security.SecureRandom;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Random;
import java.util.TimeZone;

import org.openrdf.model.Literal;
import org.openrdf.model.URI;
import org.openrdf.model.ValueFactory;
import org.openrdf.model.impl.ValueFactoryImpl;

/**
 * Useful date manipulation utils.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public final class DateUtils
{
    private static final String DATE_FORMAT_TEMPLATE = "yyyy-MM-dd'T'HH:mm:ss'Z'";
    private static final String UTC = "UTC";
    private static final ValueFactory VALUE_FACTORY_INSTANCE = ValueFactoryImpl.getInstance();
    private static final Random rng = new SecureRandom("abcdefghijklmnopqrstuvwxyz1234567890".getBytes());
    
    public static String getCurrentDate()
    {
        // SimpleDateFormat instances are not thread-safe, so we recreate them each time
        final SimpleDateFormat dateFormat = new SimpleDateFormat(DateUtils.DATE_FORMAT_TEMPLATE);
        dateFormat.setTimeZone(TimeZone.getTimeZone(DateUtils.UTC));
        
        return dateFormat.format(new Date());
    }
    
    public static Literal getCurrentDateLiteral()
    {
        // Create a new Date instance to get the current date and time
        // before formatting and converting it to a Sesame Datatyped Literal
        final Literal modifiedDate =
                DateUtils.VALUE_FACTORY_INSTANCE.createLiteral(DateUtils.getCurrentDate(), OasConstants.XSD_DATETIME);
        
        return modifiedDate;
    }
    
    /**
     * Returns a new Unique URI based on the date and the hashCode from a seed, for example, a
     * random number or a blank node identifier.
     * 
     * @param baseUri
     *            The base URI to use in the process of creating a new unique URI.
     * @param seed
     *            A seed to use to generate a unique identifier along with the current date.
     * @return A new Unique URI that can be used to overwrite blank nodes that users have sent us.
     */
    public static URI getNewUniqueUriByDate(final String baseUri, final String seed)
    {
        return DateUtils.VALUE_FACTORY_INSTANCE.createURI(baseUri, DateUtils.getCurrentDate() + "_" + seed.hashCode()
                + "_" + DateUtils.rng.nextLong());
    }
    
    /**
     * Private default constructor.
     */
    private DateUtils()
    {
        // should not need to instantiate this class
    }
    
}
