/**
 * 
 */
package com.github.ansell.oas.utils;

import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.openrdf.model.URI;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFHandlerException;
import org.restlet.Request;
import org.restlet.Response;
import org.restlet.data.MediaType;
import org.restlet.data.Reference;
import org.restlet.data.Status;
import org.restlet.ext.fileupload.RestletFileUpload;
import org.restlet.representation.Representation;
import org.restlet.resource.ResourceException;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyID;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.model.OWLOntologyRenameException;
import org.semanticweb.owlapi.model.SetOntologyID;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.data.DataHandler;
import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.objects.impl.OntologyImpl;
import com.github.ansell.propertyutil.PropertyUtil;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class OntologyUtils
{
    private static final Logger log = LoggerFactory.getLogger(OntologyUtils.class);
    
    public static void deleteOntologyInternal(final String ontologyUriString,
            final String managedOntologyVersionUriString, final String deleteOldVersions,
            final DataHandler nextDataHandler) throws ResourceException
    {
        final URI ontologyUri = nextDataHandler.createURI(ontologyUriString);
        
        // If the managedOntologyVersionUri parameter was present, then only attempt to delete that
        // exact version
        if(managedOntologyVersionUriString != null)
        {
            final URI managedOntologyVersionUri = nextDataHandler.createURI(managedOntologyVersionUriString);
            
            final Ontology managedOntology =
                    nextDataHandler.getVersionedOntology(ontologyUri, managedOntologyVersionUri);
            
            if(managedOntology == null)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                        "Could not find the ontology with the given ontology URI and managed ontology version URI.");
            }
            
            if(!nextDataHandler.deleteOntologyVersion(managedOntology))
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                        "Could not delete the ontology with the given ontology URI and managed ontology version URI.");
            }
        }
        // else try to delete all versions of the given ontology if they specified that
        else if(Boolean.valueOf(deleteOldVersions))
        {
            final List<Ontology> ontologiesByUri = nextDataHandler.getOntologiesByUri(ontologyUri);
            
            if(ontologiesByUri.size() == 0)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                        "Could not find any ontologies with the given ontology URI.");
            }
            
            boolean failureOccurred = false;
            
            OntologyUtils.log.info("About to delete {} versions of the ontology with URI={}", ontologiesByUri.size(),
                    ontologyUri);
            
            for(final Ontology nextOntology : ontologiesByUri)
            {
                if(!nextDataHandler.deleteOntologyVersion(nextOntology))
                {
                    failureOccurred = true;
                }
            }
            
            if(failureOccurred)
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                        "Failed to delete all versions of the ontology.");
            }
        }
        else
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                    "Could not delete: No version given, and deleteAllVersions parameter was not set to true");
        }
    }
    
    public static void deleteOntologyInternalMultipartFormPost(final Request nextRequest, final Response nextResponse,
            final DataHandler nextDataHandler, final Path tempDirectory, final PropertyUtil propertyUtil)
        throws ResourceException
    {
        OntologyUtils.log.debug("deleteOntology: deleteOntologyInternalMultipartFormPost");
        
        if(nextRequest.getEntity() == null)
        {
            OntologyUtils.log.error("nextRequest.getEntity() was null");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                    "Did not submit a valid multipart/form-data query");
        }
        
        OntologyUtils.log.debug("nextRequest.getEntity() was NOT null");
        
        if(MediaType.MULTIPART_FORM_DATA.equals(nextRequest.getEntity().getMediaType(), true))
        {
            OntologyUtils.log.debug("Found multipart/form-data request entity: {}", nextRequest.getEntity());
            
            List<FileItem> items;
            
            // 1: Create a factory for disk-based file items
            final DiskFileItemFactory factory = new DiskFileItemFactory(1000240, tempDirectory.toFile());
            
            // 2: Create a new file upload handler
            final RestletFileUpload upload = new RestletFileUpload(factory);
            final Map<String, String> props = new HashMap<String, String>();
            try
            {
                // 3: Request is parsed by the handler which generates a list of FileItems
                items = upload.parseRequest(nextRequest);
                
                for(final FileItem fi : items)
                {
                    final String name = fi.getName();
                    
                    if(name == null)
                    {
                        props.put(fi.getFieldName(), new String(fi.get(), StandardCharsets.UTF_8));
                    }
                    else
                    {
                        OntologyUtils.log.error("Ignoring file upload field in multipart query fi={}", fi);
                    }
                }
                
                OntologyUtils.log.debug("props={}", props.toString());
                
            }
            catch(final FileUploadException e)
            {
                OntologyUtils.log.debug("props={}", props.toString());
                
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, e);
            }
            
            String ontologyUriString;
            
            if(props.containsKey("ontologyUri"))
            {
                ontologyUriString = props.get("ontologyUri");
            }
            else
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Did not submit an ontologyUri");
            }
            
            String managedOntologyVersionUriString = null;
            
            if(props.containsKey("ontologyVersionUri"))
            {
                managedOntologyVersionUriString = props.get("ontologyVersionUri");
            }
            
            String deleteOldVersions = "false";
            
            if(props.containsKey("deleteOldVersions"))
            {
                deleteOldVersions = props.get("deleteOldVersions");
                
                OntologyUtils.log.debug("deleteOldVersions={}", deleteOldVersions);
            }
            
            OntologyUtils.deleteOntologyInternal(ontologyUriString, managedOntologyVersionUriString, deleteOldVersions,
                    nextDataHandler);
            
            nextResponse.redirectSeeOther(nextRequest.getRootRef().toString(false, false)
                    + propertyUtil.get(OasProps.PROP_ONTOLOGY_MANAGER_PATH, OasProps.DEF_ONTOLOGY_MANAGER_PATH));
        }
        else
        {
            OntologyUtils.log.debug("Passing query to deleteOntologyInternalQueryForm");
            OntologyUtils.deleteOntologyInternalQueryForm(nextRequest, nextResponse, nextDataHandler);
        }
    }
    
    /**
     * @param nextRequest
     *            TODO
     * @param nextDataHandler
     *            TODO
     * @throws ResourceException
     */
    public static void deleteOntologyInternalQueryForm(final Request nextRequest, final Response nextResponse,
            final DataHandler nextDataHandler) throws ResourceException
    {
        OntologyUtils.log.debug("deleteOntology: deleteOntologyInternalQueryForm");
        
        final String ontologyUriString =
                nextRequest.getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        OntologyUtils.log.debug("ontologyUriString={}", ontologyUriString);
        
        if(ontologyUriString == null)
        {
            OntologyUtils.log.error("ontologyUri parameter was not present");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "ontologyUri parameter must not be empty");
        }
        
        final String managedOntologyVersionUriString =
                nextRequest.getResourceRef().getQueryAsForm().getFirstValue("ontologyVersionUri", true);
        
        final String deleteAllVersionsString =
                nextRequest.getResourceRef().getQueryAsForm().getFirstValue("deleteAllVersions", true);
        
        OntologyUtils.deleteOntologyInternal(ontologyUriString, managedOntologyVersionUriString,
                deleteAllVersionsString, nextDataHandler);
        
        nextResponse.setStatus(Status.SUCCESS_NO_CONTENT);
    }
    
    /**
     * This method stored the ontology contained in the given entity based on form values set in the
     * request.
     * 
     * The response must contain a query parameter named "ontologyUri" or the request will fail. The
     * request may also optionally contain parameters named "ontologyVersionUri", "baseUri", and a
     * boolean value for "deleteOldVersions".
     * 
     * The
     * 
     * @param nextRequest
     *            TODO
     * @param nextResponse
     *            TODO
     * @param nextDataHandler
     *            TODO
     * @param entityInputStream
     *            TODO
     * @param mediaType
     *            TODO
     * @return
     * @throws ResourceException
     */
    public static Representation putOntologyRdfInternal(final Request nextRequest, final Response nextResponse,
            final DataHandler nextDataHandler, final InputStream entityInputStream, final MediaType mediaType)
        throws ResourceException
    {
        if(OntologyUtils.log.isDebugEnabled())
        {
            OntologyUtils.log.debug("received PUT request");
        }
        
        final String ontologyUri = nextRequest.getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        final String ontologyVersionUri =
                nextRequest.getResourceRef().getQueryAsForm().getFirstValue("ontologyVersionUri", true);
        
        final String baseUri = nextRequest.getResourceRef().getQueryAsForm().getFirstValue("baseUri", true);
        
        final boolean deleteOldVersions =
                Boolean.parseBoolean(nextRequest.getResourceRef().getQueryAsForm()
                        .getFirstValue("deleteOldVersions", /* ignoreCase */true));
        
        OntologyUtils.log.debug("deleteOldVersions={}", deleteOldVersions);
        
        return OntologyUtils.putOntologyRdfInternal(ontologyUri, ontologyVersionUri, baseUri, deleteOldVersions,
                nextResponse, nextDataHandler, entityInputStream, mediaType, nextRequest.getResourceRef());
    }
    
    public static Representation putOntologyRdfInternal(final String ontologyUri, final String ontologyVersionUri,
            String baseUri, final boolean deleteOldVersions, final Response nextResponse,
            final DataHandler nextDataHandler, final InputStream entityInputStream, final MediaType mediaType,
            final Reference resourceReference) throws ResourceException
    {
        try
        {
            final Ontology inputOntology = new OntologyImpl();
            
            if(ontologyUri == null)
            {
                OntologyUtils.log.error("ontologyUri parameter was not present");
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "ontologyUri parameter must not be empty");
            }
            
            inputOntology.setOntologyUri(nextDataHandler.createURI(ontologyUri));
            
            if(ontologyVersionUri != null)
            {
                inputOntology.setOntologyVersionUri(nextDataHandler.createURI(ontologyVersionUri));
            }
            
            if(baseUri == null)
            {
                baseUri = ontologyUri;
            }
            
            if(OntologyUtils.log.isDebugEnabled())
            {
                OntologyUtils.log.debug("about to store ontology=" + inputOntology.toString());
            }
            
            // TODO: Re-enable the following when Restlet supports this combination
            // See bug at github https://github.com/restlet/restlet-framework-java/issues/100
            // nextResponse.setStatus(Status.INFO_CONTINUE);
            // nextResponse.commit();
            
            final Ontology result =
                    nextDataHandler.storeOntology(inputOntology, entityInputStream, baseUri, mediaType.getName(),
                            deleteOldVersions);
            nextResponse.setStatus(Status.SUCCESS_CREATED);
            
            return RestletUtils.getRdfRepresentation(result, mediaType, resourceReference);
        }
        catch(final RepositoryException | RDFHandlerException e)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, e);
        }
    }
    
    public static OWLOntologyID replaceOWLOntologyID(final Ontology nextOntology,
            final OWLOntologyManager owlapiOntologyManager, final OWLOntology rootOntology)
        throws OWLOntologyRenameException
    {
        final OWLOntologyID owlOntologyID = nextOntology.toOWLOntologyID();
        
        final SetOntologyID changeRequest = new SetOntologyID(rootOntology, owlOntologyID);
        
        owlapiOntologyManager.applyChange(changeRequest);
        
        if(!owlOntologyID.equals(rootOntology.getOntologyID()))
        {
            throw new IllegalStateException(String.format("Unable to replace the ontologyID actual="
                    + rootOntology.getOntologyID().toString() + " expected=" + owlOntologyID.toString()));
        }
        
        return rootOntology.getOntologyID();
    }
    
    private OntologyUtils()
    {
    }
    
}
