/**
 * Utility package for OAS
 */
package com.github.ansell.oas.utils;

import info.aduna.iteration.Iterations;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.openrdf.OpenRDFUtil;
import org.openrdf.model.BNode;
import org.openrdf.model.Literal;
import org.openrdf.model.Model;
import org.openrdf.model.Resource;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.ValueFactory;
import org.openrdf.model.impl.LinkedHashModel;
import org.openrdf.model.vocabulary.OWL;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.RDFS;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryConnection;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.RepositoryResult;
import org.openrdf.repository.sail.SailRepository;
import org.openrdf.rio.RDFFormat;
import org.openrdf.rio.RDFHandlerException;
import org.openrdf.rio.RDFParseException;
import org.openrdf.rio.Rio;
import org.openrdf.rio.UnsupportedRDFormatException;
import org.openrdf.sail.memory.MemoryStore;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.resource.ResourceException;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.profiles.OWLProfile;
import org.semanticweb.owlapi.profiles.OWLProfileRegistry;
import org.semanticweb.owlapi.vocab.Namespaces;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.data.DataHandler;
import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.objects.ResourceCount;
import com.github.ansell.oas.objects.SearchResult;
import com.github.ansell.oas.objects.impl.AnnotationImpl;
import com.github.ansell.oas.objects.impl.OntologyImpl;
import com.github.ansell.oas.objects.impl.UriCountImpl;
import com.github.ansell.propertyutil.PropertyUtil;

/**
 * Wraps OpenRDF Sesame RDF actions related to serialising and deserialising objects.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public final class SesameUtils
{
    private static final Logger LOG = LoggerFactory.getLogger(SesameUtils.class);
    
    /**
     * A URI populated by the property "oas.ontologymanager.managergraph", defaulting to
     * <urn:tag:oas:ontologymanager:managergraph>.
     */
    public static final URI ONTOLOGY_MANAGER_GRAPH = OasConstants.toURI(PropertyUtil.getInstance().get(
            OasProps.PROP_ONTOLOGY_MANAGER_GRAPH, OasProps.DEF_ONTOLOGY_MANAGER_GRAPH));
    
    /**
     * A URI populated by the property "oas.annotations.managementgraph", defaulting to
     * <urn:tag:oas:annotations:managementgraph>.
     */
    public static final URI ANNOTATION_MANAGEMENT_GRAPH = OasConstants.toURI(PropertyUtil.getInstance().get(
            OasProps.PROP_ANNOTATION_MANAGMENT_GRAPH, OasProps.DEF_ANNOTATIONS_MANAGMENT_GRAPH));
    
    /**
     * A URI populated by the property "oas.usermanager.managergraph", defaulting to
     * <urn:tag:oas:usermanager:managergraph>.
     */
    public static final URI USER_MANAGER_GRAPH = OasConstants.toURI(PropertyUtil.getInstance().get(
            OasProps.PROP_USER_MANAGER_GRAPH, OasProps.DEF_USER_MANAGER_GRAPH));
    
    /**
     * A URI populated by the property "oas.auditlog.graph", defaulting to
     * &lt;urn:tag:oas:auditlog:graph&gt;.
     */
    public static final URI AUDIT_LOG_GRAPH = OasConstants.toURI(PropertyUtil.getInstance().get(
            OasProps.PROP_AUDIT_LOG_GRAPH, OasProps.DEF_AUDIT_LOG_GRAPH));
    
    public static Repository setLatestVersion(final URI ontologyUri, final URI latestOntologyVersionUri,
            final Repository myRepository, final URI... contexts)
    {
        if(ontologyUri == null || latestOntologyVersionUri == null)
        {
            throw new IllegalArgumentException("Ontology URI and latest Version URI must both not be null");
        }
        
        // copy the reference so we can create a new repository in its place if necessary
        Repository actualRepository = myRepository;
        
        RepositoryConnection conn = null;
        
        try
        {
            // create a repository for them if they did not supply one
            if(actualRepository == null)
            {
                actualRepository = new SailRepository(new MemoryStore());
                actualRepository.initialize();
            }
            
            conn = actualRepository.getConnection();
            
            conn.begin();
            conn.add(ontologyUri, RDF.TYPE, OWL.ONTOLOGY);
            
            conn.remove(ontologyUri, OasConstants.OMV_CURRENTVERSION, null, contexts);
            conn.add(ontologyUri, OasConstants.OMV_CURRENTVERSION, latestOntologyVersionUri, contexts);
            
            conn.commit();
        }
        catch(final RepositoryException ex)
        {
            try
            {
                if(conn != null && conn.isActive())
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException e)
            {
                SesameUtils.LOG.error("Found exception while rolling back commit", e);
            }
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                SesameUtils.LOG.error("Found exception while closing connection", ex);
            }
        }
        
        return actualRepository;
    }
    
    /**
     * Some OWL and RDF tools incorrectly rely on namespace prefixes being universal, even though
     * they have no literal meaning in RDF theory.
     * 
     * To reduce the possibility that a connection will have these standard prefixes overridden,
     * they are setup here based on the OWLAPI Namespaces enum.
     * 
     * @param repositoryConnection
     * @throws RepositoryException
     */
    public static void addStandardNamespaces(final RepositoryConnection repositoryConnection)
        throws RepositoryException
    {
        for(final Namespaces nextNamespace : Namespaces.values())
        {
            if(nextNamespace.isInUse())
            {
                final String currentNamespace = repositoryConnection.getNamespace(nextNamespace.getCommonPrefix());
                if(currentNamespace == null)
                {
                    repositoryConnection.setNamespace(nextNamespace.getCommonPrefix(), nextNamespace.getBaseIRI());
                }
            }
        }
    }
    
    /**
     * Fetches an annotation from the repository using the given URI.
     * 
     * @param annotationIdUri
     *            The URI of the annotation to search for, and the context to search for the
     *            annotation in.
     * @param myRepository
     *            The repository to search for the annotation in.
     * @param annotationManagementGraph
     *            The URI to use to confirm the validity of the annotation ID. If this is null the
     *            check is ignored and the annotation is returned if it exists in the context
     *            identified by its URI.
     * @return An annotation object containing the item.
     * @throws RepositoryException
     *             Any connection or other repository exceptions are thrown without being handled
     * @throws IllegalArgumentException
     *             if the URI was not found in the repository.
     */
    public static Annotation createAnnotationByUriFromRepository(final URI annotationIdUri,
            final Repository myRepository, final URI annotationManagementGraph) throws RepositoryException
    {
        Annotation result = null;
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = myRepository.getConnection();
            
            if(annotationManagementGraph != null)
            {
                // the RDF:TYPE triple is required for any valid annotation in the repository, so if
                // it
                // is not present fail early in addition, it needs to be unique for a given
                // annotation
                // URI
                final List<Statement> checkQuery =
                        Iterations.asList(conn.getStatements(annotationIdUri, RDF.TYPE, OasConstants.TAGGING_TAGGING,
                                true, annotationManagementGraph));
                if(checkQuery.size() != 1)
                {
                    SesameUtils.LOG.error("Could not find a unique managed anontation with the given URI: {}",
                            annotationIdUri);
                    return null;
                }
            }
            
            final Model statements = new LinkedHashModel();
            
            Iterations.addAll(conn.getStatements(annotationIdUri, null, null, true, annotationIdUri), statements);
            
            final List<Statement> taggedResources =
                    Iterations.asList(conn.getStatements(annotationIdUri, OasConstants.TAGGING_TAGGED_RESOURCE, null,
                            true, annotationIdUri));
            
            for(final Statement nextTaggedResource : taggedResources)
            {
                Iterations.addAll(
                        conn.getStatements((URI)nextTaggedResource.getObject(), null, null, true, annotationIdUri),
                        statements);
            }
            
            final List<Statement> termResources =
                    Iterations.asList(conn.getStatements(annotationIdUri, OasConstants.MOAT_TAG_MEANING, null, true,
                            annotationIdUri));
            
            for(final Statement nextTermResource : termResources)
            {
                Iterations.addAll(
                        conn.getStatements((URI)nextTermResource.getObject(), null, null, true, annotationIdUri),
                        statements);
            }
            
            if(!statements.isEmpty())
            {
                result = new AnnotationImpl(statements, annotationIdUri, annotationIdUri);
            }
        }
        finally
        {
            if(conn != null)
            {
                conn.close();
            }
        }
        return result;
    }
    
    /**
     * Creates annotations using the repository of RDF values, including creating unique subject
     * URIs as necessary.
     * 
     * @param myRepository
     *            The repository containing the RDF triples that define this object.
     * @param annotationManagementGraph
     *            TODO
     * @throws RepositoryException
     *             Any connection or other repository exceptions are thrown without being handled
     */
    public static Map<URI, Annotation> createAnnotationsFromRepository(final Repository myRepository,
            final String baseUri, final URI annotationManagementGraph) throws RepositoryException
    {
        final Map<URI, Annotation> results = new HashMap<URI, Annotation>();
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = myRepository.getConnection();
            
            // For each annotation, create an Annotation object and return the collection
            
            final RepositoryResult<Statement> typeStatements =
                    conn.getStatements(null, RDF.TYPE, OasConstants.TAGGING_TAGGING, true, annotationManagementGraph);
            
            try
            {
                while(typeStatements.hasNext())
                {
                    final Statement nextAnnotation = typeStatements.next();
                    
                    final Resource nextSubject = nextAnnotation.getSubject();
                    
                    final Model statements = new LinkedHashModel();
                    
                    Iterations.addAll(conn.getStatements(nextSubject, null, null, true), statements);
                    
                    final List<Statement> taggedResources =
                            Iterations.asList(conn.getStatements(nextSubject, OasConstants.TAGGING_TAGGED_RESOURCE,
                                    null, true));
                    
                    for(final Statement nextTaggedResource : taggedResources)
                    {
                        Iterations.addAll(conn.getStatements((URI)nextTaggedResource.getObject(), null, null, true),
                                statements);
                    }
                    
                    final List<Statement> termResources =
                            Iterations.asList(conn
                                    .getStatements(nextSubject, OasConstants.MOAT_TAG_MEANING, null, true));
                    
                    for(final Statement nextTermResource : termResources)
                    {
                        Iterations.addAll(conn.getStatements((URI)nextTermResource.getObject(), null, null, true),
                                statements);
                    }
                    
                    URI nextSubjectUri = null;
                    
                    if(nextSubject instanceof BNode)
                    {
                        // if the current identifier is a blank node, we create a new unique URI
                        nextSubjectUri =
                                DateUtils.getNewUniqueUriByDate(
                                        baseUri
                                                + PropertyUtil.getInstance().get(OasProps.PROP_ANNOTATION_PATH,
                                                        OasProps.DEF_ANNOTATION_PATH), ((BNode)nextSubject).getID());
                    }
                    else
                    {
                        nextSubjectUri = (URI)nextSubject;
                    }
                    
                    results.put(nextSubjectUri, new AnnotationImpl(statements, nextSubject, nextSubjectUri));
                }
            }
            finally
            {
                typeStatements.close();
            }
        }
        finally
        {
            if(conn != null)
            {
                conn.close();
            }
        }
        
        return results;
    }
    
    /**
     * 
     * @param myRepository
     *            The repository to search for ontologies in
     * @param contexts
     *            Optional list of contexts to search for ontology metadata in
     * @return A map containing the URIs of ontologies as the keys and lists of Ontology metadata
     *         objects for each version of the ontology as the values
     */
    public static Map<URI, List<Ontology>> createOntologiesFromRepository(final Repository myRepository,
            final Resource... contexts)
    {
        final Map<URI, List<Ontology>> results = new HashMap<URI, List<Ontology>>();
        
        RepositoryConnection conn = null;
        
        try
        {
            conn = myRepository.getConnection();
            
            RepositoryResult<Statement> ontologyUriList = null;
            
            try
            {
                ontologyUriList = conn.getStatements(null, RDF.TYPE, OWL.ONTOLOGY, false, contexts);
                
                while(ontologyUriList.hasNext())
                {
                    final Statement nextStatement = ontologyUriList.next();
                    
                    RepositoryResult<Statement> versionList = null;
                    try
                    {
                        versionList =
                                conn.getStatements(nextStatement.getSubject(), OasConstants.OWL_VERSIONIRI, null,
                                        false, contexts);
                        
                        final List<Ontology> nextResults = new ArrayList<Ontology>();
                        
                        while(versionList.hasNext())
                        {
                            final Statement nextVersion = versionList.next();
                            
                            if(nextVersion.getObject() instanceof URI)
                            {
                                nextResults.add(SesameUtils.createOntologyFromRepository(conn,
                                        (URI)nextStatement.getSubject(), (URI)nextVersion.getObject(), contexts));
                            }
                        }
                        
                        // TODO: if the version or the inferred ontology URI were not created so
                        // far, try to use OMV:currentVersion and OAS:currentInferredAxiomsOntology
                        // to populate them
                        
                        results.put((URI)nextStatement.getSubject(), nextResults);
                    }
                    finally
                    {
                        if(versionList != null)
                        {
                            versionList.close();
                        }
                    }
                }
            }
            finally
            {
                if(ontologyUriList != null)
                {
                    ontologyUriList.close();
                }
            }
        }
        catch(final RepositoryException re)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                    "Failed to complete fetch, as a repository exception occurred.", re);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException re)
                {
                    throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                            "Failed to complete fetch, as a repository exception occurred.", re);
                }
            }
        }
        
        return results;
    }
    
    public static Ontology createOntologyFromRepository(final RepositoryConnection conn, final URI ontologyUri,
            final URI ontologyVersionUri, final Resource... contexts) throws RepositoryException
    {
        final Ontology nextResult = new OntologyImpl();
        
        // nextResult.setOntologyUri((URI)nextStatement.getSubject());
        // nextResult.setOntologyVersionUri((URI)nextVersion.getObject());
        nextResult.setOntologyUri(ontologyUri);
        nextResult.setOntologyVersionUri(ontologyVersionUri);
        
        nextResult.setOntologyDetails(new LinkedHashModel(Iterations.asSet(conn.getStatements(ontologyVersionUri, null,
                null, true, contexts))));
        
        final RepositoryResult<Statement> inferredOntologyList =
                conn.getStatements(ontologyVersionUri, OasConstants.OAS_INFERREDAXIOMSONTOLOGY, null, false, contexts);
        
        try
        {
            while(inferredOntologyList.hasNext())
            {
                final Statement inferredStatement = inferredOntologyList.next();
                
                if(inferredStatement.getObject() instanceof URI)
                {
                    nextResult.setInferredOntologyUri((URI)inferredStatement.getObject());
                    break;
                }
                else
                {
                    SesameUtils.LOG.warn("Found non-URI for inferredOntology link: {}", inferredStatement);
                }
            }
        }
        finally
        {
            inferredOntologyList.close();
        }
        
        final RepositoryResult<Statement> profileList =
                conn.getStatements(ontologyVersionUri, OasConstants.OAS_INPROFILE, null, false, contexts);
        
        final List<OWLProfile> resultProfiles = new ArrayList<OWLProfile>();
        
        try
        {
            while(profileList.hasNext())
            {
                final Statement nextProfileStatement = profileList.next();
                
                if(nextProfileStatement.getObject() instanceof URI)
                {
                    final IRI nextProfileIRI = IRI.create((URI)nextProfileStatement.getObject());
                    resultProfiles.add(OWLProfileRegistry.getInstance().getProfile(nextProfileIRI));
                }
                else
                {
                    SesameUtils.LOG.warn("Found non-URI for profile link: {}", nextProfileStatement);
                }
            }
        }
        finally
        {
            profileList.close();
        }
        
        nextResult.setProfiles(resultProfiles);
        
        return nextResult;
    }
    
    /**
     * Generates a new Ontology Version URI based on either the last ontology version URI if it
     * exists, or the ontology URI otherwise.
     * 
     * @param ontologyUri
     *            If the last ontology version URI is not known this is required to generate a new
     *            ontology version URI.
     * @param lastOntologyVersionUri
     *            The last ontology version URI that is known. If there are no versions known this
     *            should be null.
     * @param ontologyManagerRepository
     *            The OpenRDF interface to the repository to fetch the ontology management data
     *            from.
     * @param contexts
     *            The contexts to search for current references to the ontology in. Should default
     *            to a single ontology manager context URI. NOTE: Blank node contexts are not
     *            supported.
     * @return A new URI that is different to both of the given inputs.
     * @throws RepositoryException
     *             If the repository threw an exception while trying to access it.
     * @throws IllegalArgumentException
     *             If both the ontology URI and the last ontology version URI are null.
     */
    public static URI generateNewOntologyVersionUri(final URI ontologyUri, final URI lastOntologyVersionUri,
            final Repository ontologyManagerRepository, final URI... contexts) throws RepositoryException
    {
        if(ontologyUri == null && lastOntologyVersionUri == null)
        {
            throw new IllegalArgumentException("Ontology URI and ontology version URI must not both be null");
        }
        
        final String ontologyVersionUriString;
        RepositoryConnection conn = null;
        URI managedOntologyVersionUri = null;
        
        if(lastOntologyVersionUri != null)
        {
            managedOntologyVersionUri = lastOntologyVersionUri;
            
        }
        // Could not use the version URI, so setup the managed ontology version URI using the
        // ontology URI
        else if(ontologyUri != null)
        {
            managedOntologyVersionUri =
                    OasConstants.VF.createURI(OasConstants.AUTOGENERATED_NS, ontologyUri.stringValue());
        }
        
        final ValueFactory valueFactory = ontologyManagerRepository.getValueFactory();
        
        try
        {
            conn = ontologyManagerRepository.getConnection();
            
            // while the repository has an existing version matching our current guess we keep
            // modifying the guess until we find one that does not exist
            // NOTE: This method does not reserve the URI in the repository, this method is
            // read-only on the repository. Hence, when storing the actual ontology, a check should
            // be made to ensure the URI has not been used in the meantime.
            while(conn.hasStatement(null, OasConstants.OWL_VERSIONIRI, managedOntologyVersionUri, false, contexts))
            {
                managedOntologyVersionUri = SesameUtils.generateNextURI(managedOntologyVersionUri, valueFactory);
            }
        }
        catch(final RepositoryException rex)
        {
            SesameUtils.LOG
                    .error("Found repository exception while trying to determine a suitable managed ontology version uri.");
            managedOntologyVersionUri = null;
            throw rex;
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException rex2)
                {
                    SesameUtils.LOG
                            .error("Found repository exception when trying to close repository connection", rex2);
                }
            }
        }
        
        return managedOntologyVersionUri;
    }
    
    /**
     * Helper method to generate the next URI candidate in a sequence based on the given URI.
     * 
     * Public to enable testing.
     * 
     * @param managedOntologyVersionUri
     * @param valueFactory
     * @return
     */
    public static URI generateNextURI(final URI managedOntologyVersionUri, final ValueFactory valueFactory)
    {
        String ontologyVersionUriString;
        
        ontologyVersionUriString = managedOntologyVersionUri.stringValue();
        
        // start off by assuming we will be putting on "1" to make it unique
        final String newLastChar = "1";
        
        final String suffix = SesameUtils.getNumericSuffix(ontologyVersionUriString);
        
        if(suffix.isEmpty())
        {
            // if the last character was not a digit, then just append 1 to the end,
            // essentially
            return valueFactory.createURI(ontologyVersionUriString + newLastChar);
        }
        else
        {
            final String newSuffix = SesameUtils.getNextSuffix(suffix);
            
            return valueFactory.createURI(ontologyVersionUriString.substring(0, ontologyVersionUriString.length()
                    - suffix.length())
                    + newSuffix);
        }
    }
    
    /**
     * Given a suffix that is all numeric, this method generates the string version of the next
     * number in that sequence.
     * 
     * @param suffix
     * @return
     */
    private static String getNextSuffix(String suffix)
    {
        String leadingZeroes = "";
        
        while(suffix.startsWith("0"))
        {
            if(suffix.length() > 1 && suffix.charAt(1) != '9')
            {
                leadingZeroes = leadingZeroes + "0";
            }
            if(suffix.length() == 1)
            {
                leadingZeroes = leadingZeroes + "0";
            }
            
            suffix = suffix.substring(1);
        }
        
        if(suffix.isEmpty() && leadingZeroes.isEmpty())
        {
            return "1";
        }
        else if(suffix.isEmpty())
        {
            return leadingZeroes.substring(0, leadingZeroes.length() - 1) + "1";
        }
        else
        {
            return leadingZeroes + Integer.toString(Integer.valueOf(suffix) + 1, 10);
        }
    }
    
    private static String getNumericSuffix(final String input)
    {
        String result = "";
        
        try
        {
            for(int i = input.length() - 1; i > 0; i--)
            {
                result = Integer.parseInt(input.substring(i, i + 1), 10) + result;
            }
        }
        catch(final NumberFormatException nfe)
        {
            // ignore
        }
        
        return result;
    }
    
    public static Map<URI, Annotation> parseAndStoreAnnotations(final Representation entity,
            final DataHandler nextDataHandler, final URI userUri, final String baseUri) throws ResourceException
    {
        Map<URI, Annotation> results = null;
        try
        {
            
            final Map<URI, Annotation> map =
                    SesameUtils.parseAnnotations(entity.getReader(), baseUri, entity.getMediaType().getName());
            
            for(final URI nextUri : map.keySet())
            {
                final Annotation nextAnnotation = map.get(nextUri);
                
                if(nextAnnotation.getAnnotator() == null)
                {
                    SesameUtils.LOG.debug("default annotator: {}", userUri);
                    nextAnnotation.setAnnotator(userUri);
                }
                else
                {
                    SesameUtils.LOG.debug("annotator: {}", nextAnnotation.getAnnotator());
                }
                
                // delete the old annotation using this Id
                nextDataHandler.deleteAnnotation(nextUri);
                
                // then store the new version of the annotation
                if(!nextDataHandler.storeAnnotation(nextAnnotation))
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not store annotation");
                }
            }
            
            results = map;
        }
        catch(final IOException ex)
        {
            // return HTTP 400 client error
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not store annotation", ex);
        }
        
        return results;
    }
    
    /**
     * Creates a single annotation from the reader, without previously knowing a URI.
     * 
     * This method is used when Annotations are first submitted, and they do not have a URI.
     * 
     * Wraps SesameUtils.createAnnotationsFromRepository to generate the Annotations.
     * 
     * @param inputReader
     *            The reader containing the serialised RDF.
     * @param baseUri
     *            The base URI to use for any relative references in the serialised RDF.
     * @param mimeType
     *            The serialisation format of the RDF in the inputReader.
     * @return The first annotation that is created based on the serialised RDF.
     */
    public static Map<URI, Annotation> parseAnnotations(final Reader inputReader, final String baseUri,
            final String mimeType)
    {
        final SailRepository myRepository = new SailRepository(new MemoryStore());
        RepositoryConnection conn = null;
        
        try
        {
            myRepository.initialize();
            
            conn = myRepository.getConnection();
            
            conn.begin();
            
            // Try to get a parser for their mime-type, or default to RDF/XML if that fails
            final RDFFormat inputFormat = Rio.getParserFormatForMIMEType(mimeType, RDFFormat.RDFXML);
            
            // Parse the input using the format
            conn.add(inputReader, baseUri, inputFormat, (URI)null);
            
            conn.commit();
            
            // create a set of annotations, but only return one
            final Map<URI, Annotation> results =
                    SesameUtils.createAnnotationsFromRepository(myRepository, baseUri, (URI)null);
            
            return results;
        }
        catch(final RepositoryException | RDFParseException | IOException ex)
        {
            SesameUtils.LOG.error("Found exception", ex);
            try
            {
                if(conn != null && conn.isActive())
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException e)
            {
                SesameUtils.LOG.error("Unexpected RepositoryException ", e);
            }
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                SesameUtils.LOG.error("Found RepositoryException", ex);
            }
        }
    }
    
    /**
     * Creates a single ontology from the inputStream, without previously knowing a URI.
     * 
     * @param inputStream
     *            The input stream containing the serialised RDF.
     * @param baseUri
     *            The base URI to use for any relative references in the serialised RDF.
     * @param mimeType
     *            The serialisation format of the RDF in the inputReader.
     * @return The first ontology that is created based on the serialised RDF.
     */
    public static Map<URI, List<Ontology>> parseOntologies(final InputStream inputStream, final String baseUri,
            final String mimeType)
    {
        final SailRepository myRepository = new SailRepository(new MemoryStore());
        RepositoryConnection conn = null;
        
        try
        {
            myRepository.initialize();
            
            conn = myRepository.getConnection();
            
            // Try to get a parser for their mime-type, or default to RDF/XML if that fails
            final RDFFormat inputFormat = Rio.getParserFormatForMIMEType(mimeType, RDFFormat.RDFXML);
            
            // Parse the input using the format
            conn.add(inputStream, baseUri, inputFormat);
            
            final Map<URI, List<Ontology>> results = SesameUtils.createOntologiesFromRepository(myRepository);
            
            return results;
        }
        catch(final RepositoryException ex)
        {
            SesameUtils.LOG.error("Found RepositoryException", ex);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        catch(final RDFParseException ex)
        {
            SesameUtils.LOG.error("Found RDFParseException", ex);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        catch(final IOException ex)
        {
            SesameUtils.LOG.error("Found IOException", ex);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    SesameUtils.LOG.error("Found RepositoryException", ex);
                }
            }
        }
    }
    
    /**
     * Creates a map of ontologies and their discovered versions from the reader, without previously
     * knowing a URI.
     * 
     * @param inputReader
     *            The reader containing the serialised RDF.
     * @param baseUri
     *            The base URI to use for any relative references in the serialised RDF.
     * @param mimeType
     *            The serialisation format of the RDF in the inputReader.
     * @return The first ontology that is created based on the serialised RDF.
     */
    public static Map<URI, List<Ontology>> parseOntologies(final Reader inputReader, final String baseUri,
            final String mimeType)
    {
        final SailRepository myRepository = new SailRepository(new MemoryStore());
        RepositoryConnection conn = null;
        
        try
        {
            myRepository.initialize();
            
            conn = myRepository.getConnection();
            
            // Try to get a parser for their mime-type, or default to RDF/XML if that fails
            final RDFFormat inputFormat = Rio.getParserFormatForMIMEType(mimeType, RDFFormat.RDFXML);
            
            // Parse the input using the format
            conn.add(inputReader, baseUri, inputFormat);
            
            // create a set of ontologies, but only return one
            final Map<URI, List<Ontology>> results = SesameUtils.createOntologiesFromRepository(myRepository);
            
            return results;
        }
        catch(final RepositoryException ex)
        {
            SesameUtils.LOG.error("Found RepositoryException", ex);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        catch(final RDFParseException ex)
        {
            SesameUtils.LOG.error("Found RDFParseException", ex);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        catch(final IOException ex)
        {
            SesameUtils.LOG.error("Found IOException", ex);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        finally
        {
            if(conn != null)
            {
                try
                {
                    conn.close();
                }
                catch(final RepositoryException ex)
                {
                    SesameUtils.LOG.error("Found RepositoryException", ex);
                }
            }
        }
    }
    
    /**
     * Creates a single ResourceCount based on the serialised RDF that is discovered in the given
     * Reader.
     * 
     * @param inputReader
     *            The input reader that contains serialised RDF.
     * @param baseUri
     *            The base URI to use for any relative URIs in the serialised RDF.
     * @param mimeType
     *            The mime type of the serialised RDF in the input reader.
     * @return A single ResourceCount discovered in the serialised RDF
     */
    public static ResourceCount parseUriCounts(final Reader inputReader, final String baseUri, final String mimeType)
    {
        final UriCountImpl result = new UriCountImpl();
        
        try
        {
            final Model model =
                    Rio.parse(inputReader, baseUri, Rio.getParserFormatForMIMEType(mimeType, RDFFormat.RDFXML));
            
            final Set<Value> directCounts = model.filter(null, OasConstants.OAS_URI_DIRECT_COUNT, null).objects();
            
            if(!directCounts.isEmpty())
            {
                for(final Value nextValue : directCounts)
                {
                    result.setCount(((Literal)nextValue).intValue());
                }
            }
            else
            {
                final Set<Value> recursiveCounts =
                        model.filter(null, OasConstants.OAS_URI_RECURSIVE_COUNT, null).objects();
                
                for(final Value nextValue : recursiveCounts)
                {
                    result.setCount(((Literal)nextValue).intValue());
                }
            }
            
            return result;
        }
        catch(final RDFParseException ex)
        {
            SesameUtils.LOG.error("Found RDFParseException", ex);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
        catch(final IOException ex)
        {
            SesameUtils.LOG.error("Found IOException", ex);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, ex);
        }
    }
    
    /**
     * Store the ontology from the given Reader, based on the given metadata in nextOntology, the
     * RDFFormat given by rdfFormat, in the Repository, specified by myRepository.
     * 
     * NOTE: This is currently very inefficient, as I do not want to duplicate code, so this uses a
     * hack to convert the Reader to an InputStream.
     * 
     * @param nextOntology
     *            The ontology metadata to use
     * @param inputStream
     *            The input stream containing the ontology.
     * @param baseUri
     *            The base URI to use for the ontology.
     * @param rdfFormat
     *            The RDFFormat that the ontology is encoded in.
     * @param myRepository
     *            A repository to insert the statements into. If it is null, a new in memory store
     *            will be created.
     * @param ontologyManagerGraphs
     *            The graph or graphs in the given repository to store ontology manager information
     *            in. These must not include the URIs obtained using nextOntology.getOntologyUri or
     *            nextOntology.getManagedOntologyVersionUri.
     * @return Repository containing the RDF statements.
     * @throws RepositoryException
     *             If the repository failed to complete the operation for some reason.
     * @throws RDFParseException
     *             If the inputStream did not contain a valid RDF document based on the given
     *             rdfFormat.
     * @throws IOException
     *             If the inputStream failed during the operation.
     * @throws RDFHandlerException
     *             If the loading process failed for some reason.
     * @throws IllegalArgumentException
     *             if nextOntology is null, or nextOntology.getOntologyUri() is null.
     */
    public static Repository storeOntology(final Ontology nextOntology, final InputStream inputStream,
            final String baseUri, final RDFFormat rdfFormat, final Repository myRepository,
            final URI... ontologyManagerGraphs) throws RepositoryException, RDFParseException, IOException,
        RDFHandlerException
    {
        OpenRDFUtil.verifyContextNotNull(ontologyManagerGraphs);
        
        return SesameUtils.storeOntology(nextOntology, new BufferedReader(new InputStreamReader(inputStream)), baseUri,
                rdfFormat, myRepository, ontologyManagerGraphs);
    }
    
    public static Repository storeOntology(final Ontology nextOntology, final Iterable<Statement> statements,
            final Map<String, String> prefixMap, final Iterable<Statement> inferredStatements,
            final Repository myRepository, final URI... ontologyManagerGraphs) throws RepositoryException,
        RDFParseException, IOException, RDFHandlerException
    {
        OpenRDFUtil.verifyContextNotNull(ontologyManagerGraphs);
        
        // final URI managedOntologyVersionUri =
        // nextOntology.getManagedOntologyVersionUri(myRepository, true, false,
        // ontologyManagerGraphs);
        
        final RepositoryConnection conn = myRepository.getConnection();
        
        try
        {
            SesameUtils.addStandardNamespaces(conn);
            
            if(prefixMap != null)
            {
                for(final String nextPrefixString : prefixMap.keySet())
                {
                    // Do not persist null prefix strings or those which are empty, so that
                    // there is no interfering with the default prefix, which is specified using
                    // the empty string
                    if(nextPrefixString != null && !nextPrefixString.isEmpty())
                    {
                        final String existingNamespace = conn.getNamespace(nextPrefixString);
                        
                        // do not overwrite existing prefix definitions by only setting the
                        // namespace if there was no existing namespace mapped to that prefix
                        final String nextNamespaceString = prefixMap.get(nextPrefixString);
                        if(existingNamespace == null)
                        {
                            if(SesameUtils.LOG.isDebugEnabled())
                            {
                                SesameUtils.LOG.debug(
                                        "Setting namespace prefix into repository prefix={} nextNamespaceString={}",
                                        nextPrefixString, nextNamespaceString);
                            }
                            conn.setNamespace(nextPrefixString, nextNamespaceString);
                        }
                        // if the existingNamespace was different to the one that was submitted
                        // then log it, otherwise ignore it
                        else if(SesameUtils.LOG.isTraceEnabled() && !existingNamespace.equals(nextNamespaceString))
                        {
                            SesameUtils.LOG
                                    .trace("Did not persist namespace because there was a conflicting namespace already defined prefix="
                                            + nextPrefixString
                                            + " existingNamespace="
                                            + existingNamespace
                                            + " notPersistedNamespace=" + nextNamespaceString);
                        }
                    }
                }
            }
            
            conn.add(statements, nextOntology.getOntologyVersionUri());
            
            if(inferredStatements != null && nextOntology.getInferredOntologyUri() != null)
            {
                conn.add(inferredStatements, nextOntology.getInferredOntologyUri());
            }
            
        }
        finally
        {
            if(conn != null)
            {
                conn.close();
            }
        }
        
        return SesameUtils.storeOntology(nextOntology, nextOntology.getOntologyVersionUri(), myRepository,
                ontologyManagerGraphs);
    }
    
    /**
     * Store the ontology from the given Iterable<Statement>, based on the given metadata in
     * nextOntology, the RDFFormat given by rdfFormat, in the Repository, specified by myRepository.
     * 
     * This is the best method for storing in memory statements, as it does not require
     * serialisation to a concrete RDF format.
     * 
     * @param nextOntology
     *            The ontology metadata to use
     * @param statements
     *            The RDF statements containing the ontology.
     * @param myRepository
     *            A repository to insert the statements into. If it is null, a new in memory store
     *            will be created.
     * @param ontologyManagerGraphs
     *            The graph or graphs in the given repository to store ontology manager information
     *            in. These must not include the URIs obtained using nextOntology.getOntologyUri or
     *            nextOntology.getManagedOntologyVersionUri.
     * @return Repository containing the RDF statements.
     * @throws RepositoryException
     *             If the repository failed to complete the operation for some reason.
     * @throws RDFHandlerException
     *             If the loading process failed for some reason.
     * @throws IllegalArgumentException
     *             if nextOntology is null, or nextOntology.getOntologyUri() is null.
     */
    public static Repository storeOntology(final Ontology nextOntology, final Iterable<Statement> statements,
            final Map<String, String> prefixMap, final Repository myRepository, final URI... ontologyManagerGraphs)
        throws RepositoryException, RDFParseException, IOException, RDFHandlerException
    {
        OpenRDFUtil.verifyContextNotNull(ontologyManagerGraphs);
        
        return SesameUtils
                .storeOntology(nextOntology, statements, prefixMap, null, myRepository, ontologyManagerGraphs);
    }
    
    /**
     * Store the ontology from the given Reader, based on the given metadata in nextOntology, the
     * RDFFormat given by rdfFormat, in the Repository, specified by myRepository.
     * 
     * NOTE: This is currently very inefficient, as I do not want to duplicate code, so this uses a
     * hack to convert the Reader to an InputStream.
     * 
     * @param nextOntology
     *            The ontology metadata to use
     * @param inputStream
     *            The input stream containing the ontology.
     * @param baseUri
     *            The base URI to use for the ontology.
     * @param rdfFormat
     *            The RDFFormat that the ontology is encoded in.
     * @param myRepository
     *            A repository to insert the statements into. If it is null, a new in memory store
     *            will be created.
     * @param ontologyManagerGraphs
     *            The graph or graphs in the given repository to store ontology manager information
     *            in. These must not include the URIs obtained using nextOntology.getOntologyUri or
     *            nextOntology.getManagedOntologyVersionUri.
     * @return Repository containing the RDF statements.
     * @throws RepositoryException
     *             If the repository failed to complete the operation for some reason.
     * @throws RDFParseException
     *             If the inputStream did not contain a valid RDF document based on the given
     *             rdfFormat.
     * @throws IOException
     *             If the inputStream failed during the operation.
     * @throws RDFHandlerException
     *             If the loading process failed for some reason.
     * @throws IllegalArgumentException
     *             if nextOntology is null, or nextOntology.getOntologyUri() is null.
     */
    public static Repository storeOntology(final Ontology nextOntology, final Reader inputStream, final String baseUri,
            final RDFFormat rdfFormat, final Repository myRepository, final URI... ontologyManagerGraphs)
        throws RepositoryException, RDFParseException, IOException, RDFHandlerException
    {
        OpenRDFUtil.verifyContextNotNull(ontologyManagerGraphs);
        
        Repository resultRepository = myRepository;
        
        if(nextOntology == null)
        {
            throw new IllegalArgumentException("Ontology must not be null.");
        }
        else if(nextOntology.getOntologyUri() == null)
        {
            throw new IllegalArgumentException("Ontology URI must not be null.");
        }
        
        // final URI managedOntologyVersionUri =
        // nextOntology.getManagedOntologyVersionUri(resultRepository, true, false,
        // ontologyManagerGraphs);
        
        nextOntology.ensureVersionUriUniqueness(resultRepository, true, false, ontologyManagerGraphs);
        
        if(nextOntology.getOntologyVersionUri() == null)
        {
            // NOTE: There is no way using OntologyImpl to get to this line, so to test it we need
            // to use a mock Ontology implementation that allows managed ontology version URI to be
            // null while ontologyURI is not null
            throw new IllegalArgumentException("Managed ontology URI must not be null");
        }
        
        for(final URI nextOntologyManagerGraph : ontologyManagerGraphs)
        {
            if(nextOntology.getOntologyUri().equals(nextOntologyManagerGraph))
            {
                SesameUtils.LOG
                        .error("SECURITY: Someone tried to overwrite the ontology manager graph by specifying its URI as their ontology URI");
                throw new RuntimeException("Invalid ontology URI");
            }
            
            if(nextOntology.getOntologyVersionUri().equals(nextOntologyManagerGraph))
            {
                SesameUtils.LOG
                        .error("SECURITY: Someone tried to overwrite the ontology manager graph by specifying its URI as their ontology version URI");
                throw new RuntimeException("Invalid ontology URI");
            }
            
            // FIXME: Check that the version URI does not match any existing graph URI at all
        }
        
        RepositoryConnection conn = null;
        
        try
        {
            // create a repository for them if they did not supply one
            if(resultRepository == null)
            {
                resultRepository = new SailRepository(new MemoryStore());
                resultRepository.initialize();
            }
            
            conn = resultRepository.getConnection();
            
            conn.begin();
            
            // FIXME: The asynchronous nature of the sesame-loader LoaderMain class seems to make
            // Restlet nervous and fails to complete queries for large ontology loads
            // TODO: make these parameters configurable
            // final LoaderMain loader =
            // new LoaderMain(new ArbitraryRepositoryManager(resultRepository), new Integer(20000),
            // new Integer(5), managedOntologyVersionUri);
            // loader.load(inputStream, rdfFormat, baseUri);
            conn.add(inputStream, baseUri, rdfFormat, nextOntology.getOntologyVersionUri());
            
            conn.commit();
        }
        catch(final RepositoryException | RDFParseException | IOException e)
        {
            if(conn != null && conn.isActive())
            {
                conn.rollback();
            }
            
            throw e;
        }
        finally
        {
            if(conn != null && conn.isOpen())
            {
                conn.close();
            }
        }
        return SesameUtils.storeOntology(nextOntology, nextOntology.getOntologyVersionUri(), resultRepository,
                ontologyManagerGraphs);
    }
    
    /**
     * Store the ontology from the given InputStream, based on the given metadata in nextOntology,
     * the RDFFormat given by rdfFormat, in the Repository, specified by myRepository.
     * 
     * @param nextOntology
     *            The ontology metadata to use
     * @param myRepository
     *            A repository to insert the statements into. If it is null, a new in memory store
     *            will be created.
     * @param ontologyManagerGraphs
     *            The graph or graphs in the given repository to store ontology manager information
     *            in. These must not include the URIs obtained using nextOntology.getOntologyUri or
     *            nextOntology.getManagedOntologyVersionUri.
     * @return Repository containing the RDF statements.
     * @throws RepositoryException
     *             If the repository failed to complete the operation for some reason.
     * @throws RDFParseException
     *             If the inputStream did not contain a valid RDF document based on the given
     *             rdfFormat.
     * @throws IOException
     *             If the inputStream failed during the operation.
     * @throws RDFHandlerException
     *             If the loading process failed for some reason.
     * @throws IllegalArgumentException
     *             if nextOntology is null, or nextOntology.getOntologyUri() is null.
     */
    public static Repository storeOntology(final Ontology nextOntology, final URI managedOntologyVersionUri,
            final Repository myRepository, final URI... ontologyManagerGraphs) throws RepositoryException,
        RDFParseException, IOException, RDFHandlerException
    {
        
        final RepositoryConnection conn = myRepository.getConnection();
        try
        {
            final ValueFactory vf = conn.getValueFactory();
            
            // Then once the entire ontology has been inserted, we update the metadata
            
            // Remove any ontology version information that exists in the new context so that we can
            // be sure it matches with the ontology manager graph
            conn.remove(nextOntology.getOntologyUri(), OasConstants.OWL_VERSIONIRI, null, managedOntologyVersionUri);
            conn.remove(nextOntology.getOntologyUri(), OWL.PRIORVERSION, null, managedOntologyVersionUri);
            
            conn.add(nextOntology.getOntologyUri(), RDF.TYPE, OWL.ONTOLOGY, managedOntologyVersionUri);
            
            // Overwrite the ontology version URI with our managed version to avoid conflicts if
            // people neglect to update this field between uploads and wish to still store both
            // versions simultaneously
            conn.add(nextOntology.getOntologyUri(), OasConstants.OWL_VERSIONIRI, managedOntologyVersionUri,
                    managedOntologyVersionUri);
            
            // NOTE: This is the only location where ontologyVersionUri is used on its raw form, and
            // the version URI is only inserted as a literal, so it will not be automatically
            // matched as a URI with any queries at all
            // TODO: Will this cause errors for OWL parsers if they expect VERSIONINFO to be a URI
            // and not a literal?
            if(nextOntology.getOntologyVersionUri() != null)
            {
                // The Literal attached to OWL.VERSIONINFO is typed to xsd:anyURI so that it can be
                // recognised as a URI, but not a Resource or an owl:Thing
                conn.add(
                        nextOntology.getOntologyUri(),
                        OWL.VERSIONINFO,
                        conn.getValueFactory().createLiteral(nextOntology.getOntologyVersionUri().stringValue(),
                                OasConstants.XSD_ANYURI), managedOntologyVersionUri);
            }
            
            // Then update the management graph that contains the master copy of the metadata so
            // that it can be searched quickly
            
            // if there was metadata for this ontology URI previously then we need to create the
            // list of versions in the graph
            if(conn.hasStatement(nextOntology.getOntologyUri(), null, null, false, ontologyManagerGraphs))
            {
                // Find all of the versions that are known about this ontology
                final List<Statement> priorVersions =
                        Iterations.asList(conn.getStatements(nextOntology.getOntologyUri(),
                                OasConstants.OWL_VERSIONIRI, null, false, ontologyManagerGraphs));
                
                // reset the OMV:CURRENTVERSION pointer for this ontology URI to the current managed
                // ontology version URI
                conn.remove(nextOntology.getOntologyUri(), OasConstants.OMV_CURRENTVERSION, null, ontologyManagerGraphs);
                
                // remove the list of prior versions that are known so it can be set using our local
                // knowledge
                // in particular, a case where the current ontology version was reverted to a prior
                // version may be inconsistent if we did not perform this step
                conn.remove(nextOntology.getOntologyUri(), OWL.PRIORVERSION, null, ontologyManagerGraphs);
                
                for(final Statement nextPriorVersion : priorVersions)
                {
                    if(nextPriorVersion.getObject() instanceof URI)
                    {
                        final URI nextVersionUri = (URI)nextPriorVersion.getObject();
                        
                        // only insert the prior version statement if it isn't the current version
                        if(!nextVersionUri.equals(managedOntologyVersionUri))
                        {
                            // NOTE: This may be more efficient in one operation by specifying both
                            // contexts in a single line but it would not be as easy to maintain
                            conn.add(nextOntology.getOntologyUri(), OWL.PRIORVERSION, nextVersionUri,
                                    managedOntologyVersionUri);
                            conn.add(nextOntology.getOntologyUri(), OWL.PRIORVERSION, nextVersionUri,
                                    ontologyManagerGraphs);
                        }
                    }
                }
            }
            
            // Then once the ontology graph metadata is updated, add the metadata for the current
            // version
            // of the ontology to the ontology manager
            conn.add(nextOntology.getOntologyUri(), RDF.TYPE, OWL.ONTOLOGY, ontologyManagerGraphs);
            conn.add(nextOntology.getOntologyUri(), OasConstants.OWL_VERSIONIRI, managedOntologyVersionUri,
                    ontologyManagerGraphs);
            conn.add(nextOntology.getOntologyUri(), OasConstants.OMV_CURRENTVERSION, managedOntologyVersionUri,
                    ontologyManagerGraphs);
            
            // if there was an inferred axioms ontology that is linked to from this ontology, then
            // add both a current and version specific reference to it in the ontology manager graph
            if(nextOntology.getInferredOntologyUri() != null)
            {
                conn.add(managedOntologyVersionUri, OasConstants.OAS_INFERREDAXIOMSONTOLOGY,
                        nextOntology.getInferredOntologyUri(), ontologyManagerGraphs);
                conn.add(nextOntology.getOntologyUri(), OasConstants.OAS_CURRENTINFERREDAXIOMSONTOLOGY,
                        nextOntology.getInferredOntologyUri(), ontologyManagerGraphs);
                conn.add(nextOntology.getInferredOntologyUri(), RDF.TYPE, OWL.ONTOLOGY,
                        nextOntology.getInferredOntologyUri());
                conn.add(nextOntology.getInferredOntologyUri(), OWL.IMPORTS, managedOntologyVersionUri,
                        nextOntology.getInferredOntologyUri());
            }
            
            for(final OWLProfile nextProfile : nextOntology.getProfiles())
            {
                conn.add(managedOntologyVersionUri, OasConstants.OAS_INPROFILE, nextProfile.getIRI().toOpenRDFURI(),
                        ontologyManagerGraphs);
            }
            
            if(nextOntology.getConcreteAxiomCount() > 0)
            {
                conn.add(managedOntologyVersionUri, OasConstants.OAS_CONCRETE_AXIOM_COUNT,
                        vf.createLiteral(nextOntology.getConcreteAxiomCount()), ontologyManagerGraphs);
            }
            
            if(nextOntology.getConcreteLogicalAxiomCount() > 0)
            {
                conn.add(managedOntologyVersionUri, OasConstants.OAS_CONCRETE_LOGICAL_AXIOM_COUNT,
                        vf.createLiteral(nextOntology.getConcreteLogicalAxiomCount()), ontologyManagerGraphs);
            }
            
            if(nextOntology.getIndividualCount() > 0)
            {
                conn.add(managedOntologyVersionUri, OasConstants.OAS_INDIVIDUAL_COUNT,
                        vf.createLiteral(nextOntology.getIndividualCount()), ontologyManagerGraphs);
            }
            
            if(nextOntology.getInferredAxiomCount() > 0)
            {
                conn.add(managedOntologyVersionUri, OasConstants.OAS_INFERRED_AXIOM_COUNT,
                        vf.createLiteral(nextOntology.getInferredAxiomCount()), ontologyManagerGraphs);
            }
            
            // TODO: Add more OMV support in metadata
            
            // IF and only IF everything went okay, then commit the transaction
            conn.commit();
        }
        catch(final RepositoryException ex)
        {
            conn.rollback();
            throw ex;
        }
        // catch(final RDFParseException ex)
        // {
        // conn.rollback();
        // throw ex;
        // }
        // catch(final IOException ex)
        // {
        // conn.rollback();
        // throw ex;
        // }
        // catch(final SailException ex)
        // {
        // conn.rollback();
        // throw new RepositoryException(ex);
        // }
        // catch(final RDFHandlerException ex)
        // {
        // conn.rollback();
        // throw ex;
        // }
        catch(final UnsupportedRDFormatException ex)
        {
            conn.rollback();
            throw ex;
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                throw ex;
            }
        }
        
        return myRepository;
    }
    
    /**
     * Export the annotation into a Sesame RDF repository.
     * 
     * Any null properties on the annotation will not appear in the RDF document.
     * 
     * @param annotation
     *            The annotation to generate RDF statements from. The annotations unique identifier
     *            is the context to use to find the resulting statements.
     * @param myRepository
     *            A repository to insert the statements into. If it is null, a new in memory store
     *            will be created.
     * @param annotationManagementGraph
     *            TODO
     * @return Repository containing the RDF statements.
     * @throws RepositoryException
     */
    public static void toRDFRepository(final Annotation annotation, final Repository myRepository,
            final URI annotationManagementGraph) throws RepositoryException
    {
        RepositoryConnection conn = null;
        
        try
        {
            conn = myRepository.getConnection();
            
            conn.add(annotation.getAnnotationUniqueId(), RDF.TYPE, OasConstants.TAGGING_TAGGING,
                    annotationManagementGraph);
            
            conn.add(annotation.toRdf(), annotation.getAnnotationUniqueId());
        }
        catch(final RepositoryException ex)
        {
            throw ex;
        }
        finally
        {
            try
            {
                if(conn != null)
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                throw ex;
            }
        }
    }
    
    /**
     * Serialises search results to a repsitory using the given URI as a predicate
     * 
     * @param searchResults
     *            A map of URIs from keys to result values to indicate which results came from which
     *            ontologies.
     * @param predicateUri
     *            The predicate URI to use for triples mapping the search result keys to the search
     *            result values.
     * @param myRepository
     *            The repository to serialise to, or null to dynamically create a repository.
     * @throws RepositoryException
     * @return A repository containing the search results.
     */
    public static Repository toRDFRepository(final List<SearchResult> searchResults, final URI predicateUri,
            final Repository myRepository) throws RepositoryException
    {
        if(searchResults == null)
        {
            throw new IllegalArgumentException("Search results must not be null");
        }
        
        Repository resultRepository = myRepository;
        
        RepositoryConnection conn = null;
        
        try
        {
            // create a repository for them if they did not supply one
            if(resultRepository == null)
            {
                resultRepository = new SailRepository(new MemoryStore());
                resultRepository.initialize();
            }
            
            conn = resultRepository.getConnection();
            
            conn.begin();
            
            SesameUtils.addStandardNamespaces(conn);
            
            for(final SearchResult nextResult : searchResults)
            {
                final URI nextOntology = nextResult.getOntology().getOntologyUri();
                
                conn.add(nextResult.getResultUri(), predicateUri, nextOntology);
                
                if(nextResult.getLabels() != null)
                {
                    for(final String nextLabel : nextResult.getLabels())
                    {
                        conn.add(nextResult.getResultUri(), RDFS.LABEL, conn.getValueFactory().createLiteral(nextLabel));
                    }
                }
            }
            
            conn.commit();
        }
        catch(final RepositoryException ex)
        {
            try
            {
                if(conn != null && conn.isActive())
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException ex1)
            {
                SesameUtils.LOG.error("Found repository exception when trying to rollback repository connection", ex1);
            }
            throw ex;
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                SesameUtils.LOG.error("Found repository exception when trying to close repository connection", ex);
            }
        }
        
        return resultRepository;
    }
    
    /**
     * Serialises Ontology metadata to a Repository
     * 
     * @param nextOntology
     *            The ontology metadata to be serialised
     * @param myRepository
     *            The repository to serialise to, or null to dynamically create a repository
     * @return A repository containing the ontology metadata in a context named using
     *         ontology.getManagedOntologyVersionUri()
     * @throws RepositoryException
     *             If there is a problem with the repository
     */
    public static Repository toRDFRepository(final Ontology nextOntology, final Repository myRepository)
        throws RepositoryException
    {
        if(nextOntology.getOntologyUri() == null)
        {
            throw new IllegalArgumentException("Every ontology must have a URI");
        }
        
        Repository resultRepository = myRepository;
        
        RepositoryConnection conn = null;
        
        try
        {
            // create a repository for them if they did not supply one
            if(resultRepository == null)
            {
                resultRepository = new SailRepository(new MemoryStore());
                resultRepository.initialize();
            }
            
            conn = resultRepository.getConnection();
            
            conn.begin();
            
            SesameUtils.addStandardNamespaces(conn);
            
            // final URI managedOntologyVersionUri =
            // nextOntology.getManagedOntologyVersionUri(resultRepository, false, true,
            // UriConstants.ONTOLOGY_MANAGER_GRAPH);
            
            nextOntology.ensureVersionUriUniqueness(resultRepository, false, true, SesameUtils.ONTOLOGY_MANAGER_GRAPH);
            
            conn.add(nextOntology.getOntologyUri(), RDF.TYPE, OWL.ONTOLOGY, nextOntology.getOntologyVersionUri());
            
            if(nextOntology.getOntologyVersionUri() != null)
            {
                conn.add(nextOntology.getOntologyUri(), OasConstants.OWL_VERSIONIRI,
                        nextOntology.getOntologyVersionUri(), nextOntology.getOntologyVersionUri());
            }
            
            if(nextOntology.getInferredOntologyUri() != null)
            {
                conn.add(nextOntology.getOntologyVersionUri(), OasConstants.OAS_INFERREDAXIOMSONTOLOGY,
                        nextOntology.getInferredOntologyUri(), nextOntology.getOntologyVersionUri());
            }
            
            for(final OWLProfile nextProfile : nextOntology.getProfiles())
            {
                conn.add(nextOntology.getOntologyVersionUri(), OasConstants.OAS_INPROFILE, nextProfile.getIRI()
                        .toOpenRDFURI(), nextOntology.getOntologyVersionUri());
            }
            
            conn.commit();
        }
        catch(final RepositoryException ex)
        {
            try
            {
                if(conn != null && conn.isActive())
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException ex1)
            {
                SesameUtils.LOG.error("Found repository exception when trying to rollback repository connection", ex1);
            }
            throw ex;
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                SesameUtils.LOG.error("Found repository exception when trying to close repository connection", ex);
            }
        }
        
        return resultRepository;
    }
    
    /**
     * Inserts the RDF statements representing the given ResourceCount object into myRepository,
     * creating a new repository if myRepository is not given.
     * 
     * @param nextResourceCount
     *            The ResourceCount object to create RDF statements from.
     * @param myRepository
     *            The repository to insert the RDF statements into. If it is null, a new in-memory
     *            repository is created.
     * @return If the original repository was non-null, returns a reference to this repository, or a
     *         new in-memory repository otherwise.
     */
    public static Repository toRDFRepository(final ResourceCount nextResourceCount, final Repository myRepository)
        throws RepositoryException
    {
        // copy the reference so we can create a new repository in its place if necessary
        Repository actualRepository = myRepository;
        
        RepositoryConnection conn = null;
        
        try
        {
            // create a repository for them if they did not supply one
            if(actualRepository == null)
            {
                actualRepository = new SailRepository(new MemoryStore());
                actualRepository.initialize();
            }
            
            conn = actualRepository.getConnection();
            conn.begin();
            final ValueFactory vf = conn.getValueFactory();
            
            final BNode resultNode = vf.createBNode();
            
            conn.add(resultNode, RDF.TYPE, OasConstants.URN_TAG_OBJECT_URI_COUNT);
            
            conn.add(resultNode, OasConstants.OAS_URI_DIRECT_COUNT, vf.createLiteral(nextResourceCount.getCount()));
            conn.commit();
        }
        catch(final RepositoryException ex)
        {
            try
            {
                if(conn != null && conn.isActive())
                {
                    conn.rollback();
                }
            }
            catch(final RepositoryException ex1)
            {
                SesameUtils.LOG.error("Found repository exception when trying to rollback repository connection", ex1);
            }
            
            throw ex;
        }
        finally
        {
            try
            {
                if(conn != null && conn.isOpen())
                {
                    conn.close();
                }
            }
            catch(final RepositoryException ex)
            {
                SesameUtils.LOG.error("Found repository exception when trying to close repository connection", ex);
            }
        }
        
        return actualRepository;
    }
    
    /**
     * Private default constructor.
     */
    private SesameUtils()
    {
        // should not need to instantiate this class
    }
}
