package com.github.ansell.oas.webservice.impl;

import java.util.List;

import org.restlet.Context;
import org.restlet.Request;
import org.restlet.Response;
import org.restlet.data.Method;
import org.restlet.data.Status;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;
import org.restlet.resource.ServerResource;
import org.restlet.security.User;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.data.DataHandler;
import com.github.ansell.oas.webservice.OasAction;
import com.github.ansell.oas.webservice.OasWebServiceApplication;
import com.github.ansell.propertyutil.PropertyUtil;

public abstract class AbstractOasResourceImpl extends ServerResource
{
    protected final Logger log = LoggerFactory.getLogger(this.getClass());
    
    public AbstractOasResourceImpl()
    {
        super();
    }
    
    /**
     * Checks the ability of the currently authenticated user to perform the given action.
     * 
     * @param action
     *            The OasAction that is to be performed.
     * @throws ResourceException
     *             with Status.CLIENT_ERROR_UNAUTHORIZED (HTTP 401) if the user is not authorised to
     *             perform the given action
     */
    protected void checkAuthentication(final OasAction action) throws ResourceException
    {
        String message = "";
        
        boolean result = this.getOasApplication().authenticate(action, this.getRequest(), this.getResponse());
        
        if(!result)
        {
            message = action.getErrorMessage();
        }
        
        // Only audit actions that require authentication
        if(action.isAuthRequired())
        {
            try
            {
                final String referenceUri = this.getRequest().getResourceRef().toString(true, true);
                final User currentUser = this.getRequest().getClientInfo().getUser();
                final String ipAddress = this.getRequest().getClientInfo().getUpstreamAddress();
                this.getDataHandler().addLogDetailsForRequest(message, referenceUri, action, currentUser, ipAddress);
            }
            catch(final Exception ex)
            {
                this.log.error("Found exception while trying to log request details: ", ex);
            }
        }
        
        if(!result)
        {
            throw new ResourceException(Status.CLIENT_ERROR_UNAUTHORIZED, action.getErrorMessage());
        }
        
    }
    
    /**
     * Set-up method that can be overridden in order to initialize the state of the resource. By
     * default it does nothing.
     * 
     * @see #init(Context, Request, Response)
     */
    @Override
    public void doInit()
    {
        super.doInit();
    }
    
    /**
     * Helper method to access this.getOasApplication().getDataHandler().
     * 
     * @return the dataHandler
     */
    protected DataHandler getDataHandler()
    {
        return this.getOasApplication().getDataHandler();
    }
    
    public OasWebServiceApplication getOasApplication()
    {
        final OasWebServiceApplication application = (OasWebServiceApplication)super.getApplication();
        
        return application;
    }
    
    protected PropertyUtil getPropertyUtil()
    {
        return this.getOasApplication().getPropertyUtil();
    }
    
    /**
     * Overriding broken ServerResource.getVariants method
     * 
     * NOTE: This is not a caching implementation, so the way it is used may cause it to be a
     * performance bottleneck.
     */
    @Override
    protected List<Variant> getVariants(final Method method)
    {
        return super.getVariants(method);
    }
    
    /**
     * Redirects to the given path, using an absolute URI based on this.getRequest().getRootRef().
     * 
     * @param path
     *            The path to redirect to.
     */
    protected void redirectToPath(final String path)
    {
        final String rootRef = this.getRequest().getRootRef().toString(false, false);
        
        final boolean rootHasSlash = rootRef.endsWith("/");
        final boolean pathHasSlash = path.startsWith("/");
        
        if(rootHasSlash && pathHasSlash && path.length() > 1)
        {
            this.getResponse().redirectSeeOther(rootRef + path.substring(1));
        }
        else
        // if((rootHasSlash && !pathHasSlash) || (!rootHasSlash && pathHasSlash))
        {
            this.getResponse().redirectSeeOther(rootRef + path);
        }
    }
}