/**
 * Package for implementations of the resource handlers in the OAS API
 */
package com.github.ansell.oas.webservice.impl;

import java.util.Arrays;
import java.util.Map;

import org.openrdf.model.URI;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.sail.SailRepository;
import org.openrdf.rio.RDFHandlerException;
import org.openrdf.sail.memory.MemoryStore;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.utils.SesameUtils;
import com.github.ansell.oas.utils.UserUtils;
import com.github.ansell.oas.webservice.AnnotationByIdResource;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.restletutils.RestletUtilUser;

/**
 * Creates, fetches and updates annotations.
 * 
 * Provides responses in HTML, Talis RDF/JSON and RDF/XML.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class AnnotationByIdResourceImpl extends AbstractOasResourceImpl implements AnnotationByIdResource
{
    /**
     * Default constructor for Annotation Resource Implementation.
     */
    public AnnotationByIdResourceImpl()
    {
        super();
    }
    
    @Override
    public void deleteAnnotation() throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_DELETE);
        
        final String inputAnnotationId =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue(OasProps.PARAM_ANNOTATIONID, true);
        
        if(inputAnnotationId == null)
        {
            this.log.error("No annotation specified in request");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "No annotation specified.");
        }
        
        try
        {
            final URI annotationId = this.getDataHandler().createURI(inputAnnotationId);
            
            // Get the annotation based on the identifier URI and load it into memory
            final Annotation annotation = this.getDataHandler().getAnnotation(annotationId);
            
            if(annotation == null)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find annotation to delete");
            }
            
            if(this.getDataHandler().deleteAnnotation(annotationId))
            {
                // HTTP 204 No Content if the delete succeeded
                this.getResponse().setStatus(Status.SUCCESS_NO_CONTENT);
            }
            else
            {
                // if there were no exceptions, but the delete failed, assume it was the clients
                // fault
                // HTTP Status.CLIENT_ERROR_BAD_REQUEST Client Error if the delete did not succeed
                // but
                // there were no exceptions
                // thrown
                this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST, "Delete annotation failed");
            }
        }
        catch(final IllegalArgumentException iae)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid annotation ID");
        }
        
    }
    
    @Override
    public Representation getHtml(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final String inputAnnotationId =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue(OasProps.PARAM_ANNOTATIONID, true);
        
        if(inputAnnotationId == null)
        {
            this.log.error("No annotation specified in request");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "No annotation specified.");
        }
        
        try
        {
            final URI annotationId = this.getDataHandler().createURI(inputAnnotationId);
            
            // Get the annotation based on the identifier URI and load it into memory
            final Annotation annotation = this.getDataHandler().getAnnotation(annotationId);
            
            if(annotation == null)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find annotation");
            }
            
            final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
            dataModel.put(
                    "contentTemplate",
                    this.getOasApplication().getPropertyUtil()
                            .get(OasProps.PROP_TMPL_ANNOTATIONS_LIST, OasProps.DEF_TMPL_ANNOTATIONS_LIST));
            dataModel.put(
                    "pageTitle",
                    this.getOasApplication().getPropertyUtil()
                            .get(OasProps.PROP_TMPL_ANNOTATIONS_TITLE, OasProps.DEF_TMPL_ANNOTATIONS_TITLE));
            
            dataModel.put("annotations", Arrays.asList(annotation));
            dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
            dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
            
            return RestletUtils.getHtmlRepresentation(
                    this.getOasApplication().getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE),
                    dataModel, MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
        }
        catch(final IllegalArgumentException iae)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid annotation ID");
        }
    }
    
    @Override
    public Representation getRdf(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final String inputAnnotationId =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("annotationId", true);
        
        if(inputAnnotationId == null)
        {
            this.log.error("No annotation specified in request");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "No annotation specified.");
        }
        
        try
        {
            final URI annotationId = this.getDataHandler().createURI(inputAnnotationId);
            
            // Get the annotation based on the identifier URI and load it into memory
            final Annotation annotation = this.getDataHandler().getAnnotation(annotationId);
            
            if(annotation == null)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find annotation");
            }
            
            try
            {
                return RestletUtils.getRdfRepresentation(annotation, variant.getMediaType(), this.getRequest()
                        .getResourceRef());
            }
            catch(final RepositoryException ex)
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
            }
            catch(final RDFHandlerException ex)
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
            }
        }
        catch(final IllegalArgumentException iae)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid annotation ID");
        }
    }
    
    @Override
    public Representation putAnnotationHtml(final Representation entity, final Variant variant)
        throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_CREATE);
        
        URI currentUserUri = null;
        
        if(this.getRequest().getClientInfo().isAuthenticated())
        {
            final RestletUtilUser findUser =
                    this.getOasApplication().getRealm()
                            .findUser(this.getRequest().getClientInfo().getUser().getIdentifier());
            
            if(findUser != null)
            {
                try
                {
                    currentUserUri =
                            this.getOasApplication().getRealm()
                                    .getUserUri(this.getRequest().getClientInfo().getUser().getIdentifier());
                }
                catch(final RepositoryException e)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find user for annotation");
                }
            }
        }
        
        final Map<URI, Annotation> storeAnnotations =
                SesameUtils.parseAndStoreAnnotations(entity, this.getDataHandler(), currentUserUri, this.getRequest()
                        .getRootRef().toString(false, false));
        
        // HTTP 201 Created
        this.getResponse().setStatus(Status.SUCCESS_CREATED);
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        
        dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        
        // If the PUT instruction resulted in multiple annotations being created then display them
        // as a list
        if(storeAnnotations.size() > 1)
        {
            dataModel.put(
                    "contentTemplate",
                    this.getOasApplication().getPropertyUtil()
                            .get(OasProps.PROP_TMPL_ANNOTATIONS_LIST, OasProps.DEF_TMPL_ANNOTATIONS_LIST));
            dataModel.put(
                    "pageTitle",
                    this.getOasApplication().getPropertyUtil()
                            .get(OasProps.PROP_TMPL_ANNOTATIONS_LIST_TITLE, OasProps.DEF_TMPL_ANNOTATIONS_LIST_TITLE));
            
            dataModel.put("annotations", storeAnnotations.values());
        }
        else
        {
            // else display details just about the given annotation
            dataModel.put(
                    "contentTemplate",
                    this.getOasApplication().getPropertyUtil()
                            .get(OasProps.PROP_TMPL_ANNOTATIONS, OasProps.DEF_TMPL_ANNOTATIONS));
            dataModel.put(
                    "pageTitle",
                    this.getOasApplication().getPropertyUtil()
                            .get(OasProps.PROP_TMPL_ANNOTATIONS_TITLE, OasProps.DEF_TMPL_ANNOTATIONS_TITLE));
            if(storeAnnotations.isEmpty())
            {
                dataModel.put("annotation", null);
            }
            else
            {
                dataModel.put("annotation", storeAnnotations.values().iterator().next());
            }
        }
        return RestletUtils.getHtmlRepresentation(
                this.getOasApplication().getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE),
                dataModel, variant.getMediaType(), this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    public Representation putAnnotationRdf(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_CREATE);
        
        URI currentUserUri = null;
        
        if(this.getRequest().getClientInfo().isAuthenticated())
        {
            final RestletUtilUser findUser =
                    this.getOasApplication().getRealm()
                            .findUser(this.getRequest().getClientInfo().getUser().getIdentifier());
            
            if(findUser != null)
            {
                try
                {
                    currentUserUri =
                            this.getOasApplication().getRealm()
                                    .getUserUri(this.getRequest().getClientInfo().getUser().getIdentifier());
                }
                catch(final RepositoryException e)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find user for annotation");
                }
            }
        }
        
        final Map<URI, Annotation> storeAnnotations =
                SesameUtils.parseAndStoreAnnotations(entity, this.getDataHandler(), currentUserUri, this.getRequest()
                        .getRootRef().toString(false, false));
        
        // HTTP 201 Created
        this.getResponse().setStatus(Status.SUCCESS_CREATED);
        
        try
        {
            final Repository nextRepository = new SailRepository(new MemoryStore());
            nextRepository.initialize();
            
            for(final URI nextAnnotationUri : storeAnnotations.keySet())
            {
                SesameUtils.toRDFRepository(storeAnnotations.get(nextAnnotationUri), nextRepository,
                        SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            }
            
            return RestletUtils.toRDFSerialisation(variant.getMediaType().getName(), nextRepository, storeAnnotations
                    .keySet().toArray(new URI[0]));
        }
        catch(final RepositoryException ex)
        {
            // return HTTP Status.CLIENT_ERROR_BAD_REQUEST client error
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final RDFHandlerException ex)
        {
            // return HTTP Status.CLIENT_ERROR_BAD_REQUEST client error
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
    }
}
