/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.util.Collection;
import java.util.Map;

import org.openrdf.model.URI;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFHandlerException;
import org.restlet.data.CharacterSet;
import org.restlet.data.Form;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.objects.ResourceCount;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.utils.UserUtils;
import com.github.ansell.oas.webservice.AnnotationCountsByOntologyTermUriResource;
import com.github.ansell.oas.webservice.BaseOasActions;

/**
 * Implementation of the webservice for fetching annotations based on the URI of the ontology term.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class AnnotationCountsByOntologyTermUriResourceImpl extends AbstractOasResourceImpl implements
        AnnotationCountsByOntologyTermUriResource
{
    
    /**
     * Default constructor for webservice for fetching annotations based on the URI of the ontology.
     * term
     */
    public AnnotationCountsByOntologyTermUriResourceImpl()
    {
        super();
    }
    
    @Override
    public Representation getHtmlAnnotationCountByOntologyTermUri(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final Form queryParams = this.getReference().getQueryAsForm(CharacterSet.UTF_8);
        
        // get the query parameter, ignoring the case, so OntologyTermURI and objectTermURI will
        // both match this
        final String ontologyTerm = queryParams.getFirstValue("ontologytermuri", true);
        
        if(ontologyTerm == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                    "OntologyTermUri query parameter was not present");
        }
        
        final URI ontologyTermUri = this.getDataHandler().createURI(ontologyTerm);
        
        final Collection<Annotation> results = this.getDataHandler().getAnnotationsByOntologyTermUri(ontologyTermUri);
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST, OasProps.DEF_TMPL_ANNOTATIONS_LIST));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST_TITLE,
                        OasProps.DEF_TMPL_ANNOTATIONS_LIST_TITLE));
        
        dataModel.put("annotations", results);
        dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    public Representation getRdfAnnotationCountByOntologyTermUri(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final Form queryParams = this.getReference().getQueryAsForm(CharacterSet.UTF_8);
        
        final String ontologyTerm = queryParams.getFirstValue("ontologyTermUri", true);
        
        if(ontologyTerm == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                    "OntologyTermUri query parameter was not present");
        }
        
        final URI ontologyTermUri = this.getDataHandler().createURI(ontologyTerm);
        
        final String ontologyUri = queryParams.getFirstValue("ontologyUri", true);
        final String ontologyVersionUri = queryParams.getFirstValue("ontologyVersionUri", true);
        final String recursiveString = queryParams.getFirstValue("recursive", true);
        
        Ontology nextOntology = null;
        
        if(ontologyVersionUri != null && ontologyUri != null)
        {
            nextOntology =
                    this.getDataHandler().getVersionedOntology(this.getDataHandler().createURI(ontologyUri),
                            this.getDataHandler().createURI(ontologyVersionUri));
        }
        
        if(nextOntology == null && ontologyUri != null)
        {
            nextOntology = this.getDataHandler().getLatestOntology(this.getDataHandler().createURI(ontologyUri));
        }
        
        boolean recursive = false;
        if(recursiveString != null)
        {
            recursive = Boolean.valueOf(recursiveString);
        }
        
        final ResourceCount results =
                this.getDataHandler().getAnnotationCountsByOntologyTermUri(ontologyTermUri, nextOntology, recursive);
        
        try
        {
            return RestletUtils.getRdfRepresentation(results, variant, this.getRequest().getResourceRef());
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final RDFHandlerException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
    }
}
