/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.openrdf.model.Literal;
import org.openrdf.model.Model;
import org.openrdf.model.Resource;
import org.openrdf.model.URI;
import org.openrdf.model.util.GraphUtil;
import org.openrdf.model.util.GraphUtilException;
import org.openrdf.model.vocabulary.RDFS;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFHandlerException;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.Get;
import org.restlet.resource.ResourceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasConstants;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class AuditLogResourceImpl extends AbstractOasResourceImpl
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private List<String> formatLogDetailsForHtml(final Model model)
    {
        final List<String> results = new ArrayList<String>();
        
        for(final Resource nextRecord : model.subjects())
        {
            try
            {
                final Literal date =
                        GraphUtil.getOptionalObjectLiteral(model, nextRecord, OasConstants.DC_TERMS_MODIFIED);
                final Literal message = GraphUtil.getOptionalObjectLiteral(model, nextRecord, RDFS.COMMENT);
                final Literal action =
                        GraphUtil.getOptionalObjectLiteral(model, nextRecord, OasConstants.URN_TAG_LOGACTION);
                final URI method =
                        GraphUtil.getOptionalObjectURI(model, nextRecord, OasConstants.URN_TAG_LOGOPERATION_VERB);
                final Literal user = GraphUtil.getOptionalObjectLiteral(model, nextRecord, OasConstants.FOAF_MAKER);
                final Literal reference =
                        GraphUtil.getOptionalObjectLiteral(model, nextRecord, OasConstants.URN_TAG_AUDITLOG_REF);
                final Literal ipaddress =
                        GraphUtil.getOptionalObjectLiteral(model, nextRecord, OasConstants.URN_TAG_USERSIPADDRESS);
                
                final StringBuilder nextLine = new StringBuilder(128);
                nextLine.append("<tr>\n<td>\n");
                if(date != null)
                {
                    nextLine.append(date.stringValue());
                }
                nextLine.append("</td>\n<td>\n");
                if(message != null)
                {
                    nextLine.append(message.stringValue());
                }
                nextLine.append("</td>\n<td>\n");
                if(action != null)
                {
                    nextLine.append(action.stringValue());
                }
                nextLine.append("</td>\n<td>\n");
                if(method != null)
                {
                    nextLine.append(method.stringValue());
                }
                nextLine.append("</td>\n<td>\n");
                if(user != null)
                {
                    nextLine.append(user.stringValue());
                }
                nextLine.append("</td>\n<td>\n");
                if(reference != null)
                {
                    nextLine.append(reference.stringValue());
                }
                nextLine.append("</td>\n<td>\n");
                if(ipaddress != null)
                {
                    nextLine.append(ipaddress.stringValue());
                }
                nextLine.append("</td>\n</tr>\n");
                results.add(nextLine.toString());
            }
            catch(final GraphUtilException re)
            {
                this.log.error(
                        "Failed to process audit log due to graph structure being unexpected for record={} message={}",
                        nextRecord.stringValue(), re.getMessage());
            }
        }
        
        return results;
        
    }
    
    @Get(":html")
    public Representation getAuditLogHtml(final Representation entity) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.AUDITLOG_GET);
        
        this.log.debug("in getAuditLogHtml");
        
        final String offsetString = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("offset", true);
        
        final String limitString = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("limit", true);
        
        int offset = 0;
        int limit = 100;
        
        if(offsetString != null)
        {
            offset = Integer.parseInt(offsetString);
            
            if(offset < 0)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid offset requested");
            }
        }
        
        if(limitString != null)
        {
            limit = Integer.parseInt(limitString);
            
            if(limit > 10000 || limit < 1)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid limit requested");
            }
        }
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_AUDITLOG, OasProps.DEF_TMPL_AUDITLOG));
        dataModel.put("pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_AUDITLOG_TITLE, OasProps.DEF_TMPL_AUDITLOG_TITLE));
        
        final Model model = this.getDataHandler().getLogDetails(offset, limit);
        dataModel.put("auditLog", this.formatLogDetailsForHtml(model));
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    @Get(":rdf|rj|json|ttl")
    public Representation getAuditLogRdf(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.AUDITLOG_GET);
        
        this.log.debug("in getAuditLogHtml");
        
        final String offsetString = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("offset", true);
        
        final String limitString = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("limit", true);
        
        int offset = 0;
        int limit = 100;
        
        if(offsetString != null)
        {
            offset = Integer.parseInt(offsetString);
            
            if(offset < 0)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid offset requested");
            }
        }
        
        if(limitString != null)
        {
            limit = Integer.parseInt(limitString);
            
            if(limit > 10000 || limit < 1)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid limit requested");
            }
        }
        
        final Model model = this.getDataHandler().getLogDetails(offset, limit);
        try
        {
            return RestletUtils.toRDFSerialisation(variant.getMediaType().getName(), model);
        }
        catch(RepositoryException | RDFHandlerException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Failed to generate output.", ex);
        }
    }
    
}
