/**
 * Package for implementations of the resource handlers in the OAS API
 */
package com.github.ansell.oas.webservice.impl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;

import org.openrdf.model.Model;
import org.openrdf.model.Resource;
import org.openrdf.model.URI;
import org.openrdf.model.impl.LinkedHashModel;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFFormat;
import org.openrdf.rio.RDFHandlerException;
import org.openrdf.rio.RDFParseException;
import org.openrdf.rio.RDFParser;
import org.openrdf.rio.Rio;
import org.openrdf.rio.helpers.StatementCollector;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasConstants;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.utils.UserUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.BulkFetchAnnotationResource;

/**
 * Fetches multiple annotations, based on URIs given in submitted RDF documents.
 * 
 * Provides responses in HTML, Talis RDF/JSON and RDF/XML.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class BulkFetchAnnotationResourceImpl extends AbstractOasResourceImpl implements BulkFetchAnnotationResource
{
    /**
     * Default constructor for Annotation Resource Implementation.
     */
    public BulkFetchAnnotationResourceImpl()
    {
        super();
    }
    
    private Collection<Annotation> getAnnotations() throws ResourceException
    {
        final String offsetString = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("offset", true);
        
        final String limitString = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("limit", true);
        
        int offset = 0;
        int limit = 100;
        
        if(offsetString != null)
        {
            offset = Integer.parseInt(offsetString);
            
            if(offset < 0)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid offset requested");
            }
        }
        
        if(limitString != null)
        {
            limit = Integer.parseInt(limitString);
            
            if(limit > 10000 || limit < 1)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Invalid limit requested");
            }
        }
        
        return this.getDataHandler().listAnnotations(offset, limit);
    }
    
    @Override
    public Representation getToHtml(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final Collection<Annotation> annotations = this.getAnnotations();
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST, OasProps.DEF_TMPL_ANNOTATIONS_LIST));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST_TITLE,
                        OasProps.DEF_TMPL_ANNOTATIONS_LIST_TITLE));
        
        dataModel.put("annotations", annotations);
        dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    public Representation getToRdf(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final Collection<Annotation> annotations = this.getAnnotations();
        
        try
        {
            return RestletUtils.getRdfRepresentation(annotations, variant.getMediaType(), this.getRequest()
                    .getResourceRef());
        }
        catch(final RepositoryException | RDFHandlerException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Failed to generate output.", ex);
        }
    }
    
    private Collection<Annotation> parseInput(final Representation entity) throws ResourceException
    {
        final Model input = new LinkedHashModel();
        final RDFFormat inputFormat = Rio.getParserFormatForMIMEType(entity.getMediaType().getName());
        
        if(inputFormat == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Did not recognise input format");
        }
        
        final RDFParser parser = Rio.createParser(inputFormat);
        parser.setRDFHandler(new StatementCollector(input));
        try
        {
            parser.parse(entity.getStream(), "");
        }
        catch(RDFParseException | RDFHandlerException | IOException e)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Failed to parse input");
        }
        
        final Collection<Annotation> annotations = new ArrayList<Annotation>();
        
        for(final Resource nextSubject : input.filter(null, RDF.TYPE, OasConstants.TAGGING_TAGGING).subjects())
        {
            if(nextSubject instanceof URI)
            {
                // Get the annotation based on the identifier URI and load it into memory
                final Annotation annotation = this.getDataHandler().getAnnotation((URI)nextSubject);
                
                if(annotation == null)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find annotation. uri="
                            + nextSubject.stringValue());
                }
                
                annotations.add(annotation);
            }
        }
        
        return annotations;
    }
    
    @Override
    public Representation postToHtml(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final Collection<Annotation> annotations = this.parseInput(entity);
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST, OasProps.DEF_TMPL_ANNOTATIONS_LIST));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST_TITLE,
                        OasProps.DEF_TMPL_ANNOTATIONS_LIST_TITLE));
        
        dataModel.put("annotations", annotations);
        dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    public Representation postToRdf(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final Collection<Annotation> annotations = this.parseInput(entity);
        
        try
        {
            return RestletUtils.getRdfRepresentation(annotations, variant.getMediaType(), this.getRequest()
                    .getResourceRef());
        }
        catch(final RepositoryException | RDFHandlerException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Failed to generate output.", ex);
        }
    }
}
