/**
 * Package for implementations of the resource handlers in the OAS API
 */
package com.github.ansell.oas.webservice.impl;

import java.util.Map;

import org.openrdf.model.URI;
import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryException;
import org.openrdf.repository.sail.SailRepository;
import org.openrdf.rio.RDFHandlerException;
import org.openrdf.sail.memory.MemoryStore;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.utils.SesameUtils;
import com.github.ansell.oas.utils.UserUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.BulkUpdateAnnotationResource;
import com.github.ansell.restletutils.RestletUtilUser;

/**
 * Updates multiple annotations, based on URIs given in submitted RDF documents.
 * 
 * Provides responses in HTML, Talis RDF/JSON and RDF/XML.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class BulkUpdateAnnotationResourceImpl extends AbstractOasResourceImpl implements BulkUpdateAnnotationResource
{
    /**
     * Default constructor for Annotation Resource Implementation.
     */
    public BulkUpdateAnnotationResourceImpl()
    {
        super();
    }
    
    @Override
    public Representation postRdfToHtml(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_CREATE);
        
        URI currentUserUri = null;
        
        if(this.getRequest().getClientInfo().isAuthenticated())
        {
            final RestletUtilUser findUser =
                    this.getOasApplication().getRealm()
                            .findUser(this.getRequest().getClientInfo().getUser().getIdentifier());
            
            if(findUser != null)
            {
                try
                {
                    currentUserUri =
                            this.getOasApplication().getRealm()
                                    .getUserUri(this.getRequest().getClientInfo().getUser().getIdentifier());
                }
                catch(final RepositoryException e)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find user for annotation");
                }
            }
        }
        
        final Map<URI, Annotation> storeAnnotations =
                SesameUtils.parseAndStoreAnnotations(entity, this.getDataHandler(), currentUserUri, this.getRequest()
                        .getRootRef().toString(false, false));
        
        // HTTP 201 Created
        this.getResponse().setStatus(Status.SUCCESS_CREATED);
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST, OasProps.DEF_TMPL_ANNOTATIONS_LIST));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST_TITLE,
                        OasProps.DEF_TMPL_ANNOTATIONS_LIST_TITLE));
        
        dataModel.put("annotations", storeAnnotations.values());
        dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                variant.getMediaType(), this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    public Representation postRdfToRdf(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_CREATE);
        
        URI currentUserUri = null;
        
        if(this.getRequest().getClientInfo().isAuthenticated())
        {
            final RestletUtilUser findUser =
                    this.getOasApplication().getRealm()
                            .findUser(this.getRequest().getClientInfo().getUser().getIdentifier());
            
            if(findUser != null)
            {
                try
                {
                    currentUserUri =
                            this.getOasApplication().getRealm()
                                    .getUserUri(this.getRequest().getClientInfo().getUser().getIdentifier());
                }
                catch(final RepositoryException e)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find user for annotation");
                }
            }
        }
        
        final Map<URI, Annotation> storeAnnotations =
                SesameUtils.parseAndStoreAnnotations(entity, this.getDataHandler(), currentUserUri, this.getRequest()
                        .getRootRef().toString(false, false));
        
        // HTTP 201 Created
        this.getResponse().setStatus(Status.SUCCESS_CREATED);
        
        try
        {
            final Repository nextRepository = new SailRepository(new MemoryStore());
            nextRepository.initialize();
            
            for(final URI nextAnnotationUri : storeAnnotations.keySet())
            {
                SesameUtils.toRDFRepository(storeAnnotations.get(nextAnnotationUri), nextRepository,
                        SesameUtils.ANNOTATION_MANAGEMENT_GRAPH);
            }
            
            return RestletUtils.toRDFSerialisation(variant.getMediaType().getName(), nextRepository, storeAnnotations
                    .keySet().toArray(new URI[0]));
        }
        catch(final RepositoryException ex)
        {
            // return HTTP Status.CLIENT_ERROR_BAD_REQUEST client error
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final RDFHandlerException ex)
        {
            // return HTTP Status.CLIENT_ERROR_BAD_REQUEST client error
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
    }
    
}
