/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.any23.mime.MIMEType;
import org.apache.any23.mime.TikaMIMETypeDetector;
import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.ext.fileupload.RestletFileUpload;
import org.restlet.representation.Representation;
import org.restlet.representation.StringRepresentation;
import org.restlet.resource.Get;
import org.restlet.resource.ResourceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.OntologyUtils;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.CreateOntologyResource;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class CreateOntologyResourceImpl extends AbstractOasResourceImpl implements CreateOntologyResource
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private final Path tempDirectory;
    
    /**
     * 
     */
    public CreateOntologyResourceImpl()
    {
        super();
        
        try
        {
            this.tempDirectory = Files.createTempDirectory("oas-ontologymanageruploads");
        }
        catch(final IOException e)
        {
            throw new RuntimeException("Could not create temporary directory", e);
        }
    }
    
    @Override
    public Representation createOntologyMultiPartByPost(final Representation entity) throws ResourceException
    {
        this.log.debug("in createOntologyMultiPartByPost");
        
        this.checkAuthentication(BaseOasActions.ONTOLOGY_CREATE);
        
        Representation rep = null;
        if(entity != null)
        {
            if(MediaType.MULTIPART_FORM_DATA.equals(entity.getMediaType(), true))
            {
                List<FileItem> items;
                Path file = null;
                String filename = null;
                String contentType = null;
                
                // 1: Create a factory for disk-based file items
                final DiskFileItemFactory factory = new DiskFileItemFactory(1000240, this.tempDirectory.toFile());
                
                // 2: Create a new file upload handler
                final RestletFileUpload upload = new RestletFileUpload(factory);
                final Map<String, String> props = new HashMap<String, String>();
                try
                {
                    // 3: Request is parsed by the handler which generates a list of FileItems
                    items = upload.parseRequest(this.getRequest());
                    
                    for(final FileItem fi : items)
                    {
                        final String name = fi.getName();
                        
                        if(name == null)
                        {
                            props.put(fi.getFieldName(), new String(fi.get(), StandardCharsets.UTF_8));
                        }
                        else
                        {
                            // FIXME: Strip everything up to the last . out of the filename so that
                            // the filename can be used for content type determination where
                            // possible.
                            // Note: These are Java-7 APIs
                            file = Files.createTempFile(this.tempDirectory, "ontologyupload-", name);
                            filename = name;
                            contentType = fi.getContentType();
                            // InputStream uploadedFileInputStream = fi.getInputStream();
                            try
                            {
                                fi.write(file.toFile());
                            }
                            catch(final IOException ioe)
                            {
                                throw ioe;
                            }
                            catch(final Exception e)
                            {
                                // avoid throwing a generic exception just because the apache
                                // commons library throws Exception
                                throw new IOException(e);
                            }
                        }
                    }
                    
                }
                catch(final IOException | FileUploadException e)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, e);
                }
                
                // this.log.error("props={}", props.toString());
                
                if(file == null || filename == null)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                            "Did not submit a valid file and filename");
                }
                else
                {
                    String ontologyUri;
                    
                    if(props.containsKey("ontologyUri"))
                    {
                        ontologyUri = props.get("ontologyUri");
                    }
                    else
                    {
                        throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Did not submit an ontologyUri");
                    }
                    
                    String ontologyVersionUri = null;
                    
                    if(props.containsKey("ontologyVersionUri"))
                    {
                        ontologyVersionUri = props.get("ontologyVersionUri");
                    }
                    
                    this.log.debug("ontology version URI: {}", ontologyVersionUri);
                    
                    if(ontologyVersionUri != null && ontologyVersionUri.trim().isEmpty())
                    {
                        this.log.debug("replacing empty version URI with null");
                        ontologyVersionUri = null;
                    }
                    
                    String baseUri = "";
                    
                    if(props.containsKey("baseUri"))
                    {
                        baseUri = props.get("baseUri");
                    }
                    
                    boolean deleteOldVersions = false;
                    
                    if(props.containsKey("deleteOldVersions"))
                    {
                        deleteOldVersions = Boolean.parseBoolean(props.get("deleteOldVersions"));
                        
                        this.log.debug("deleteOldVersions=" + deleteOldVersions);
                    }
                    
                    InputStream inputStream;
                    try
                    {
                        inputStream = new BufferedInputStream(Files.newInputStream(file, StandardOpenOption.READ));
                    }
                    catch(final IOException e)
                    {
                        throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "File IO error occurred", e);
                    }
                    
                    final MIMEType parsedContentType = MIMEType.parse(contentType);
                    
                    final MIMEType guessedMIMEType =
                            new TikaMIMETypeDetector().guessMIMEType(filename, inputStream, parsedContentType);
                    
                    final MediaType finalMimeType = MediaType.valueOf(guessedMIMEType.toString());
                    
                    OntologyUtils.putOntologyRdfInternal(ontologyUri, ontologyVersionUri, baseUri, deleteOldVersions,
                            this.getResponse(), this.getDataHandler(), inputStream, finalMimeType, this.getRequest()
                                    .getResourceRef());
                    
                    this.redirectToPath(this.getPropertyUtil().get(OasProps.PROP_ONTOLOGY_MANAGER_PATH,
                            OasProps.DEF_ONTOLOGY_MANAGER_PATH));
                }
            }
            else
            {
                // other format != multipart form data
                this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
                rep = new StringRepresentation("Multipart/form-data required", MediaType.TEXT_PLAIN);
            }
        }
        else
        {
            // POST request with no entity.
            this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
            rep = new StringRepresentation("Error", MediaType.TEXT_PLAIN);
        }
        
        return rep;
        
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see
     * com.github.ansell.oas.webservice.OntologyResource#putOntologyRdf(org.restlet.representation
     * .Representation)
     */
    @Override
    public Representation createOntologyRdfByPost(final Representation entity) throws ResourceException
    {
        this.log.debug("in createOntologyRdfByPost");
        
        this.checkAuthentication(BaseOasActions.ONTOLOGY_CREATE);
        
        if(entity == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Ontology cannot be null");
        }
        
        try
        {
            return OntologyUtils.putOntologyRdfInternal(this.getRequest(), this.getResponse(), this.getDataHandler(),
                    entity.getStream(), entity.getMediaType());
        }
        catch(final IOException e)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, e);
        }
    }
    
    @Override
    public Representation createOntologyRdfByPut(final Representation entity) throws ResourceException
    {
        this.log.debug("in createOntologyRdfByPut");
        
        this.checkAuthentication(BaseOasActions.ONTOLOGY_CREATE);
        
        if(entity == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Ontology cannot be null");
        }
        
        try
        {
            return OntologyUtils.putOntologyRdfInternal(this.getRequest(), this.getResponse(), this.getDataHandler(),
                    entity.getStream(), entity.getMediaType());
        }
        catch(final IOException e)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, e);
        }
    }
    
    @Override
    @Get("html")
    public Representation getUploadPageHtml(final Representation entity) throws ResourceException
    {
        this.log.debug("in createOntologyRdfByPut");
        
        this.checkAuthentication(BaseOasActions.ONTOLOGY_CREATE);
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put(
                "contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_UPLOAD,
                        OasProps.DEF_TMPL_ONTOLOGYMANAGER_UPLOAD));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_TITLE,
                        OasProps.DEF_TMPL_ONTOLOGYMANAGER_TITLE));
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
}
