/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.openrdf.model.URI;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.ext.fileupload.RestletFileUpload;
import org.restlet.representation.Representation;
import org.restlet.representation.StringRepresentation;
import org.restlet.resource.Get;
import org.restlet.resource.ResourceException;
import org.restlet.security.User;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.CreateUserResource;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class CreateUserResourceImpl extends AbstractOasResourceImpl implements CreateUserResource
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private final Path tempDirectory;
    
    /**
     * 
     */
    public CreateUserResourceImpl()
    {
        super();
        
        try
        {
            this.tempDirectory = Files.createTempDirectory("oas-createuseruploads");
        }
        catch(final IOException e)
        {
            throw new RuntimeException("Could not create temporary directory", e);
        }
    }
    
    @Override
    @Get("html")
    public Representation getCreateUserHtml(final Representation entity) throws ResourceException
    {
        this.log.debug("in getCreateUserHtml");
        
        this.checkAuthentication(BaseOasActions.USER_CREATE);
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_USER_CREATE, OasProps.DEF_TMPL_USER_CREATE));
        dataModel.put("pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_USER_CREATE_TITLE, OasProps.DEF_TMPL_USER_CREATE_TITLE));
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    public Representation postCreateUserMultipart(final Representation entity) throws ResourceException
    {
        this.log.debug("in postCreateUserMultipart");
        
        this.checkAuthentication(BaseOasActions.USER_CREATE);
        
        Representation rep = null;
        if(entity != null)
        {
            if(MediaType.MULTIPART_FORM_DATA.equals(entity.getMediaType(), true))
            {
                List<FileItem> items;
                Path file = null;
                String filename = null;
                String contentType = null;
                
                // 1: Create a factory for disk-based file items
                final DiskFileItemFactory factory = new DiskFileItemFactory(1000240, this.tempDirectory.toFile());
                
                // 2: Create a new file upload handler
                final RestletFileUpload upload = new RestletFileUpload(factory);
                final Map<String, String> props = new HashMap<String, String>();
                try
                {
                    // 3: Request is parsed by the handler which generates a list of FileItems
                    items = upload.parseRequest(this.getRequest());
                    
                    for(final FileItem fi : items)
                    {
                        final String name = fi.getName();
                        
                        if(name == null)
                        {
                            props.put(fi.getFieldName(), new String(fi.get(), StandardCharsets.UTF_8));
                        }
                        else
                        {
                            // FIXME: Strip everything up to the last . out of the filename so that
                            // the filename can be used for content type determination where
                            // possible.
                            // Note: These are Java-7 APIs
                            file = Files.createTempFile(this.tempDirectory, "ontologyupload-", name);
                            filename = name;
                            contentType = fi.getContentType();
                            // InputStream uploadedFileInputStream = fi.getInputStream();
                            try
                            {
                                fi.write(file.toFile());
                            }
                            catch(final IOException ioe)
                            {
                                throw ioe;
                            }
                            catch(final Exception e)
                            {
                                // avoid throwing a generic exception just because the apache
                                // commons library throws Exception
                                throw new IOException(e);
                            }
                        }
                    }
                    
                }
                catch(final IOException | FileUploadException e)
                {
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, e);
                }
                
                // this.log.debug("props={}", props.toString());
                
                String userName = null;
                
                if(props.containsKey("username"))
                {
                    userName = props.get("username");
                }
                else
                {
                    this.log.error("Did not submit a username");
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Did not submit a username");
                }
                
                String password = null;
                
                if(props.containsKey("password"))
                {
                    password = props.get("password");
                }
                else
                {
                    this.log.error("Did not submit a password");
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Did not submit a password");
                }
                
                String email = "";
                
                if(props.containsKey("email"))
                {
                    email = props.get("email");
                }
                else
                {
                    this.log.error("Did not submit an email address");
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Did not submit an email address");
                }
                
                String lastName = "";
                
                if(props.containsKey("lastName"))
                {
                    lastName = props.get("lastName");
                }
                
                String firstName = "";
                
                if(props.containsKey("firstName"))
                {
                    firstName = props.get("firstName");
                }
                
                final User nextUser = new User(userName, password, firstName, lastName, email);
                
                // TODO: Redirect user based on the URI that was returned
                final URI addUser = this.getOasApplication().getRealm().addUser(nextUser);
                
                this.redirectToPath(this.getPropertyUtil().get(OasProps.PROP_GET_USER_PATH, OasProps.DEF_GET_USER_PATH));
            }
            else
            {
                // other format != multipart form data
                this.getResponse().setStatus(Status.CLIENT_ERROR_UNSUPPORTED_MEDIA_TYPE);
                rep = new StringRepresentation("Multipart/form-data required", MediaType.TEXT_PLAIN);
            }
        }
        else
        {
            this.log.error("No content was sent with request");
            // POST request with no entity.
            this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
            rep = new StringRepresentation("Error: No content was sent with request", MediaType.TEXT_PLAIN);
        }
        
        return rep;
        
    }
    
}
