/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;

import org.openrdf.model.URI;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.Get;
import org.restlet.resource.ResourceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.OntologyUtils;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.DeleteOntologyResource;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class DeleteOntologyResourceImpl extends AbstractOasResourceImpl implements DeleteOntologyResource
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private final Path tempDirectory;
    
    /**
     * 
     */
    public DeleteOntologyResourceImpl()
    {
        super();
        
        try
        {
            this.tempDirectory = Files.createTempDirectory("oas-ontologymanagerdeletions");
        }
        catch(final IOException e)
        {
            throw new RuntimeException("Could not create temporary directory", e);
        }
    }
    
    @Override
    public void deleteOntologyByDelete(final Representation entity, final Variant variant) throws ResourceException
    {
        this.log.debug("deleteOntology: deleteOntologyByDelete");
        
        this.checkAuthentication(BaseOasActions.ONTOLOGY_DELETE);
        
        OntologyUtils.deleteOntologyInternalQueryForm(this.getRequest(), this.getResponse(), this.getDataHandler());
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see com.github.ansell.oas.webservice.OntologyResource#deleteOntology()
     */
    @Override
    public void deleteOntologyByPost(final Representation entity, final Variant variant) throws ResourceException
    {
        this.log.debug("deleteOntology: deleteOntologyByPost");
        
        this.checkAuthentication(BaseOasActions.ONTOLOGY_DELETE);
        
        OntologyUtils.deleteOntologyInternalMultipartFormPost(this.getRequest(), this.getResponse(),
                this.getDataHandler(), this.tempDirectory, this.getPropertyUtil());
        
    }
    
    @Override
    @Get("html")
    public Representation getDeletePageHtml(final Representation entity) throws ResourceException
    {
        this.log.debug("deleteOntology: getDeletePageHtml");
        
        this.checkAuthentication(BaseOasActions.ONTOLOGY_DELETE);
        
        final String ontologyUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        this.log.debug("ontologyUriString={}", ontologyUriString);
        
        if(ontologyUriString == null)
        {
            this.log.error("ontologyUri parameter was not present");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "ontologyUri parameter must not be empty");
        }
        
        final URI ontologyUri = this.getDataHandler().createURI(ontologyUriString);
        
        final String ontologyVersionUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyVersionUri", true);
        
        if(ontologyVersionUriString == null)
        {
            this.log.error("ontologyVersionUriString parameter was not present");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                    "ontologyVersionUriString parameter must not be empty");
        }
        
        final URI ontologyVersionUri = this.getDataHandler().createURI(ontologyVersionUriString);
        
        final Ontology managedOntology = this.getDataHandler().getVersionedOntology(ontologyUri, ontologyVersionUri);
        
        if(managedOntology == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find the specified ontology.");
        }
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put(
                "contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_DELETE,
                        OasProps.DEF_TMPL_ONTOLOGYMANAGER_DELETE));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_TITLE,
                        OasProps.DEF_TMPL_ONTOLOGYMANAGER_TITLE));
        
        dataModel.put("ontologyToDelete", managedOntology);
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
}
