/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;

import org.openrdf.model.URI;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.ext.html.FormDataSet;
import org.restlet.representation.Representation;
import org.restlet.representation.StringRepresentation;
import org.restlet.representation.Variant;
import org.restlet.resource.Get;
import org.restlet.resource.ResourceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.DeleteUserResource;
import com.github.ansell.restletutils.RestletUtilUser;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class DeleteUserResourceImpl extends AbstractOasResourceImpl implements DeleteUserResource
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private final Path tempDirectory;
    
    /**
     * 
     */
    public DeleteUserResourceImpl()
    {
        super();
        
        try
        {
            this.tempDirectory = Files.createTempDirectory("oas-deleteuseruploads");
        }
        catch(final IOException e)
        {
            throw new RuntimeException("Could not create temporary directory", e);
        }
    }
    
    @Override
    public void deleteUserByPost(final Representation entity, final Variant variant) throws ResourceException
    {
        this.log.debug("in deleteUserByPost");
        
        this.checkAuthentication(BaseOasActions.USER_DELETE);
        
        Representation rep = null;
        if(entity != null)
        {
            final FormDataSet postedForm = new FormDataSet(entity);
            this.log.trace("postedForm={}", postedForm);
            this.log.trace("postedForm.getEntries={}", postedForm.getEntries());
            final String userName = postedForm.getEntries().getFirstValue("username", true);
            
            // FIXME: Integrate this code with UserResourceImpl.deleteUser
            if(userName == null)
            {
                this.log.error("Did not submit a username");
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Did not submit a username");
            }
            
            final RestletUtilUser findUser = this.getOasApplication().getRealm().findUser(userName);
            
            if(findUser == null)
            {
                this.log.error("Could not delete user as they were not found: {}", userName);
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST);
            }
            else if(findUser.getIdentifier().equals(this.getRequest().getClientInfo().getUser().getIdentifier()))
            {
                this.log.error("Cannot delete ones own user", userName);
                throw new ResourceException(Status.CLIENT_ERROR_LOCKED, "Cannot delete the currently logged in user");
            }
            
            // TODO: Redirect user based on the URI that was returned
            // Returns the URI of the deleted user, or null if no user was found
            final URI deleteUser = this.getOasApplication().getRealm().deleteUser(findUser);
            
            this.redirectToPath(this.getPropertyUtil().get(OasProps.PROP_GET_USER_PATH, OasProps.DEF_GET_USER_PATH));
        }
        else
        {
            this.log.error("No content was sent with request");
            // POST request with no entity.
            this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
            rep = new StringRepresentation("Error: No content was sent with request", MediaType.TEXT_PLAIN);
        }
        
    }
    
    @Override
    @Get("html")
    public Representation getDeletePageHtml(final Representation entity) throws ResourceException
    {
        this.log.debug("in getDeletePageHtml");
        
        this.checkAuthentication(BaseOasActions.USER_DELETE);
        
        // Include parameter for the username in the query entity
        final String usernameString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("username", true);
        
        if(usernameString == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Did not submit username for deletion");
        }
        
        final RestletUtilUser findUser = this.getOasApplication().getRealm().findUser(usernameString);
        
        if(findUser == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "User does not exist");
        }
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_USER_DELETE, OasProps.DEF_TMPL_USER_DELETE));
        dataModel.put("pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_USER_DELETE_TITLE, OasProps.DEF_TMPL_USER_DELETE_TITLE));
        
        dataModel.put("deleteUsername", usernameString);
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
}
