/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.restlet.data.Form;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.Get;
import org.restlet.resource.ResourceException;
import org.restlet.security.Role;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.EditUserRolesResource;
import com.github.ansell.restletutils.RestletUtilRoles;
import com.github.ansell.restletutils.RestletUtilSesameRealm;
import com.github.ansell.restletutils.RestletUtilUser;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class EditUserRolesResourceImpl extends AbstractOasResourceImpl implements EditUserRolesResource
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private final Path tempDirectory;
    
    /**
     * 
     */
    public EditUserRolesResourceImpl()
    {
        super();
        
        try
        {
            this.tempDirectory = Files.createTempDirectory("oas-createuseruploads");
        }
        catch(final IOException e)
        {
            throw new RuntimeException("Could not create temporary directory", e);
        }
    }
    
    @Override
    public void editUserRolesByPost(final Representation entity, final Variant variant) throws ResourceException
    {
        this.log.debug("in editUserRolesByPost");
        
        this.checkAuthentication(BaseOasActions.ROLE_EDIT);
        
        final Representation rep = null;
        
        String username = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("username", true);
        
        final Form data = new Form(this.getRequest().getEntity());
        
        final List<String> submittedRoles = new ArrayList<String>();
        
        submittedRoles.addAll(Arrays.asList(this.getRequest().getResourceRef().getQueryAsForm()
                .getValuesArray("activeRoles", true)));
        
        this.log.debug("formdata: {}", data);
        for(final String nextEntry : data.getValuesMap().keySet())
        {
            this.log.trace("nextEntry: {}", nextEntry);
            if(nextEntry.equalsIgnoreCase("activeRoles"))
            {
                submittedRoles.addAll(Arrays.asList(data.getValuesArray("activeRoles", true)));
            }
            else if(nextEntry.equalsIgnoreCase("username"))
            {
                username = data.getFirstValue("username", true);
            }
        }
        
        if(username == null)
        {
            this.log.error("Username was not submitted");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Username was not submitted");
        }
        
        final RestletUtilSesameRealm realm = this.getOasApplication().getRealm();
        
        final RestletUtilUser findUser = realm.findUser(username);
        
        if(findUser == null)
        {
            this.log.error("Could not find the user");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find the user");
        }
        
        final Set<Role> currentRoles = realm.findRoles(findUser);
        
        this.log.debug("current roles: {}", currentRoles);
        
        this.log.debug("next active roles: {}", submittedRoles);
        
        // log.info("inactive roles: {}", inactiveRoles);
        
        // For each existing role, if the role was not in the submitted set, unmap it in the
        // realm
        for(final Role nextCurrentRole : currentRoles)
        {
            boolean found = false;
            for(final String nextSubmittedRole : submittedRoles)
            {
                if(nextCurrentRole.getName().equalsIgnoreCase(nextSubmittedRole))
                {
                    found = true;
                    break;
                }
            }
            
            if(!found)
            {
                realm.unmap(findUser, nextCurrentRole);
            }
        }
        
        // for each of the submitted roles, if the role was not in the existing set, map it in
        // the realm
        for(final String nextSubmittedRole : submittedRoles)
        {
            if(nextSubmittedRole == null || nextSubmittedRole.trim().isEmpty())
            {
                // skip null or empty strings in submitted roles
                continue;
            }
            boolean found = false;
            for(final Role nextCurrentRole : currentRoles)
            {
                if(nextCurrentRole.getName().equalsIgnoreCase(nextSubmittedRole))
                {
                    found = true;
                    break;
                }
            }
            
            if(!found)
            {
                final List<Role> assignableRoles = RestletUtilRoles.getAssignableRoles();
                
                Role matchingRole = null;
                
                for(final Role nextAssignableRole : assignableRoles)
                {
                    if(nextAssignableRole.getName().equalsIgnoreCase(nextSubmittedRole))
                    {
                        matchingRole = nextAssignableRole;
                        realm.map(findUser, nextAssignableRole);
                        break;
                    }
                }
                
                if(matchingRole == null)
                {
                    this.log.error("Could not find a role with the name: " + nextSubmittedRole);
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                            "Could not find a role with the name: " + nextSubmittedRole);
                }
            }
        }
        
        final Set<Role> afterRoles = realm.findRoles(findUser);
        
        if(currentRoles.contains(RestletUtilRoles.ADMIN.getRole())
                && !afterRoles.contains(RestletUtilRoles.ADMIN.getRole()))
        {
            boolean foundAdmin = false;
            // Verify that there is still at least one user who has the admin role, or add the admin
            // role back to this user
            for(final RestletUtilUser nextUser : realm.getUsers())
            {
                if(!nextUser.getIdentifier().equals(findUser.getIdentifier()))
                {
                    if(realm.findRoles(nextUser).contains(RestletUtilRoles.ADMIN.getRole()))
                    {
                        foundAdmin = true;
                    }
                }
            }
            
            if(!foundAdmin)
            {
                // https://github.com/ansell/oas/issues/50
                this.log.warn(
                        "Remapping user to admin role to prevent situation where there are no administrators: original details(user={} currentRoles={} afterRoles={})",
                        findUser, currentRoles, afterRoles);
                realm.map(findUser, RestletUtilRoles.ADMIN.getRole());
            }
        }
        
        this.log.debug("afterRoles: {}", afterRoles);
        
        this.redirectToPath(this.getPropertyUtil().get(OasProps.PROP_GET_USER_PATH, OasProps.DEF_GET_USER_PATH)
                + "?username=" + username);
    }
    
    @Override
    @Get("html")
    public Representation getEditUserRolesPageHtml(final Representation entity) throws ResourceException
    {
        this.log.debug("in getEditUserRolesPageHtml");
        
        this.checkAuthentication(BaseOasActions.ROLE_EDIT);
        
        // TODO: Include parameter for the username in the query entity
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ROLE_EDIT, OasProps.DEF_TMPL_ROLE_EDIT));
        dataModel.put("pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ROLE_EDIT_TITLE, OasProps.DEF_TMPL_ROLE_EDIT_TITLE));
        
        final List<Role> possibleRoles = RestletUtilRoles.getAssignableRoles();
        dataModel.put("possibleRoles", possibleRoles);
        this.log.debug("possibleRoles: {}", possibleRoles);
        
        String editRoleUsername = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("username");
        if(editRoleUsername == null)
        {
            this.log.info("Username was not submitted, editing roles for current user");
            editRoleUsername = this.getClientInfo().getUser().getIdentifier();
        }
        
        final RestletUtilSesameRealm realm = this.getOasApplication().getRealm();
        final RestletUtilUser findUser = realm.findUser(editRoleUsername);
        
        if(findUser == null)
        {
            this.log.error("Could not find the user: editRoleUsername={}", editRoleUsername);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find the user: " + editRoleUsername);
        }
        
        dataModel.put("editRoleUser", findUser);
        
        final Set<Role> existingRoles = realm.findRoles(findUser);
        dataModel.put("existingRoles", existingRoles);
        
        this.log.debug("existingRoles: {}", existingRoles);
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
}
