package com.github.ansell.oas.webservice.impl;

import org.restlet.Context;
import org.restlet.Request;
import org.restlet.Response;
import org.restlet.Restlet;
import org.restlet.data.CharacterSet;
import org.restlet.data.MediaType;
import org.restlet.data.Parameter;
import org.restlet.data.Protocol;
import org.restlet.routing.Router;
import org.restlet.security.ChallengeAuthenticator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.data.DataHandler;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.webservice.OasAction;
import com.github.ansell.oas.webservice.OasWebServiceApplication;
import com.github.ansell.propertyutil.PropertyUtil;
import com.github.ansell.restletutils.CrossOriginResourceSharingFilter;
import com.github.ansell.restletutils.RestletUtilMediaType;
import com.github.ansell.restletutils.RestletUtilSesameRealm;

import freemarker.template.Configuration;

/**
 * This class handles all requests from clients to the OAS Web Service.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class OasWebServiceApplicationImpl extends OasWebServiceApplication
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    /**
     * The Freemarker template configuration.
     */
    private volatile Configuration freemarkerConfiguration;
    private volatile DataHandler dataHandler;
    private volatile ChallengeAuthenticator auth;
    private volatile RestletUtilSesameRealm realm;
    
    private volatile PropertyUtil propertyUtil;
    
    /**
     * Default Constructor.
     * 
     * Adds the necessary file protocols and sets up the template location.
     */
    public OasWebServiceApplicationImpl(Context context)
    {
        super(context);
        
        // List of protocols required by the application
        this.getConnectorService().getClientProtocols().add(Protocol.HTTP);
        this.getConnectorService().getClientProtocols().add(Protocol.CLAP);
        
        this.getMetadataService().setDefaultCharacterSet(CharacterSet.UTF_8);
        
        // Define extensions for RDF and Javascript
        // These extensions are also used to identify mediatypes in services
        // For example: @Get("owl") will not be processed without the declaration below
        this.getMetadataService().addExtension("rdf", MediaType.APPLICATION_RDF_XML, true);
        this.getMetadataService().addExtension("rj", RestletUtilMediaType.APPLICATION_RDF_JSON, true);
        this.getMetadataService().addExtension("owl", MediaType.APPLICATION_RDF_XML, false);
        this.getMetadataService().addExtension("json", MediaType.APPLICATION_JSON, true);
        this.getMetadataService().addExtension("ttl", MediaType.APPLICATION_RDF_TURTLE, true);
        this.getMetadataService().addExtension("n3", MediaType.TEXT_RDF_N3, true);
        this.getMetadataService().addExtension("nt", MediaType.TEXT_RDF_NTRIPLES, true);
        this.getMetadataService().addExtension("nq",
                MediaType.register("text/nquads", "The NQuads extension to the NTriples RDF serialisation"), true);
        
        this.getMetadataService().addExtension("js", MediaType.TEXT_JAVASCRIPT, true);
        this.getMetadataService().addExtension("css", MediaType.TEXT_CSS, true);
        
        this.getMetadataService().addExtension("multipart", MediaType.MULTIPART_FORM_DATA, true);
        this.getMetadataService().addExtension("form", MediaType.APPLICATION_WWW_FORM, false);
        
        // Automagically tunnel client preferences for extensions through the
        // tunnel
        this.getTunnelService().setExtensionsTunnel(true);
    }
    
    public OasWebServiceApplicationImpl(final Context context, final String propertyFilePrefix)
    {
        this(context);
        
        this.propertyUtil = new PropertyUtil(propertyFilePrefix);
    }
    
    @Override
    public boolean authenticate(final OasAction authenticationScope, final Request request, final Response response)
    {
        if(authenticationScope.isAuthRequired() && !request.getClientInfo().isAuthenticated())
        {
            if(this.getAuthenticator() == null)
            {
                throw new RuntimeException("Could not find authentication method");
            }
            
            // add challenges to the response and set the status to HTTP 401 Unauthorized
            this.getAuthenticator().challenge(response, false);
            
            // Return false after the challenge and HTTP 401 response have been added to the
            // response
            return false;
        }
        else if(authenticationScope.isAuthRequired() && request.getClientInfo().isAuthenticated()
                && authenticationScope.isRoleRequired()
                && !authenticationScope.matchesForRoles(request.getClientInfo().getRoles()))
        {
            this.log.error("Authenticated user does not have enough privileges to execute the given action");
            
            // FIXME: Implement auditing here
            // this.getDataHandler().addLogDetailsForRequest(message, referenceUri,
            // authenticationScope, get, currentUser, currentRole);
            
            return false;
        }
        
        if(request.getClientInfo().isAuthenticated() && request.getClientInfo().getRoles().isEmpty())
        {
            this.log.warn("Authenticated user did not have any roles: user={}", request.getClientInfo().getUser());
        }
        
        return true;
    }
    
    /**
     * Create the necessary connections between the application and its handlers.
     */
    @Override
    public Restlet createInboundRoot()
    {
        final CrossOriginResourceSharingFilter corsFilter = new CrossOriginResourceSharingFilter();
        
        final ChallengeAuthenticator authenticator = this.getAuthenticator();
        
        if(authenticator == null)
        {
            throw new RuntimeException("Could not find authentication method");
        }
        
        final Router router = new Router(this.getContext());
        
        // Add a route for Annotations resource for creating a new annotation without previously
        // knowing its ID
        final String annotationCreatePath =
                this.getPropertyUtil().get(OasProps.PROP_ANNOTATION_CREATE_PATH, OasProps.DEF_ANNOTATIONS_CREATE_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotation create service to path=" + annotationCreatePath);
        }
        router.attach(annotationCreatePath, BulkUpdateAnnotationResourceImpl.class);
        
        // Add a route for fetching multiple annotations
        final String bulkFetchAnnotationsPath =
                this.getPropertyUtil().get(OasProps.PROP_BULK_FETCH_ANNOTATION_PATH,
                        OasProps.DEF_BULK_FETCH_ANNOTATION_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotation bulk fetch service to path=" + bulkFetchAnnotationsPath);
        }
        router.attach(bulkFetchAnnotationsPath, BulkFetchAnnotationResourceImpl.class);
        
        // Add a route for deleting multiple annotations
        final String bulkDeleteAnnotationsPath =
                this.getPropertyUtil().get(OasProps.PROP_BULK_DELETE_ANNOTATION_PATH,
                        OasProps.DEF_BULK_DELETE_ANNOTATION_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotation bulk delete service to path=" + bulkDeleteAnnotationsPath);
        }
        router.attach(bulkDeleteAnnotationsPath, BulkDeleteAnnotationResourceImpl.class);
        
        // Add a route for getting, updating and deleting Annotations by ID
        final String annotationByIdPath =
                this.getPropertyUtil().get(OasProps.PROP_ANNOTATION_PATH, OasProps.DEF_ANNOTATION_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotation by id service to path=" + annotationByIdPath);
        }
        router.attach(annotationByIdPath, AnnotationByIdResourceImpl.class);
        
        // Add a route for getting a list of annotations based on the annotated object type
        final String annotationsByObjectTypePath =
                this.getPropertyUtil().get(OasProps.PROP_ANNOTATIONS_BY_OBJECT_TYPE_PATH,
                        OasProps.DEF_ANNOTATIONS_BY_OBJECT_TYPE_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotations by object type service to path=" + annotationsByObjectTypePath);
        }
        router.attach(annotationsByObjectTypePath, AnnotationsByAnnotatedObjectTypeResourceImpl.class);
        
        // Add a route for getting a list of annotations based on the annotated object uri
        final String annotationsByObjectUriPath =
                this.getPropertyUtil().get(OasProps.PROP_ANNOTATIONS_BY_OBJECT_URI_PATH,
                        OasProps.DEF_ANNOTATIONS_BY_OBJECT_URI_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotations by object uri service to path=" + annotationsByObjectUriPath);
        }
        router.attach(annotationsByObjectUriPath, AnnotationsByAnnotatedObjectUriResourceImpl.class);
        
        // Add a route for getting a list of annotations based on the ontology term uri
        final String annotationsByOntologyTermPath =
                this.getPropertyUtil().get(OasProps.PROP_ANNOTATIONS_BY_ONTOLOGY_TERM_URI_PATH,
                        OasProps.DEF_ANNOTATIONS_BY_ONTOLOGY_TERM_URI_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotations by ontology term uri service to path="
                    + annotationsByOntologyTermPath);
        }
        router.attach(annotationsByOntologyTermPath, AnnotationsByOntologyTermUriResourceImpl.class);
        
        // Add a route for getting a list of annotations based on the ontology term uri
        final String annotationsCountByOntologyTermUriPath =
                this.getPropertyUtil().get(OasProps.PROP_ANNOTATION_COUNTS_BY_ONTOLOGY_TERM_URI_PATH,
                        OasProps.DEF_ANNOTATION_COUNTS_BY_ONTOLOGY_TERM_URI_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotations count by ontology term uri service to path="
                    + annotationsCountByOntologyTermUriPath);
        }
        router.attach(annotationsCountByOntologyTermUriPath, AnnotationCountsByOntologyTermUriResourceImpl.class);
        
        // Add a route for getting a list of annotations based on the annotated object URI
        final String annotationsCountByAnnotatedObjectUriPath =
                this.getPropertyUtil().get(OasProps.PROP_ANNOTATION_COUNTS_BY_OBJECT_URI_PATH,
                        OasProps.DEF_ANNOTATION_COUNTS_BY_OBJECT_URI_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching annotations count by annotated object uri service to path="
                    + annotationsCountByAnnotatedObjectUriPath);
        }
        router.attach(annotationsCountByAnnotatedObjectUriPath, AnnotationCountsByAnnotatedObjectUriResourceImpl.class);
        
        // Add a route for uploading ontologies using HTTP POST
        final String ontologyManagerUploadPath =
                this.getPropertyUtil().get(OasProps.PROP_ONTOLOGY_MANAGER_UPLOAD_PATH,
                        OasProps.DEF_ONTOLOGY_MANAGER_UPLOAD_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching HTTP POST upload ontology resource service to path=" + ontologyManagerUploadPath);
        }
        router.attach(ontologyManagerUploadPath, CreateOntologyResourceImpl.class);
        
        // Add a route for deleting ontologies using HTTP POST
        final String ontologyManagerDeletePath =
                this.getPropertyUtil().get(OasProps.PROP_ONTOLOGY_MANAGER_DELETE_PATH,
                        OasProps.DEF_ONTOLOGY_MANAGER_DELETE_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching HTTP POST delete ontology resource service to path=" + ontologyManagerDeletePath);
        }
        router.attach(ontologyManagerDeletePath, DeleteOntologyResourceImpl.class);
        
        // Add a route for Creating ontologies using HTTP PUT, fetching using HTTP GET, and deleting
        // using HTTP DELETE
        final String ontologyManager =
                this.getPropertyUtil().get(OasProps.PROP_ONTOLOGY_MANAGER_PATH, OasProps.DEF_ONTOLOGY_MANAGER_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching restful ontology resource service to path=" + ontologyManager);
        }
        router.attach(ontologyManager, OntologyManagerResourceImpl.class);
        
        final String setCurrentVersion =
                this.getPropertyUtil().get(OasProps.PROP_SET_CURRENT_VERSION_PATH,
                        OasProps.DEF_SET_CURRENT_VERSION_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching set current ontology version service to path=" + setCurrentVersion);
        }
        router.attach(setCurrentVersion, SetCurrentOntologyResourceImpl.class);
        
        // Add a route for browsing ontologies using HTTP GET
        final String ontologyBrowser =
                this.getPropertyUtil().get(OasProps.PROP_ONTOLOGY_BROWSER_PATH, OasProps.DEF_ONTOLOGY_BROWSER_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching restful ontology browser service to path=" + ontologyBrowser);
        }
        router.attach(ontologyBrowser, OntologyBrowserResourceImpl.class);
        
        // Add a route for searching annotations
        final String searchAnnotations =
                this.getPropertyUtil().get(OasProps.PROP_SEARCH_ANNOTATION_PATH, OasProps.DEF_SEARCH_ANNOTATION_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching search annotations service to path=" + searchAnnotations);
        }
        router.attach(searchAnnotations, SearchAnnotationResourceImpl.class);
        
        // Add a route for searching ontologies
        final String searchOntologies =
                this.getPropertyUtil().get(OasProps.PROP_SEARCH_ONTOLOGIES_PATH, OasProps.DEF_SEARCH_ONTOLOGIES_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching search ontologies service to path=" + searchOntologies);
        }
        router.attach(searchOntologies, SearchOntologyResourceImpl.class);
        
        // Add a route for searching ontology labels
        final String searchOntologyLabels =
                this.getPropertyUtil().get(OasProps.PROP_SEARCH_ONTOLOGY_LABELS_PATH,
                        OasProps.DEF_SEARCH_ONTOLOGY_LABELS_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching search labels of ontologies service to path=" + searchOntologyLabels);
        }
        router.attach(searchOntologyLabels, SearchOntologyLabelsResourceImpl.class);
        
        // Add a route for Login form. Login service is handled by the authenticator
        final String login = this.getPropertyUtil().get(OasProps.PROP_LOGIN_FORM_PATH, OasProps.DEF_LOGIN_FORM_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching login service to path=" + login);
        }
        // NOTE: This only displays the login form. All HTTP POST requests to the login path should
        // be handled by the Authenticator
        router.attach(login, CookieLoginResourceImpl.class);
        
        // Add a route for Create User form and to handle POST requests
        final String createUser =
                this.getPropertyUtil().get(OasProps.PROP_CREATE_USER_PATH, OasProps.DEF_CREATE_USER_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching create user service to path=" + createUser);
        }
        router.attach(createUser, CreateUserResourceImpl.class);
        
        // Add a route for Delete User form and to handle POST requests
        final String deleteUser =
                this.getPropertyUtil().get(OasProps.PROP_DELETE_USER_PATH, OasProps.DEF_DELETE_USER_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching delete user service to path=" + deleteUser);
        }
        router.attach(deleteUser, DeleteUserResourceImpl.class);
        
        // Add a route for Update User form and to handle POST requests
        final String updateUser =
                this.getPropertyUtil().get(OasProps.PROP_UPDATE_USER_PATH, OasProps.DEF_UPDATE_USER_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching update user service to path=" + updateUser);
        }
        router.attach(updateUser, UpdateUserResourceImpl.class);
        
        // Add a route for Edit User Roles form and to handle POST requests
        final String editUserRoles =
                this.getPropertyUtil().get(OasProps.PROP_EDIT_USER_ROLES_PATH, OasProps.DEF_EDIT_USER_ROLES_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching edit user roles service to path=" + editUserRoles);
        }
        router.attach(editUserRoles, EditUserRolesResourceImpl.class);
        
        // Add a route for Get User resource
        final String getUser = this.getPropertyUtil().get(OasProps.PROP_GET_USER_PATH, OasProps.DEF_GET_USER_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching get user service to path=" + getUser);
        }
        router.attach(getUser, UserResourceImpl.class);
        
        // Add a route for Index page
        final String indexPath = this.getPropertyUtil().get(OasProps.PROP_INDEX_PATH, OasProps.DEF_INDEX_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching index page to path=" + indexPath);
        }
        router.attach(indexPath, IndexResourceImpl.class);
        
        // Add a route for Help subpage service
        final String helpSubPagePath =
                this.getPropertyUtil().get(OasProps.PROP_HELP_SUBPAGE_PATH, OasProps.DEF_HELP_SUBPAGE_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching helpSubPagePath page to path=" + helpSubPagePath);
        }
        router.attach(helpSubPagePath, HelpResourceImpl.class);
        
        // Add a route for Help base page service
        final String helpBasePath =
                this.getPropertyUtil().get(OasProps.PROP_HELP_BASE_PATH, OasProps.DEF_HELP_BASE_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching helpBasePath page to path=" + helpBasePath);
        }
        router.attach(helpBasePath, HelpResourceImpl.class);
        
        // Add a route for Audit Log service
        final String auditlogPath = this.getPropertyUtil().get(OasProps.PROP_AUDITLOG_PATH, OasProps.DEF_AUDITLOG_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching auditlog page to path=" + auditlogPath);
        }
        router.attach(auditlogPath, AuditLogResourceImpl.class);
        
        // Add a route for WhoAmI service
        final String whoamiPath = this.getPropertyUtil().get(OasProps.PROP_WHOAMI_PATH, OasProps.DEF_WHOAMI_PATH);
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching whoami page to path=" + whoamiPath);
        }
        router.attach(whoamiPath, WhoAmIResourceImpl.class);
        
        // Add a route for the testannotations page
        final String testAnnotationPath = "testannotations";
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching test annotation page to path=" + testAnnotationPath);
        }
        router.attach(testAnnotationPath, TestAnnotationResource.class);
        
        // Add a route for the testala page
        final String testAnnotationAlaPath = "testala";
        if(this.log.isDebugEnabled())
        {
            this.log.debug("attaching test ala page to path=" + testAnnotationAlaPath);
        }
        router.attach(testAnnotationAlaPath, TestAlaResource.class);
        
        // // Add a route for Logout service
        // final String logout =
        // PropertyUtils.getProperty(PropertyUtils.PROPERTY_LOGOUT_FORM_PATH,
        // PropertyUtils.DEFAULT_LOGOUT_FORM_PATH);
        //
        // if(this.log.isDebugEnabled())
        // {
        // this.log.info("attaching logout service to path=" + logout);
        // }
        
        // FIXME: Switch between the logout resource implementations here based on the authenticator
        // router.attach(logout, CookieLogoutResourceImpl.class);
        
        this.log.info("routes={}", router.getRoutes().toString());
        
        // put the authenticator in front of the resource router so it can handle challenge
        // responses and forward them on to the right location after locking in the authentication
        // data Authentication of individual methods on individual resources is handled using calls
        // to OasWebServiceApplication.authenticate
        authenticator.setNext(router);
        
        corsFilter.setNext(authenticator);
        
        return corsFilter;
    }
    
    /**
     * Fetches a ChallengeAuthenticator based on the key defined in
     * PropertyUtils.PROPERTY_CHALLENGE_AUTH_METHOD.
     * 
     * Currently defaults to a DigestAuthenticator.
     * 
     * @return A ChallengeAuthenticator that can be used to challenge unauthenticated requests to
     *         resources that need authenticated access.
     */
    @Override
    public ChallengeAuthenticator getAuthenticator()
    {
        return this.auth;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see com.github.ansell.oas.webservice.impl.OasWebServiceApplicationInterface#getDataHandler()
     */
    @Override
    public DataHandler getDataHandler()
    {
        return this.dataHandler;
    }
    
    @Override
    public PropertyUtil getPropertyUtil()
    {
        if(this.propertyUtil == null)
        {
            synchronized(this)
            {
                if(this.propertyUtil == null)
                {
                    this.log.trace("Dumping context attributes...");
                    for(String parameter : getContext().getAttributes().keySet())
                    {
                        this.log.trace("context attribute: <{}> <{}>", parameter,
                                getContext().getAttributes().get(parameter));
                    }
                    
                    this.log.trace("Dumping context parameters...");
                    for(Parameter parameter : getContext().getParameters())
                    {
                        this.log.trace("context parameter: <{}> <{}>", parameter.getName(), parameter.getValue());
                    }
                    
                    String propertyBundleName =
                            this.getContext().getParameters().getFirstValue("oasproperties", true, "oas");
                    this.log.info("Setting up property util using: {}", propertyBundleName);
                    this.propertyUtil = new PropertyUtil(propertyBundleName);
                }
            }
        }
        
        return this.propertyUtil;
    }
    
    @Override
    public RestletUtilSesameRealm getRealm()
    {
        return this.realm;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see
     * com.github.ansell.oas.webservice.impl.OasWebServiceApplicationInterface#getTemplateConfiguration
     * ()
     */
    @Override
    public Configuration getTemplateConfiguration()
    {
        return this.freemarkerConfiguration;
    }
    
    /**
     * @param auth
     *            the auth to set
     */
    @Override
    public void setAuthenticator(final ChallengeAuthenticator auth)
    {
        this.auth = auth;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see
     * com.github.ansell.oas.webservice.impl.OasWebServiceApplicationInterface#setDataHandler(net
     * .maenad .oas.data.DataHandler)
     */
    @Override
    public void setDataHandler(final DataHandler nextDataHandler)
    {
        this.dataHandler = nextDataHandler;
    }
    
    @Override
    public void setRealm(final RestletUtilSesameRealm nextRealm)
    {
        this.realm = nextRealm;
    }
    
    @Override
    public void setTemplateConfiguration(final Configuration nextFreemarkerConfiguration)
    {
        this.freemarkerConfiguration = nextFreemarkerConfiguration;
    }
    
}
