/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.io.ByteArrayOutputStream;
import java.util.Arrays;
import java.util.Map;
import java.util.concurrent.ConcurrentMap;

import org.openrdf.model.URI;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.ByteArrayRepresentation;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.objects.ResourceCount;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasConstants;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.OntologyBrowserResource;

/**
 * A resource that can be used to browse ontologies.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class OntologyBrowserResourceImpl extends AbstractOasResourceImpl implements OntologyBrowserResource
{
    private final Logger log = LoggerFactory.getLogger(OntologyBrowserResourceImpl.class);
    
    /**
     * 
     */
    public OntologyBrowserResourceImpl()
    {
        super();
    }
    
    @Override
    public Representation getHtml(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_READ);
        
        // if(entity == null)
        // {
        // this.log.error("Entity was null variant={}", variant);
        // }
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        
        final String ontologyUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        this.log.debug("ontologyUriString={}", ontologyUriString);
        
        if(ontologyUriString == null)
        {
            this.log.error("No ontology URI specified");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "No ontology URI specified in request");
        }
        
        final URI ontologyUri = this.getDataHandler().createURI(ontologyUriString);
        
        Ontology nextOntology = null;
        
        final String managedOntologyVersionUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyVersionUri", true);
        
        if(managedOntologyVersionUriString != null)
        {
            final URI managedOntologyVersionUri = this.getDataHandler().createURI(managedOntologyVersionUriString);
            
            nextOntology = this.getDataHandler().getVersionedOntology(ontologyUri, managedOntologyVersionUri);
        }
        
        // backup is to attempt to get the latest ontology for the ontology URI
        if(nextOntology == null)
        {
            nextOntology = this.getDataHandler().getLatestOntology(ontologyUri);
        }
        
        // if the nextOntology is still null then return early with an exception
        if(nextOntology == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find the specified ontology");
        }
        
        // TODO: In the AJAX version this setting is not supported here, and would be better
        // supported as a javascript function
        boolean recursive = false;
        
        final String directChildrenOnlyString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("directChildrenOnly", true);
        
        if(directChildrenOnlyString != null)
        {
            recursive = !Boolean.valueOf(directChildrenOnlyString);
        }
        
        URI parentUri;
        
        final String parentUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("parentUri", true);
        
        // Default to browsing from OWL:Thing
        if(parentUriString == null)
        {
            parentUri = OasConstants.OWL_THING;
        }
        else
        {
            parentUri = this.getDataHandler().createURI(parentUriString);
        }
        
        final ConcurrentMap<URI, ResourceCount> directChildrenCount =
                this.getDataHandler().getChildrenCount(nextOntology, Arrays.asList(parentUri), false);
        final ConcurrentMap<URI, ResourceCount> allChildrenCount =
                this.getDataHandler().getChildrenCount(nextOntology, Arrays.asList(parentUri), true);
        
        final int annotationCounts =
                this.getDataHandler().getAnnotationCountsByOntologyTermUri(parentUri, null, false).getCount();
        
        dataModel.put("ontology", nextOntology);
        dataModel.put("parentUri", parentUri);
        
        dataModel.put("childrenCount", directChildrenCount.get(parentUri).getCount());
        dataModel.put("allChildrenCount", allChildrenCount.get(parentUri).getCount());
        dataModel.put("annotationCounts", annotationCounts);
        
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYBROWSER, OasProps.DEF_TMPL_ONTOLOGYBROWSER));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYBROWSER_TITLE,
                        OasProps.DEF_TMPL_ONTOLOGYBROWSER_TITLE));
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see
     * com.github.ansell.oas.webservice.OntologyResource#getRdf(org.restlet.representation.Variant)
     */
    @Override
    public Representation getRdf(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_READ);
        
        return this.getRdfInternal(variant);
    }
    
    private Representation getRdfInternal(final Variant variant)
    {
        final String ontologyUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        if(ontologyUriString == null)
        {
            this.log.error("ontologyUri parameter was not present");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "ontologyUri parameter must not be empty");
        }
        
        final String ontologyVersionUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyVersionUri", true);
        
        final URI ontologyUri = this.getDataHandler().createURI(ontologyUriString);
        
        URI parentUri = OasConstants.OWL_THING;
        
        final String parentUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("parentUri", true);
        
        if(parentUriString != null)
        {
            parentUri = this.getDataHandler().createURI(parentUriString);
        }
        
        boolean recursive = false;
        
        final String directChildrenOnlyString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("directChildrenOnly", true);
        
        if(directChildrenOnlyString != null)
        {
            recursive = !Boolean.valueOf(directChildrenOnlyString);
        }
        
        boolean addCounts = true;
        
        final String addCountsString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("addCounts", true);
        
        if(addCountsString != null)
        {
            addCounts = Boolean.valueOf(addCountsString);
        }
        
        Ontology result = null;
        
        if(ontologyVersionUriString != null)
        {
            this.log.debug("version was specified, so returning all data for the given version if possible");
            
            final URI ontologyVersionUri = this.getDataHandler().createURI(ontologyVersionUriString);
            
            result = this.getDataHandler().getVersionedOntology(ontologyUri, ontologyVersionUri);
            
        }
        
        // Fallback to the latest version if they didn't specify a version or we didn't successfully
        // find the version
        if(result == null)
        {
            result = this.getDataHandler().getLatestOntology(ontologyUri);
        }
        
        // if the result is still null then return early with an exception
        if(result == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find the specified ontology");
        }
        
        // TODO: Determine how to return subsets of the ontology using RDF. Possibly use CBD and
        // SPARQL DESCRIBE, but the behaviour of DESCRIBE is not configurable due to SPARQL team not
        // thinking it would be useful.
        final ByteArrayOutputStream stream = new ByteArrayOutputStream();
        
        if(parentUriString == null)
        {
            this.log.debug("Downloading entire ontology");
            this.getDataHandler().downloadOntologyWithInferences(result, stream, variant.getMediaType().getName());
        }
        else
        {
            this.log.debug("Downloading ontology subset with parent: {}", parentUriString);
            this.getDataHandler().downloadOntologySubset(result, stream, variant.getMediaType().getName(), parentUri,
                    recursive, addCounts);
        }
        
        return new ByteArrayRepresentation(stream.toByteArray());
    }
    
}
