/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.openrdf.model.URI;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFHandlerException;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.ByteArrayRepresentation;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.utils.CurrentVersionUtils;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.OntologyUtils;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.OntologyManagerResource;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class OntologyManagerResourceImpl extends AbstractOasResourceImpl implements OntologyManagerResource
{
    private final Logger log = LoggerFactory.getLogger(OntologyManagerResourceImpl.class);
    
    /**
     * 
     */
    public OntologyManagerResourceImpl()
    {
        super();
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see com.github.ansell.oas.webservice.OntologyResource#deleteOntology()
     */
    @Override
    public void deleteOntology() throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_DELETE);
        
        OntologyUtils.deleteOntologyInternalQueryForm(this.getRequest(), this.getResponse(), this.getDataHandler());
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see
     * com.github.ansell.oas.webservice.OntologyResource#getHtml(org.restlet.representation.Variant)
     */
    @Override
    public Representation getHtml(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_READ);
        
        // if(entity == null)
        // {
        // this.log.error("Entity was null variant={}", variant);
        // }
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        dataModel.put("currentversionutil", new CurrentVersionUtils(this.getDataHandler()));
        
        final String ontologyUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        this.log.debug("ontologyUriString={}", ontologyUriString);
        
        final String managedOntologyVersionUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyVersionUri", true);
        
        final String startOffsetString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("offset", true);
        
        final String limitString = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("limit", true);
        
        if(ontologyUriString != null && managedOntologyVersionUriString != null)
        {
            final URI ontologyUri = this.getDataHandler().createURI(ontologyUriString);
            final URI managedOntologyVersionUri = this.getDataHandler().createURI(managedOntologyVersionUriString);
            
            final Ontology versionedOntology =
                    this.getDataHandler().getVersionedOntology(ontologyUri, managedOntologyVersionUri);
            
            if(versionedOntology == null)
            {
                throw new ResourceException(Status.CLIENT_ERROR_NOT_FOUND, "Could not find given ontology");
            }
            
            dataModel.put("ontologies", Collections.singletonList(versionedOntology));
        }
        else if(ontologyUriString != null)
        {
            final URI ontologyUri = this.getDataHandler().createURI(ontologyUriString);
            dataModel.put("ontologies", this.getDataHandler().getOntologiesByUri(ontologyUri));
        }
        else
        {
            int startOffset = 0;
            int limit = 20;
            if(startOffsetString != null)
            {
                try
                {
                    startOffset = Integer.parseInt(startOffsetString, 10);
                }
                catch(final NumberFormatException nfe)
                {
                    this.log.error("Could not parse startOffset={}", startOffsetString);
                    startOffset = 0;
                }
            }
            
            if(limitString != null)
            {
                try
                {
                    limit = Integer.parseInt(limitString, 10);
                }
                catch(final NumberFormatException nfe)
                {
                    this.log.error("Could not parse limit={}", limitString);
                    limit = 20;
                }
            }
            
            if(limit > 100)
            {
                this.log.warn("Requested limit was greater than 100, resetting it to 100 original limit={}", limit);
                limit = 100;
            }
            
            final List<Ontology> baseOntologies = this.getDataHandler().getOntologies(startOffset, limit);
            
            this.log.debug("Returning metadata about the following ontologies: {}", baseOntologies);
            
            final List<Ontology> allOntologies = new ArrayList<Ontology>();
            
            for(final Ontology nextOntology : baseOntologies)
            {
                if(nextOntology != null && nextOntology.getOntologyUri() != null)
                {
                    final List<Ontology> list = this.getDataHandler().getOntologiesByUri(nextOntology.getOntologyUri());
                    
                    if(list != null)
                    {
                        allOntologies.addAll(list);
                    }
                }
            }
            
            Collections.sort(allOntologies);
            
            dataModel.put("ontologies", allOntologies);
        }
        
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER, OasProps.DEF_TMPL_ONTOLOGYMANAGER));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_TITLE,
                        OasProps.DEF_TMPL_ONTOLOGYMANAGER_TITLE));
        
        if(this.getDataHandler().getOntologies(0, 1).isEmpty())
        {
            dataModel.put("hasOntologies", true);
        }
        else
        {
            dataModel.put("hasOntologies", false);
        }
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    private Map<URI, URI> getLatestVersionsForOntologies(final Collection<Ontology> ontologies)
    {
        // eliminate duplicates in the given list
        final Set<URI> ontologyUris = new HashSet<URI>(ontologies.size());
        
        for(final Ontology nextOntology : ontologies)
        {
            ontologyUris.add(nextOntology.getOntologyUri());
        }
        
        return this.getLatestVersionsForOntologyUris(ontologyUris);
    }
    
    private Map<URI, URI> getLatestVersionsForOntologyUris(final Collection<URI> ontologies)
    {
        final Map<URI, URI> latestOntologyVersions = new HashMap<URI, URI>();
        
        for(final URI ontologyUri : ontologies)
        {
            final Ontology latestOntology = this.getDataHandler().getLatestOntology(ontologyUri);
            
            if(latestOntology != null && latestOntology.getOntologyUri() != null
                    && latestOntology.getOntologyVersionUri() != null)
            {
                latestOntologyVersions.put(latestOntology.getOntologyUri(), latestOntology.getOntologyVersionUri());
            }
        }
        
        return latestOntologyVersions;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see
     * com.github.ansell.oas.webservice.OntologyResource#getRdf(org.restlet.representation.Variant)
     */
    @Override
    public Representation getRdf(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_READ);
        
        final String ontologyUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        if(ontologyUriString == null)
        {
            this.log.debug("No ontology URI specified, returning directory information about the known ontologies");
            // getResponse().setStatus(Status.SERVER_ERROR_NOT_IMPLEMENTED);
            // return new AppendableRepresentation(
            // "<error>No default RDF/XML representation for ontology manager</error>",
            // variant.getMediaType());
            
            // FIXME: Remove hardcoded offset and limit
            final List<Ontology> baseOntologies = this.getDataHandler().getOntologies(0, 100);
            
            this.log.debug("Returning metadata about the following ontologies: {}", baseOntologies);
            
            final List<Ontology> allOntologies = new ArrayList<Ontology>();
            
            for(final Ontology nextOntology : baseOntologies)
            {
                allOntologies.addAll(this.getDataHandler().getOntologiesByUri(nextOntology.getOntologyUri()));
            }
            
            Collections.sort(allOntologies);
            
            final Map<URI, URI> latestOntologyVersions = this.getLatestVersionsForOntologies(allOntologies);
            
            // TODO: Check authentication for this query using a new OasAction for directory
            // information to enable blocking of directory listing using this method to
            // unauthenticated or underprivileged users
            try
            {
                return RestletUtils.getRdfRepresentationShortOntologies(allOntologies, latestOntologyVersions,
                        variant.getMediaType(), this.getRequest().getResourceRef());
            }
            catch(RepositoryException | RDFHandlerException e)
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Could not generate ontology list");
            }
        }
        else
        {
            return this.getRdfInternal(variant);
        }
    }
    
    private Representation getRdfInternal(final Variant variant)
    {
        final String ontologyUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyUri", true);
        
        if(ontologyUriString == null)
        {
            this.log.error("ontologyUri parameter was not present");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "ontologyUri parameter must not be empty");
        }
        
        final String ontologyVersionUriString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("ontologyVersionUri", true);
        
        final URI ontologyUri = this.getDataHandler().createURI(ontologyUriString);
        
        Ontology result = null;
        
        if(ontologyVersionUriString != null)
        {
            this.log.debug("version was specified, so returning all data for the given version if possible");
            
            final URI ontologyVersionUri = this.getDataHandler().createURI(ontologyVersionUriString);
            
            final List<Ontology> ontologiesByUri = this.getDataHandler().getOntologiesByUri(ontologyUri);
            
            for(final Ontology nextOntology : ontologiesByUri)
            {
                final URI nextOntologyVersionUri = nextOntology.getOntologyVersionUri();
                
                if(ontologyVersionUri.equals(nextOntologyVersionUri))
                {
                    result = nextOntology;
                    break;
                }
            }
            
            // if they said they wanted a specific version and it was not found fallback to the
            // latest ontology with the given URI to provide forwards compatibility instead of
            // throwing an error
            if(result == null)
            {
                result = this.getDataHandler().getLatestOntology(ontologyUri);
            }
            
            // if the result is still null then return early with an exception
            if(result == null)
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                        "Could not find the specified ontology, including the version IRI as specified");
            }
            
            // try
            // {
            // FIXME: Is there anyway of just streaming the ontology straight out using Restlet??
            final ByteArrayOutputStream stream = new ByteArrayOutputStream();
            this.getDataHandler().downloadOntologyWithInferences(result, stream, variant.getMediaType().getName());
            
            return new ByteArrayRepresentation(stream.toByteArray());
            // return RestletUtils.getRdfRepresentation(result, variant.getMediaType(),
            // this.getRequest()
            // .getResourceRef());
            // }
            // catch(final RepositoryException e)
            // {
            // throw new ResourceException(Status.SERVER_ERROR_INTERNAL, e);
            // }
            // catch(final RDFHandlerException e)
            // {
            // throw new ResourceException(Status.SERVER_ERROR_INTERNAL, e);
            // }
            
        }
        else
        {
            this.log.debug("Returning metadata for all ontologies with the given ontology URI, as no version was specified, so we assume they are browsing the ontology manager but have not yet decided on a version");
            // version was not specified, so dump the metadata for the available versions to the
            // results, without dumping the actual ontologies in this case, they will need to choose
            // a version from the list to proceed
            
            // TODO: Make this behaviour configurable to switch between returning metadata about all
            // ontologies to returning the full ontology for the latest version
            
            final List<Ontology> ontologiesByUri = this.getDataHandler().getOntologiesByUri(ontologyUri);
            
            if(ontologiesByUri.isEmpty())
            {
                throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                        "Could not find the specified ontology, and no version was specified");
            }
            
            final Map<URI, URI> latestOntologyVersions = this.getLatestVersionsForOntologies(ontologiesByUri);
            
            try
            {
                return RestletUtils.getRdfRepresentationShortOntologies(ontologiesByUri, latestOntologyVersions,
                        variant.getMediaType(), this.getRequest().getResourceRef());
            }
            catch(final RepositoryException e)
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL, e);
            }
            catch(final RDFHandlerException e)
            {
                throw new ResourceException(Status.SERVER_ERROR_INTERNAL, e);
            }
            
        }
        
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see
     * com.github.ansell.oas.webservice.OntologyResource#putOntologyRdf(org.restlet.representation
     * .Representation)
     */
    @Override
    public Representation putOntologyRdf(final Representation entity) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_CREATE);
        
        if(entity == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Ontology cannot be null");
        }
        
        try
        {
            return OntologyUtils.putOntologyRdfInternal(this.getRequest(), this.getResponse(), this.getDataHandler(),
                    entity.getStream(), entity.getMediaType());
        }
        catch(final IOException e)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, e);
        }
    }
    
}
