/**
 * Package for implementations of the resource handlers in the OAS API
 */
package com.github.ansell.oas.webservice.impl;

import java.util.Collection;
import java.util.Map;

import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFHandlerException;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Annotation;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.utils.UserUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.SearchAnnotationResource;

/**
 * Searches for a text string in annotations
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class SearchAnnotationResourceImpl extends AbstractOasResourceImpl implements SearchAnnotationResource
{
    /**
     * Default constructor for Annotation Resource Implementation.
     */
    public SearchAnnotationResourceImpl()
    {
        super();
    }
    
    @Override
    public Representation getHtml(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final String inputSearch =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("searchterm", true);
        
        if(inputSearch != null)
        {
            // Get the annotation based on the identifier URI and load it into memory
            final Collection<Annotation> annotations = this.getDataHandler().searchAnnotations(inputSearch, 1000, 0);
            
            final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
            dataModel
                    .put("contentTemplate",
                            this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST,
                                    OasProps.DEF_TMPL_ANNOTATIONS_LIST));
            dataModel.put(
                    "pageTitle",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_LIST_TITLE,
                            OasProps.DEF_TMPL_ANNOTATIONS_LIST_TITLE));
            
            dataModel.put("annotations", annotations);
            dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
            dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
            
            return RestletUtils.getHtmlRepresentation(
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                    MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
        }
        else
        {
            final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
            dataModel.put(
                    "contentTemplate",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_SEARCH,
                            OasProps.DEF_TMPL_ANNOTATIONS_SEARCH));
            dataModel.put(
                    "pageTitle",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_ANNOTATIONS_SEARCH_TITLE,
                            OasProps.DEF_TMPL_ANNOTATIONS_SEARCH_TITLE));
            
            dataModel.put("userutil", new UserUtils(this.getOasApplication().getRealm()));
            
            return RestletUtils.getHtmlRepresentation(
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                    MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
            
        }
    }
    
    @Override
    public Representation getRdf(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ANNOTATION_READ);
        
        final String inputSearch =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("searchterm", true);
        
        // Get the annotation based on the identifier URI and load it into memory
        final Collection<Annotation> annotations = this.getDataHandler().searchAnnotations(inputSearch, 1000, 0);
        
        if(inputSearch == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST,
                    "Did not find a searchTerm parameter in the request");
        }
        
        try
        {
            return RestletUtils.getRdfRepresentation(annotations, variant.getMediaType(), this.getRequest()
                    .getResourceRef());
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final RDFHandlerException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
    }
    
}
