/**
 * Package for implementations of the resource handlers in the OAS API
 */
package com.github.ansell.oas.webservice.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFHandlerException;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.Ontology;
import com.github.ansell.oas.objects.SearchResult;
import com.github.ansell.oas.utils.CurrentVersionUtils;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasConstants;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.utils.SesameUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.SearchOntologyLabelsResource;

/**
 * Searches for a text string in ontology labels
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class SearchOntologyLabelsResourceImpl extends AbstractOasResourceImpl implements SearchOntologyLabelsResource
{
    /**
     * Default constructor for Search Ontology Labels.
     */
    public SearchOntologyLabelsResourceImpl()
    {
        super();
    }
    
    @Override
    public Representation getHtml(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_READ);
        
        final String inputSearch =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("searchterm", true);
        
        if(inputSearch != null)
        {
            final List<SearchResult> searchResults = this.getDataHandler().searchOntologyLabels(inputSearch, 1000, 0);
            
            final List<Ontology> ontologies = new ArrayList<Ontology>(searchResults.size());
            
            for(final SearchResult nextResult : searchResults)
            {
                ontologies.add(nextResult.getOntology());
            }
            
            final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
            dataModel.put("contentTemplate",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER, OasProps.DEF_TMPL_ONTOLOGYMANAGER));
            dataModel.put(
                    "pageTitle",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_TITLE,
                            OasProps.DEF_TMPL_ONTOLOGYMANAGER_TITLE));
            
            dataModel.put("ontologies", ontologies);
            dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
            dataModel.put("currentversionutil", new CurrentVersionUtils(this.getDataHandler()));
            
            if(this.getDataHandler().getOntologies(0, 1).isEmpty())
            {
                dataModel.put("hasOntologies", true);
            }
            else
            {
                dataModel.put("hasOntologies", false);
            }
            
            return RestletUtils.getHtmlRepresentation(
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                    MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
        }
        else
        {
            final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
            dataModel.put(
                    "contentTemplate",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_LABELSEARCH,
                            OasProps.DEF_TMPL_ONTOLOGYMANAGER_LABELSEARCH));
            dataModel.put(
                    "pageTitle",
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_LABELSEARCH_TITLE,
                            OasProps.DEF_TMPL_ONTOLOGYMANAGER_LABELSEARCH_TITLE));
            dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
            
            return RestletUtils.getHtmlRepresentation(
                    this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                    MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
            
        }
    }
    
    @Override
    public Representation getRdf(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_READ);
        
        final String inputSearch =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("searchterm", true);
        
        if(inputSearch == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "No search term specified");
        }
        
        try
        {
            final List<SearchResult> searchResults = this.getDataHandler().searchOntologyLabels(inputSearch, 1000, 0);
            
            final Repository repository =
                    SesameUtils.toRDFRepository(searchResults, OasConstants.OAS_ONTOLOGY_SEARCH_RESULT, null);
            
            return RestletUtils.toRDFSerialisation(variant.getMediaType().getName(), repository);
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final RDFHandlerException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
    }
    
}
