/**
 * Package for implementations of the resource handlers in the OAS API
 */
package com.github.ansell.oas.webservice.impl;

import java.util.List;
import java.util.Map;

import org.openrdf.repository.Repository;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFHandlerException;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.ResourceException;

import com.github.ansell.oas.objects.SearchResult;
import com.github.ansell.oas.utils.LabelUtils;
import com.github.ansell.oas.utils.OasConstants;
import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.utils.SesameUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.SearchOntologyResource;

/**
 * Searches for a text string in ontologies.
 * 
 * @author Peter Ansell p_ansell@yahoo.com
 */
public class SearchOntologyResourceImpl extends AbstractOasResourceImpl implements SearchOntologyResource
{
    /**
     * Default constructor for Annotation Resource Implementation.
     */
    public SearchOntologyResourceImpl()
    {
        super();
    }
    
    @Override
    public Representation getHtml(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_READ);
        
        final String inputSearch =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("searchterm", true);
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put(
                "contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_SEARCH,
                        OasProps.DEF_TMPL_ONTOLOGYMANAGER_SEARCH));
        dataModel.put(
                "pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_ONTOLOGYMANAGER_SEARCH_TITLE,
                        OasProps.DEF_TMPL_ONTOLOGYMANAGER_SEARCH_TITLE));
        dataModel.put("labelutil", new LabelUtils(this.getDataHandler()));
        
        if(inputSearch != null)
        {
            final List<SearchResult> searchResults = this.getDataHandler().searchAllOntologies(inputSearch, 1000, 0);
            dataModel.put("searchResults", searchResults);
        }
        
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
        
    }
    
    @Override
    public Representation getRdf(final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.ONTOLOGY_READ);
        
        final String inputSearch =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("searchterm", true);
        
        // Get the annotation based on the identifier URI and load it into memory
        final List<SearchResult> searchResults = this.getDataHandler().searchAllOntologies(inputSearch, 1000, 0);
        
        try
        {
            final Repository repository =
                    SesameUtils.toRDFRepository(searchResults, OasConstants.OAS_ONTOLOGY_SEARCH_RESULT, null);
            
            return RestletUtils.toRDFSerialisation(variant.getMediaType().getName(), repository);
        }
        catch(final RepositoryException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
        catch(final RDFHandlerException ex)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, ex);
        }
    }
    
}
