/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.util.Map;

import org.openrdf.model.URI;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.ext.html.FormDataSet;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.Get;
import org.restlet.resource.Post;
import org.restlet.resource.ResourceException;
import org.restlet.security.User;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.UpdateUserResource;
import com.github.ansell.restletutils.RestletUtilRoles;
import com.github.ansell.restletutils.RestletUtilUser;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class UpdateUserResourceImpl extends AbstractOasResourceImpl implements UpdateUserResource
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    /**
     * 
     */
    public UpdateUserResourceImpl()
    {
        super();
    }
    
    private void checkUpdateUserAuthentication(String usernameString) throws ResourceException
    {
        if(this.getRequest().getClientInfo().isAuthenticated())
        {
            if(this.getRequest().getClientInfo().getUser().getIdentifier().equals(usernameString))
            {
                // Always allow users to update their own information
                return;
            }
        }
        
        // Check for user create permission if the update would be on another user
        this.checkAuthentication(BaseOasActions.USER_CREATE);
        
        // For system safety, all user updates must be authenticated, even if the
        // OasAction.USER_CREATE is setup to allow users to register themselves into the system
        if(!this.getRequest().getClientInfo().isAuthenticated())
        {
            this.log.error("User was not authenticated");
            throw new ResourceException(Status.CLIENT_ERROR_UNAUTHORIZED,
                    "Only authenticated users can update user details");
        }
    }
    
    @Override
    @Get("html")
    public Representation getUpdatePageHtml(final Representation entity) throws ResourceException
    {
        this.log.debug("in getUpdatePageHtml");
        
        // Include parameter for the username in the query entity
        final String usernameString =
                this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("username", true);
        
        // If the username was not submitted default to updating the current user information
        if(usernameString == null || usernameString.trim().isEmpty())
        {
            this.log.error("Username was not submitted");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Username was not submitted");
        }
        
        this.checkUpdateUserAuthentication(usernameString);
        
        final RestletUtilUser findUser = this.getOasApplication().getRealm().findUser(usernameString);
        
        if(findUser == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "User does not exist");
        }
        
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_USER_UPDATE, OasProps.DEF_TMPL_USER_UPDATE));
        dataModel.put("pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_USER_UPDATE_TITLE, OasProps.DEF_TMPL_USER_UPDATE_TITLE));
        
        dataModel.put("updateUser", findUser);
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    @Post
    public Representation updateUserByPost(final Representation entity, final Variant variant) throws ResourceException
    {
        this.log.debug("in updateUserByPost");
        
        final FormDataSet form = new FormDataSet(entity);
        final String usernameString = form.getEntries().getFirstValue("username", true);
        final String password = form.getEntries().getFirstValue("password", true);
        String email = form.getEntries().getFirstValue("email", true);
        String firstName = form.getEntries().getFirstValue("firstname", true);
        String lastName = form.getEntries().getFirstValue("lastname", true);
        
        if(usernameString == null || usernameString.trim().isEmpty())
        {
            this.log.error("Username was not submitted");
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Username was not submitted");
        }
        
        this.checkUpdateUserAuthentication(usernameString);
        
        final User existingUser = this.getOasApplication().getRealm().findUser(usernameString);
        
        if(existingUser != null)
        {
            final char[] secret;
            
            // If they did not submit a password reuse the existing secret
            if(password == null)
            {
                secret = existingUser.getSecret();
                
                if(secret.length < 6)
                {
                    this.log.error("Password was not submitted, and existing secret was not long enough. Must be 6 characters long");
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Password was not submitted");
                }
            }
            else
            {
                // Trim the password to avoid whitespace errors
                final String trimmedPassword = password.trim();
                
                if(trimmedPassword.length() < 6)
                {
                    this.log.error("Password was not long enough. Must be 6 characters long");
                    throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Password was not long enough");
                }
                
                secret = trimmedPassword.toCharArray();
            }
            
            // Allow them to leave email out in a valid request
            if(email == null)
            {
                email = existingUser.getEmail();
            }
            
            // Allow them to leave first name out in a valid request
            if(firstName == null)
            {
                firstName = existingUser.getFirstName();
            }
            
            // Allow them to leave last name out in a valid request
            if(lastName == null)
            {
                lastName = existingUser.getLastName();
            }
            
            // Update user information always allows for a user to their own information, and for an
            // admin to any user account
            if(this.getRequest().getClientInfo().getUser().getIdentifier().equalsIgnoreCase(usernameString)
                    || this.getRequest().getClientInfo().getRoles().contains(RestletUtilRoles.ADMIN.getRole()))
            {
                
                final URI updateUser = this.getOasApplication().getRealm().deleteUser(existingUser);
                
                if(updateUser != null)
                {
                    final RestletUtilUser newUser =
                            new RestletUtilUser(usernameString, secret, firstName, lastName, email);
                    
                    final URI addUser = this.getOasApplication().getRealm().addUser(newUser);
                    
                    if(addUser == null)
                    {
                        this.log.error("Failed to update user in add step");
                        throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                                "Could not create user for some reason");
                    }
                    
                    this.redirectToPath(this.getPropertyUtil().get(OasProps.PROP_GET_USER_PATH,
                            OasProps.DEF_GET_USER_PATH));
                    
                    return null;
                }
                else
                {
                    this.log.error("Failed to update user in delete step");
                    throw new ResourceException(Status.SERVER_ERROR_INTERNAL,
                            "An error occurred updating the user information");
                }
            }
            else
            {
                this.log.error("Non-admin users are not allowed to modify other users information");
                throw new ResourceException(Status.CLIENT_ERROR_FORBIDDEN, "Not allowed to update user information");
            }
        }
        else
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find user to update");
        }
    }
}
