/**
 * 
 */
package com.github.ansell.oas.webservice.impl;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Map;

import org.openrdf.model.URI;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.ext.html.FormDataSet;
import org.restlet.representation.Representation;
import org.restlet.representation.Variant;
import org.restlet.resource.Delete;
import org.restlet.resource.Get;
import org.restlet.resource.Put;
import org.restlet.resource.ResourceException;
import org.restlet.security.User;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.github.ansell.oas.utils.OasProps;
import com.github.ansell.oas.utils.RestletUtils;
import com.github.ansell.oas.webservice.BaseOasActions;
import com.github.ansell.oas.webservice.UserResource;
import com.github.ansell.restletutils.RestletUtilRoles;
import com.github.ansell.restletutils.RestletUtilUser;

/**
 * @author Peter Ansell p_ansell@yahoo.com
 * 
 */
public class UserResourceImpl extends AbstractOasResourceImpl implements UserResource
{
    private final Logger log = LoggerFactory.getLogger(this.getClass());
    
    private final Path tempDirectory;
    
    /**
     * 
     */
    public UserResourceImpl()
    {
        super();
        
        try
        {
            this.tempDirectory = Files.createTempDirectory("oas-createuseruploads");
        }
        catch(final IOException e)
        {
            throw new RuntimeException("Could not create temporary directory", e);
        }
    }
    
    @Override
    @Delete
    public Representation deleteUser(final Representation entity) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.USER_DELETE);
        
        final String username = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("username");
        
        final RestletUtilUser findUser = this.getOasApplication().getRealm().findUser(username);
        
        if(findUser == null)
        {
            this.log.error("Could not delete user as they were not found: {}", username);
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Could not find user");
        }
        
        // TODO: Redirect user based on the URI that was returned
        // Returns the URI of the deleted user, or null if no user was found
        final URI deleteUser = this.getOasApplication().getRealm().deleteUser(findUser);
        
        if(deleteUser != null)
        {
            this.getResponse().setStatus(Status.SUCCESS_NO_CONTENT, "User deleted");
            
            return null;
        }
        else
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Could not delete user");
        }
    }
    
    @Override
    @Get("html")
    public Representation getUserHtml(final Representation entity, final Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.USER_GET);
        
        return this.getUserInfoInternal(entity, variant);
    }
    
    private Representation getUserInfoInternal(final Representation entity, final Variant variant)
        throws ResourceException
    {
        final Map<String, Object> dataModel = RestletUtils.getBaseDataModel(this.getRequest());
        
        // FIXME: When checkAuthentication supports checking authentication based on the user
        // requesting, and the user being requested, change this to support it
        // For now, all requests for getUserInfo require authentication, and if the user is not
        // requesting their own information, they must be an administrator.
        if(!this.getRequest().getClientInfo().isAuthenticated())
        {
            this.log.error("Must be authenticated to request user information");
            throw new ResourceException(Status.CLIENT_ERROR_FORBIDDEN,
                    "Must be authenticated to request user information");
        }
        else
        {
            final String username = this.getRequest().getResourceRef().getQueryAsForm().getFirstValue("username");
            
            if(this.getRequest().getClientInfo().getRoles().contains(RestletUtilRoles.ADMIN.getRole()))
            {
                if(username == null)
                {
                    // list all users
                    // dataModel.put("requestedUsers",
                    // this.getOasApplication().getRealm().getUsers());
                    
                    dataModel.put("requestedUsers", this.getOasApplication().getRealm().getUsers());
                }
                else
                {
                    // administrators can request all user information
                    // Make sure we cast to Restlet User class so that Arrays.asList does not
                    // generate a list with a different generics type that may confuse Freemarker
                    final User requestedUser = this.getOasApplication().getRealm().findUser(username);
                    dataModel.put("requestedUsers", Arrays.asList(requestedUser));
                }
            }
            else if(username == null || username.equals(this.getRequest().getClientInfo().getUser().getIdentifier()))
            {
                dataModel.put("requestedUsers", Arrays.asList(this.getRequest().getClientInfo().getUser()));
            }
            else
            {
                this.log.error("Could not request user information due to missing privileges: {}", this.getRequest()
                        .getClientInfo().getRoles());
                throw new ResourceException(Status.CLIENT_ERROR_FORBIDDEN,
                        "Could not request user information due to missing privileges");
            }
        }
        
        dataModel.put("contentTemplate",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_USER_GET, OasProps.DEF_TMPL_USER_GET));
        dataModel.put("pageTitle",
                this.getPropertyUtil().get(OasProps.PROP_TMPL_USER_GET_TITLE, OasProps.DEF_TMPL_USER_GET_TITLE));
        
        // Output the base template, with contentTemplate from the dataModel defining the template
        // to use for the content in the body of the page
        return RestletUtils.getHtmlRepresentation(
                this.getPropertyUtil().get(OasProps.PROP_TMPL_BASE, OasProps.DEF_TMPL_BASE), dataModel,
                MediaType.TEXT_HTML, this.getOasApplication().getTemplateConfiguration());
    }
    
    @Override
    @Get("rdf")
    public Representation getUserRdfXml(final Representation entity, Variant variant) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.USER_GET);
        
        return this.getUserInfoInternal(entity, variant);
    }
    
    @Override
    @Put
    public Representation putCreateUserMultipart(final Representation entity) throws ResourceException
    {
        this.checkAuthentication(BaseOasActions.USER_CREATE);
        
        final FormDataSet form = new FormDataSet(entity);
        final String username = form.getEntries().getFirstValue("username", true);
        final String password = form.getEntries().getFirstValue("password", true);
        final String email = form.getEntries().getFirstValue("email", true);
        final String firstName = form.getEntries().getFirstValue("firstname", true);
        final String lastName = form.getEntries().getFirstValue("lastname", true);
        
        if(username == null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_BAD_REQUEST, "Username was not submitted");
        }
        
        final User existingUser = this.getOasApplication().getRealm().findUser(username);
        
        if(existingUser != null)
        {
            throw new ResourceException(Status.CLIENT_ERROR_CONFLICT,
                    "Could not create user as there is an existing user with that username");
        }
        
        final RestletUtilUser newUser = new RestletUtilUser(username, password, firstName, lastName, email);
        
        final URI addUser = this.getOasApplication().getRealm().addUser(newUser);
        
        if(addUser == null)
        {
            throw new ResourceException(Status.SERVER_ERROR_INTERNAL, "Could not create user for some reason");
        }
        
        this.redirectToPath(this.getPropertyUtil().get(OasProps.PROP_GET_USER_PATH, OasProps.DEF_GET_USER_PATH));
        
        return null;
    }
    
}
